From Coq Require Import ssreflect ssrfun ssrbool.
Require Import stdpp.tactics.


Class Proset (X : Type) := {pro_le : X -> X -> Prop; pro_pro :> PreOrder pro_le}.
Hint Mode Proset ! : typeclass_instances.
Infix "⊢" := pro_le (no associativity, at level 70).
Arguments pro_le {_ _} !_ !_ /.
Definition core {X} (R : X -> X -> Prop) : X -> X -> Prop := fun a1 a2 => R a1 a2 /\ R a2 a1.
Instance core_sub1 {X} {R : X -> X -> Prop} : subrelation (core R) R | 10.
Proof. firstorder. Qed.
(*
Instance core_sub2 {X} {R : X -> X -> Prop} : subrelation (core R) (flip R) | 10.
Proof. firstorder. Qed.
*)
Instance core_symmetric {X} {R : X -> X -> Prop} : Symmetric (core R).
Proof. firstorder. Qed.
Instance core_equivalence `{PreOrder X R} : Equivalence (core R).
Proof.
  constructor.
  1,2: firstorder.
  move=> a1 a2 a3 [H1 H1'] [H2 H2']; split; etransitivity; eauto.
Qed.
Infix "⟛" := (core pro_le) (no associativity, at level 70).
Lemma core_eq : forall {X} {R : X -> X -> Prop} `{!Antisymmetric X (=) R} {A B},
    core R A B -> A = B.
Proof. firstorder. Qed.
(* TODO Hmmmmm. Is this good to have around? *)
Instance proper_wrt_core `{!Antisymmetric X (=) R} {Y} {f : X -> Y}
  : Proper (core R ==> (=)) f.
Proof. firstorder. Qed.
Record op (A : Type) := Op {get_op : A}.
Add Printing Constructor op.
Arguments Op {_}.
Arguments get_op {_}.
Definition pro_op {X} (R : X -> X -> Prop) : op X -> op X -> Prop :=
  fun a1 a2 => R (get_op a2) (get_op a1).
Arguments pro_op {X} R !a1 !a2 /.
Instance pro_op_pro `{PreOrder X R} : PreOrder (pro_op R).
Proof.
  constructor; first by firstorder.
  move=> [x] [y] [z] /=; etransitivity; eauto.
Qed.
Instance op_proset `{Proset X} : Proset (op X) := {| pro_le := pro_op pro_le |}.
Instance pw_pro {Y} `{PreOrder X R} : PreOrder (pointwise_relation Y R) :=
  {| PreOrder_Reflexive f y := PreOrder_Reflexive (f y);
     PreOrder_Transitive f g h E1 E2 y := PreOrder_Transitive _ _ _ (E1 y) (E2 y) |}.
Instance pw_proset {Y} `{Proset X} : Proset (Y -> X) :=
  {| pro_le := pointwise_relation Y pro_le |}.
Lemma pw_core0 : forall {X} {R : X -> X -> Prop} {Y} {F G : Y -> X},
    core (pointwise_relation Y R) F G <-> (forall A, core R (F A) (G A)).
Proof. firstorder. Qed.
Lemma pw_core : forall `{Proset X} {Y} {F G : Y -> X}, F ⟛ G -> (forall A, F A ⟛ G A).
Proof. firstorder. Qed.
Lemma pw_core' : forall `{Proset X} {Y} {F G : Y -> X}, F ⟛ G <-> (forall A, F A ⟛ G A).
Proof. firstorder. Qed.
Instance pw_core_subrel `{Proset X} {Y}
  : subrelation (core (pro_le (X:=Y -> X))) (pointwise_relation Y (core (pro_le (X:=X)))).
Proof. firstorder. Qed.
Local Arguments prod_relation {A B} R1 R2 !x !y /.
Instance prod_relation_pro `{PreOrder X R, PreOrder Y R'}
  : PreOrder (prod_relation R R').
Proof. firstorder. Qed.
Instance prod_proset `{Proset X, Proset Y} : Proset (X * Y) :=
  {| pro_le := prod_relation pro_le pro_le |}.

Class Monoidal {X} (R : X -> X -> Prop) (T : X -> X -> X) (U : X) :=
  {bimap : forall {A A' B B'}, R A A' -> R B B' -> R (T A B) (T A' B');
   lunit : forall A, core R (T U A) A;
   runit : forall A, core R (T A U) A;
   massoc : forall A B C, core R (T A (T B C)) (T (T A B) C)}.
Hint Mode Monoidal ! ! ! - : typeclass_instances.
Class Sym {X} (R : X -> X -> Prop) (T : X -> X -> X) :=
   sym : forall A B, R (T A B) (T B A).
Hint Mode Sym ! ! ! : typeclass_instances.
Class MonSet (X : Type) `{Proset X} :=
  {memp : X; pro_tens : X -> X -> X; pro_monoidal :> Monoidal pro_le pro_tens memp}.
Arguments pro_tens {_ _ _} !_ !_ /.
Hint Mode MonSet ! - : typeclass_instances.
Infix "⊗" := pro_tens (at level 40).
Definition pro_lunit `{MonSet X} (A : X) : memp ⊗ A ⟛ A := lunit A.
Definition pro_runit `{MonSet X} (A : X) : A ⊗ memp ⟛ A := runit A.
Definition pro_massoc `{MonSet X} (A B C : X) : A ⊗ (B ⊗ C) ⟛ (A ⊗ B) ⊗ C
  := massoc A B C.
Class SymMonSet (X : Type) `{MonSet X} :=
  {pro_sym :> Sym pro_le pro_tens}.
Hint Mode SymMonSet ! - - : typeclass_instances.
Definition pro_sym' `{SymMonSet X} (A B : X) : A ⊗ B ⟛ B ⊗ A :=
  conj (pro_sym A B) (pro_sym B A).
Definition tens_op {X} (T : X -> X -> X) (a1 a2 : op X) : op X :=
  Op (T (get_op a1) (get_op a2)).
Arguments tens_op {X} T !a1 !a2 /.
Instance pro_op_monoidal `{Monoidal X R T U} : Monoidal (pro_op R) (tens_op T) (Op U).
Proof.
  constructor.
  all: rewrite /pro_op /core /= => *; rewrite -/(core R _ _).
  - by apply: bimap.
  - symmetry; apply: lunit.
  - symmetry; apply: runit.
  - symmetry; apply: massoc.
Qed.
Instance pro_op_sym `{Sym X R T} : Sym (pro_op R) (tens_op T).
Proof. move=> [A] [B] /=; apply: sym. Qed.
Instance op_mset `{MonSet X} : MonSet (op X) := {| pro_tens := tens_op pro_tens |}.
Instance op_smset `{SymMonSet X} : SymMonSet (op X) := {}.
Instance pw_monoidal {Y} `{Monoidal X R T U} :
  Monoidal (pointwise_relation Y R) (fun f g y => T (f y) (g y)) (const U).
Proof.
  constructor.
  - move=> A A' B B' D1 D2 y; apply: bimap; [apply: D1 | apply: D2].
  - move=> A; apply/pw_core0 => y; apply: lunit.
  - move=> A; apply/pw_core0 => y; apply: runit.
  - move=> A B C; apply/pw_core0 => y; apply: massoc.
Qed.
Instance pw_sym {Y} `{Sym X R T}
  : Sym (pointwise_relation Y R) (fun f g y => T (f y) (g y)).
Proof. move=> A B y; apply: sym. Qed.
Instance pw_mset {Y} `{MonSet X} : MonSet (Y -> X) :=
  {| pro_tens := fun f g y => f y ⊗ g y |}.
Instance pw_smset {Y} `{SymMonSet X} : SymMonSet (Y -> X) := {}.
Definition tens_prod {X Y} (T : X -> X -> X) (T' : Y -> Y -> Y) (p1 p2 : X * Y) : X * Y :=
  (T p1.1 p2.1, T' p1.2 p2.2).
Arguments tens_prod {X Y} T T' !p1 !p2 /.
Instance prod_monoidal `{Monoidal X R T U, Monoidal Y R' T' U'}
  : Monoidal (prod_relation R R') (tens_prod T T') (U, U').
Proof. compute; firstorder. Qed.
Instance prod_sym `{Sym X R T, Sym Y R' T'} : Sym (prod_relation R R') (tens_prod T T').
Proof. by compute. Qed.
Instance prod_mset `{MonSet X, MonSet Y} : MonSet (X * Y)
  := {| pro_tens := tens_prod pro_tens pro_tens |}.
Instance prod_smset `{SymMonSet X, SymMonSet Y} : SymMonSet (X * Y) := {}.

Polymorphic Instance prop_proset : Proset Prop := {| pro_le := impl |}.
Proof. firstorder. Defined.

Class Functor `{Proset X, Proset Y} (F : X -> Y) :=
  fmap' : forall {A B}, A ⊢ B -> F A ⊢ F B.
Hint Mode Functor ! - ! - ! : typeclass_instances.
Class Full `{Proset X, Proset Y} (F : X -> Y) :=
  unfmap : forall {A B}, F A ⊢ F B -> A ⊢ B.
Hint Mode Full ! - ! - ! : typeclass_instances.
(* TODO Double-check this definition. *)
Class LaxMon `{MonSet X, MonSet Y} (F : X -> Y) :=
  {pres_memp : memp ⊢ F memp; pres_tens : forall {A B}, F A ⊗ F B ⊢ F (A ⊗ B)}.
Hint Mode LaxMon ! - - ! - - ! : typeclass_instances.
Class OplaxMon `{MonSet X, MonSet Y} (F : X -> Y) :=
  {pres_memp_op : F memp ⊢ memp; pres_tens_op : forall {A B}, F (A ⊗ B) ⊢ F A ⊗ F B}.
Hint Mode OplaxMon ! - - ! - - ! : typeclass_instances.
(* This definition is OK, since we are in prosets---I think??? *)
Class StrongMon `{MonSet X, MonSet Y} (F : X -> Y) :=
  {strong_lax :> LaxMon F; strong_oplax :> OplaxMon F}.
Hint Mode StrongMon ! - - ! - - ! : typeclass_instances.

Lemma unfmap_fmap : forall `{Proset X, Proset Y} {F : X -> Y} `{!Functor F, !Full F} {A B},
    F A ⊢ F B <-> A ⊢ B.
Proof. split; [apply: unfmap | apply: fmap']. Qed.
Lemma fmap_core : forall `{Functor X Y F} {A B}, A ⟛ B -> F A ⟛ F B.
Proof. firstorder. Qed.
Instance functor_proper `{Functor X Y F} : Proper (pro_le ++> pro_le) F.
Proof. move=> ? ?; apply: fmap'. Qed.
(* flipped instances due to https://gitlab.mpi-sws.org/iris/stdpp/issues/58 *)
Instance functor_proper' `{Functor X Y F} : Proper (pro_le --> flip pro_le) F.
Proof. move=> ? ?; apply: fmap'. Qed.
Instance functor_proper_core `{Functor X Y F} : Proper (core pro_le ==> core pro_le) F.
Proof. move=> ? ? [? ?]; split; by apply fmap'. Qed.
Instance functor_proper_core' `{Functor X Y F} : Proper (core pro_le --> core pro_le) F.
Proof. move=> ? ? [? ?]; split; by apply fmap'. Qed.
Instance monoidal_proper `{Monoidal X R T} : Proper (R ++> R ++> R) T.
Proof. move=> ? ? ? ? ? ?; by apply: bimap. Qed.
Instance monoidal_proper' `{Monoidal X R T} : Proper (R --> R --> flip R) T.
Proof. move=> ? ? ? ? ? ? /=; by apply: bimap. Qed.
(* TODO This should probably be falling out of functor_proper...
Instance compose_proper `{Proset X, Proset Y, Proset Z}
  : Proper (pro_le ++> pro_le ++> pro_le) (@compose X Y Z).
Proof. move=> F F' E1 G G' E2 A /=. setoid_rewrite E1. setoid_rewrite E2.  ? ? ? ? ? ? ? /=. firstorder. Qed.
*)

Definition cone `{Proset X} {R} (A : X) (J : R -> X) : Prop :=
  forall x, A ⊢ J x.
Definition lim_cone `{Proset X} {R} (A : X) (J : R -> X) : Prop :=
  cone A J /\ forall A', cone A' J -> A' ⊢ A.
Definition cocone `{Proset X} {R} (J : R -> X) (A : X) : Prop :=
  forall x, J x ⊢ A.
Definition colim_cocone `{Proset X} {R} (J : R -> X) (A : X) : Prop :=
  cocone J A /\ forall A', cocone J A' -> A ⊢ A'.
Lemma lim_unique : forall `{Proset X} {R} {A A'} {J : R -> X},
    lim_cone A J -> lim_cone A' J <-> A ⟛ A'.
Proof.
  move=> X Pr_X R A A' J LC; split; first by firstorder.
  move: LC => [C U] [H1 H2]; split=> [r | A'' C'].
  - by etransitivity.
  - setoid_rewrite <- H1; firstorder.
Qed.
Lemma colim_unique : forall `{Proset X} {R} {J : R -> X} {A A'},
    colim_cocone J A -> colim_cocone J A' <-> A ⟛ A'.
Proof.
  move=> X Pr_X R J A A' LC; split; first by firstorder.
  move: LC => [C U] [H1 H2]; split=> [r | A'' C'].
  - setoid_rewrite <- H1; firstorder.
  - setoid_rewrite H2; firstorder.
Qed.

Class Bicomplete (X : Type) `{Proset X} :=
  {lim R (J : R -> X) : X;
   is_lim : forall {R} J, lim_cone (lim R J) J;
   colim R (J : R -> X) : X;
   is_colim : forall {R} J, colim_cocone J (colim R J)}.
Hint Mode Bicomplete ! - : typeclass_instances.
Arguments lim {_ _ _ _} _.
Arguments colim {_ _ _ _} _.

(* TODO Lots of stuff! *)
Definition lim_left `{Bicomplete X} {R} (J : R -> X) : cone (lim J) J
  := proj1 (is_lim J).
Definition lim_right `{Bicomplete X} {R} (A : X) (J : R -> X) (C : cone A J)
  : A ⊢ lim J
  := proj2 (is_lim J) A C.
Lemma lim_universal : forall `{Bicomplete X} {R} (A : X) (J : R -> X),
    cone A J <-> A ⊢ lim J.
Proof.
  move=> X Pr Comp R A J; split; first by apply: lim_right.
  move=> D r; etransitivity; [eassumption | apply: lim_left].
Qed.
Definition colim_right `{Bicomplete X} {R} (J : R -> X) : cocone J (colim J)
  := proj1 (is_colim J).
Definition colim_left `{Bicomplete X} {R} (A : X) (J : R -> X) (C : cocone J A)
  : colim J ⊢ A
  := proj2 (is_colim J) A C.
Lemma colim_universal : forall `{Bicomplete X} {R} (J : R -> X) (A : X),
    cocone J A <-> colim J ⊢ A.
Proof.
  move=> X Pr Cocomp R J A; split; first by apply: colim_left.
  move=> D r; etransitivity; [apply: colim_right | eassumption].
Qed.

(* This could probably follow by some abstract nonsense if we moved it *after*
   presheaf stuff... Hmm. Oh well. *)
Definition colim_from_lim `{Proset X} (lim' : forall {R} (J : R -> X), X) {R} (J : R -> X) : X
  := lim' (fun A : sig (cocone J) => `A).
Lemma complete_sufficient : forall `{Proset X} (lim' : forall {R}, (R -> X) -> X),
    (forall {R} (J : R -> X), lim_cone (lim' J) J) -> Bicomplete X.
Proof.
  move=> X Pr lim' is_lim'.
  refine {| lim := lim';
            is_lim := is_lim';
            colim R J := lim' _ (fun A : sig (fun A0 => cocone J A0) => `A); |}.
  move=> R J; split=> [r | A Cocone].
  - apply/(proj2 (is_lim' _ _)) => -[A Cocone] //=.
  - change A with (sval ((A ↾ Cocone) : sig (fun A0 => cocone J A0))).
    apply: (proj1 (is_lim' _ _)).
Qed.
Lemma cocomplete_sufficient : forall `{Proset X} (colim' : forall {R}, (R -> X) -> X),
    (forall {R} (J : R -> X), colim_cocone J (colim' J)) -> Bicomplete X.
Proof.
  move=> X Pr colim' is_colim'.
  refine {| lim R J := colim' _ (fun A : sig (fun A0 => cone A0 J) => `A);
            colim := colim';
            is_colim := is_colim' |}.
  move=> R J; split=> [r | A Cone].
  - apply/(proj2 (is_colim' _ _)) => -[A Cone] //=.
  - change A with (sval ((A ↾ Cone) : sig (fun A0 => cone A0 J))).
    apply: (proj1 (is_colim' _ _)).
Qed.

Lemma lim_unique' : forall `{Bicomplete X} {R} {A} {J : R -> X},
    lim_cone A J <-> A ⟛ lim J.
Proof.
  move=> ? ? ? ? ? J; generalize (is_lim J).
  rewrite (symmetry_iff (core _)); apply: lim_unique.
Qed.
Lemma colim_unique' : forall `{Bicomplete X} {R} {J : R -> X} {A},
    colim_cocone J A <-> A ⟛ colim J.
Proof.
  move=> ? ? ? ? J ?; generalize (is_colim J).
  rewrite (symmetry_iff (core _)); apply: colim_unique.
Qed.

Polymorphic Instance prop_bicomplete : Bicomplete Prop := {| lim := all; colim := ex |}.
Proof.
  - firstorder.
  - move=> R J; firstorder.
Defined.
Program Instance pw_bicomplete `{Bicomplete X} {Y} : Bicomplete (Y -> X)
  := {| lim R J y := lim (fun r => J r y); colim R J y := colim (fun r => J r y) |}.
Next Obligation.
  move=> X Pr Comp Y R J; split=> [r | A' Cone] y.
  - apply: lim_left.
  - apply: lim_right; firstorder.
Qed.
Next Obligation.
  move=> X Pr Comp Y R J; split=> [r | A' Cone] y.
  - apply: colim_right.
  - apply: colim_left; firstorder.
Qed.
Program Instance op_bicomplete `{Bicomplete X} : Bicomplete (op X)
  := {| lim R J := Op (colim (get_op ∘ J)); colim R J := Op (lim (get_op ∘ J)) |}.
Next Obligation.
  move=> X Pr Comp R J; split=> [r | [A'] Cone] /=.
  - replace (J r) with (Op (get_op (J r))) by by case: J.
    apply: (colim_right (get_op ∘ J)).
  - apply: colim_left; firstorder.
Qed.
Next Obligation.
  move=> X Pr Comp R J; split=> [r | [A'] Cone] /=.
  - replace (J r) with (Op (get_op (J r))) by by case: J.
    apply: (lim_left (get_op ∘ J)).
  - apply: lim_right; firstorder.
Qed.
Lemma curry_dep : forall {A B} {P : A * B -> Prop}, (forall p, P p) <-> (forall a b, P (a, b)).
Proof. move=> A B P; split=> [H a b | H [a b]] //. Qed.
Lemma uncurry_dep : forall {A B} {P : A -> B -> Prop}, (forall a b, P a b) <-> (forall p, P p.1 p.2).
Proof. move=> A B P; split=> [H p | H a b] //; apply: (H (a, b)). Qed.
Lemma prod_cone : forall `{Proset X, Proset Y} {R} {p} {J : R -> X * Y},
    cone p J <-> cone p.1 (fst ∘ J) /\ cone p.2 (snd ∘ J).
Proof. firstorder. Qed.
Lemma prod_cocone : forall `{Proset X, Proset Y} {R} {J : R -> X * Y} {p},
    cocone J p <-> cocone (fst ∘ J) p.1 /\ cocone (snd ∘ J) p.2.
Proof. firstorder. Qed.
Lemma prod_lim_cone : forall `{Proset X, Proset Y} {R} {p} {J : R -> X * Y},
    lim_cone p J <-> lim_cone p.1 (fst ∘ J) /\ lim_cone p.2 (snd ∘ J).
Proof.
  move=> X Pr_X Y Pr_Y R [A B] J /=.
  rewrite /lim_cone curry_dep /=; setoid_rewrite (prod_cone (J:=J)); firstorder.
Qed.
Lemma prod_colim_cocone : forall `{Proset X, Proset Y} {R} {J : R -> X * Y} {p},
    colim_cocone J p <-> colim_cocone (fst ∘ J) p.1 /\ colim_cocone (snd ∘ J) p.2.
Proof.
  move=> X Pr_X Y Pr_Y R J [A B] /=.
  rewrite /colim_cocone curry_dep /=; setoid_rewrite (prod_cocone (J:=J)); firstorder.
Qed.
Program Instance prod_bicomplete `{Bicomplete X, Bicomplete Y} : Bicomplete (X * Y)
  := {| lim R J := (lim (fst ∘ J), lim (snd ∘ J));
        colim R J := (colim (fst ∘ J), colim (snd ∘ J)) |}.
Next Obligation.
  move=> X Pr_X Comp_X Y Pr_Y Comp_Y R J /=.
  rewrite prod_lim_cone /=; split; apply: is_lim.
Qed.
Next Obligation.
  move=> X Pr_X Comp_X Y Pr_Y Comp_Y R J /=.
  rewrite prod_colim_cocone /=; split; apply: is_colim.
Qed.
(*
Lemma op_cone_cocone : forall `{Proset X} {R},
*)

Definition terminal (X : Type) `{Bicomplete X} : X :=
  lim (fun H : Empty_set => match H with end).
Definition initial (X : Type) `{Bicomplete X} : X :=
  colim (fun H : Empty_set => match H with end).
Notation "⊤" := (terminal _).
Notation "⊥" := (initial _).
Definition product `{Bicomplete X} (A B : X) : X :=
  lim (fun b : bool => if b then A else B).
Definition coproduct `{Bicomplete X} (A B : X) : X :=
  colim (fun b : bool => if b then A else B).
Infix "×" := product (at level 40).
Definition embed_prop `{Bicomplete X} (P : Prop) : X := colim (fun H : P => ⊤).
Arguments terminal : simpl never.
Arguments initial : simpl never.
Arguments product : simpl never.
Arguments coproduct : simpl never.
Arguments embed_prop : simpl never.

Definition terminal_right `{Bicomplete X} {A : X} : A ⊢ ⊤ :=
  lim_right _ _ (fun e : Empty_set => match e with end).
Definition initial_left `{Bicomplete X} {A : X} : ⊥ ⊢ A :=
  colim_left _ _ (fun e : Empty_set => match e with end).
Definition product_left1 `{Bicomplete X} {A B : X} : A × B ⊢ A
  := lim_left _ true.
Definition product_left2 `{Bicomplete X} {A B : X} : A × B ⊢ B
  := lim_left _ false.
Definition product_right `{Bicomplete X} {C A B : X} (D1 : C ⊢ A) (D2 : C ⊢ B)
  : C ⊢ A × B
  := lim_right _ _ (fun b => if b as b0 return C ⊢ (if b0 then A else B) then D1 else D2).
Definition coproduct_left `{Bicomplete X} {A B C : X} (D1 : A ⊢ C) (D2 : B ⊢ C)
  : coproduct A B ⊢ C
  := colim_left _ _ (fun b => if b as b0 return (if b0 then A else B) ⊢ C then D1 else D2).
Definition coproduct_right1 `{Bicomplete X} {A B : X} : A ⊢ coproduct A B
  := colim_right _ true.
Definition coproduct_right2 `{Bicomplete X} {A B : X} : B ⊢ coproduct A B
  := colim_right _ false.
Lemma embed_prop_left : forall `{Bicomplete X} {P : Prop} {Q : X},
    (P -> ⊤ ⊢ Q) -> embed_prop P ⊢ Q.
Proof. move=> X Pr Comp P Q D; by apply: colim_left. Qed.
(* TODO Put this in the right place *)
Lemma embed_prop_right : forall `{Bicomplete X} {P : X} {Q : Prop},
    Q -> P ⊢ embed_prop Q.
Proof.
  move=> X Pr Comp P Q H.
  setoid_rewrite <- colim_right.
  - apply: terminal_right.
  - done.
Qed.

Class Continuous `{Proset X, Proset Y} (F : X -> Y) :=
  distrib_lim_cone : forall {R} A (J : R -> X), lim_cone A J -> lim_cone (F A) (F ∘ J).
Class Cocontinuous `{Proset X, Proset Y} (F : X -> Y) :=
  distrib_colim_cocone :
    forall {R} (J : R -> X) A, colim_cocone J A -> colim_cocone (F ∘ J) (F A).
Hint Mode Continuous ! - ! - ! : typeclass_instances.
Hint Mode Cocontinuous ! - ! - ! : typeclass_instances.
Lemma create_lim_cone : forall `{Bicomplete X, Proset Y} {F : X -> Y} `{!Continuous F}
                          {R} (J : R -> X),
    lim_cone (F (lim J)) (F ∘ J).
Proof. move=> *; apply: distrib_lim_cone; apply: is_lim. Qed.
Lemma distrib_lim : forall `{Bicomplete X, Bicomplete Y} {F : X -> Y} `{!Continuous F}
                      {R} (J : R -> X),
    F (lim J) ⟛ lim (F ∘ J).
Proof. move=> *; apply/lim_unique'/create_lim_cone. Qed.
Lemma distrib_lim_sufficient : forall `{Bicomplete X, Bicomplete Y} {F : X -> Y} `{!Functor F},
    (forall {R} (J : R -> X), lim (F ∘ J) ⊢ F (lim J)) -> Continuous F.
Proof.
  move=> X Pr_X Comp_X Y Pr_Y Comp_Y F Funct_F Distr R A J /lim_unique' E.
  apply/lim_unique'; rewrite E; split.
  - apply/lim_right => r /=; apply/fmap'/lim_left.
  - apply/Distr.
Qed.
Lemma create_colim_cocone : forall `{Bicomplete X, Proset Y} {F : X -> Y} `{!Cocontinuous F}
                              {R} (J : R -> X),
    colim_cocone (F ∘ J) (F (colim J)).
Proof. move=> *; apply: distrib_colim_cocone; apply: is_colim. Qed.
Lemma distrib_colim : forall `{Bicomplete X, Bicomplete Y} {F : X -> Y} `{!Cocontinuous F}
                        {R} (J : R -> X),
    F (colim J) ⟛ colim (F ∘ J).
Proof. move=> *; apply/colim_unique'/create_colim_cocone. Qed.
Lemma distrib_colim_sufficient : forall `{Bicomplete X, Bicomplete Y}
                                   {F : X -> Y} `{!Functor F},
    (forall {R} (J : R -> X), F (colim J) ⊢ colim (F ∘ J)) -> Cocontinuous F.
Proof.
  move=> X Pr_X Comp_X Y Pr_Y Comp_Y F Funct_F Distr R A J /colim_unique' E.
  apply/colim_unique'; rewrite E; split.
  - apply/Distr.
  - apply/colim_left => r /=; apply/fmap'/colim_right.
Qed.

(*
Instance op_continuous1 `{Functor X Y F, !Continuous F}
  : Cocontinuous (Op ∘ F ∘ get_op).
Proof.
  move=> R J A [C U].
  split=> [r | [A']] /=.
  - move: (C r); move: (A) (J r) => [?] [?] /=; apply/fmap'.
  - apply: C'. (U (Op (F (get_op A)))). A')).
*)

Instance id_functor `{Proset X} : Functor (@id X).
Proof. firstorder. Qed.
Instance id_strongmon `{MonSet X} : StrongMon (@id X).
Proof. firstorder. Qed.
Instance compose_functor `{Proset X, Proset Y, Proset Z'} {F : Y -> Z'} {G : X -> Y}
         `{!Functor F, !Functor G} : Functor (F ∘ G) | 0.
Proof. firstorder. Qed.
Instance compose_laxmon `{MonSet X, MonSet Y, MonSet Z'}
         {F : Y -> Z'} {G : X -> Y} `{!Functor F, !Functor G, !LaxMon F, !LaxMon G}
  : LaxMon (F ∘ G).
Proof.
  constructor=> [| A B] /=.
  - setoid_rewrite <- pres_memp; apply: pres_memp.
  - setoid_rewrite <- pres_tens; apply: pres_tens.
Qed.
Instance compose_oplaxmon `{MonSet X, MonSet Y, MonSet Z'}
         {F : Y -> Z'} {G : X -> Y} `{!Functor F, !Functor G, !OplaxMon F, !OplaxMon G}
  : OplaxMon (F ∘ G).
Proof.
  constructor=> [| A B] /=.
  - setoid_rewrite pres_memp_op; apply: pres_memp_op.
  - setoid_rewrite pres_tens_op; apply: pres_tens_op.
Qed.
Instance compose_strongmon `{MonSet X, MonSet Y, MonSet Z'}
         {F : Y -> Z'} {G : X -> Y} `{!Functor F, !Functor G, !StrongMon F, !StrongMon G}
  : StrongMon (F ∘ G).
Proof. constructor; typeclasses eauto. Qed.
Instance const_functor {Y} `{Proset X} : Functor (@const X Y).
Proof. firstorder. Qed.
Instance const_strongmon {Y} `{MonSet X} : StrongMon (@const X Y).
Proof. constructor; by constructor. Qed.
Instance subst_functor1 {Y Y'} `{Proset X} {f : Y' -> Y} : Functor (X:=Y -> X) (.∘ f).
Proof. firstorder. Qed.
Instance subst_strongmon1 {Y Y'} `{MonSet X} {f : Y' -> Y} :
  StrongMon (X:=Y -> X) (.∘ f).
Proof. constructor; by constructor. Qed.
Instance postcomp_functor {Y} `{Proset X, Proset X'} {f : X -> X'} `{!Functor f} :
  Functor (X:=Y -> X) (f ∘.).
Proof. firstorder. Qed.
Instance postcomp_laxmon {Y} `{MonSet X, MonSet X'}
         {f : X -> X'} `{!Functor f, !LaxMon f} : LaxMon (X:=Y -> X) (f ∘.).
Proof. firstorder. Qed.
Instance postcomp_oplaxmon {Y} `{MonSet X, MonSet X'}
         {f : X -> X'} `{!Functor f, !OplaxMon f} : OplaxMon (X:=Y -> X) (f ∘.).
Proof. firstorder. Qed.
Instance postcomp_strongmon {Y} `{MonSet X, MonSet X'}
         {f : X -> X'} `{!Functor f, !StrongMon f} : StrongMon (X:=Y -> X) (f ∘.).
Proof. firstorder. Qed.
Definition eval_at {X Y} (y : Y) : (Y -> X) -> X := fun f => f y.
Instance subst_functor2 {Y} `{Proset X} {y : Y} : Functor (X:=Y -> X) (eval_at y).
Proof. firstorder. Qed.
Instance subst_strongmon2 {Y} `{MonSet X} {y : Y} : StrongMon (X:=Y -> X) (eval_at y).
Proof. constructor; by constructor. Qed.
Instance fst_functor `{Proset X, Proset Y} : Functor (@fst X Y).
Proof. firstorder. Qed.
Instance fst_strongmon `{MonSet X, MonSet Y} : StrongMon (@fst X Y).
Proof. by compute. Qed.
Instance fst_continuous `{Proset X, Proset Y} : Continuous (@fst X Y).
Proof. move=> R [A B] J /=; rewrite prod_lim_cone /=; tauto. Qed.
Instance fst_cocontinuous `{Proset X, Proset Y} : Cocontinuous (@fst X Y).
Proof. move=> R J [A B] /=; rewrite prod_colim_cocone /=; tauto. Qed.
Instance snd_functor `{Proset X, Proset Y} : Functor (@snd X Y).
Proof. firstorder. Qed.
Instance snd_strongmon `{MonSet X, MonSet Y} : StrongMon (@snd X Y).
Proof. by compute. Qed.
Instance snd_continuous `{Proset X, Proset Y} : Continuous (@snd X Y).
Proof. move=> R [A B] J /=; rewrite prod_lim_cone /=; tauto. Qed.
Instance snd_cocontinuous `{Proset X, Proset Y} : Cocontinuous (@snd X Y).
Proof. move=> R J [A B] /=; rewrite prod_colim_cocone /=; tauto. Qed.
Instance pair_functor `{Proset X, Proset Y} : Functor (@pair X Y).
Proof. by compute. Qed.
Instance pair_functor' `{Proset X, Proset Y} {A} : Functor (@pair X Y A).
Proof. by compute. Qed.
Definition r_pair {X Y} (B : Y) : X -> X * Y := fun A => (A, B).
Arguments r_pair {_ _} _ _ /.
Instance r_pair_functor' `{Proset X, Proset Y} {B} : Functor (@r_pair X Y B).
Proof. by compute. Qed.
Instance memp_pair_strongmon `{MonSet X, MonSet Y} : StrongMon (@pair X Y memp).
Proof.
  constructor; constructor=> [| B B']; rewrite /= /prod_relation //=;
    (split; [apply (pro_lunit memp) | done]).
Qed.
Instance pair_memp_strongmon `{MonSet X, MonSet Y} : StrongMon (@r_pair X Y memp).
Proof.
  constructor; constructor=> [| A A']; rewrite /= /prod_relation //=;
    (split; [done | apply (pro_lunit memp)]).
Qed.
Instance prod_map_functor `{Proset X, Proset X', Proset Y, Proset Y'}
         {F : X -> X'} {G : Y -> Y'} `{!Functor F, !Functor G} : Functor (prod_map F G).
Proof. firstorder. Qed.
Instance prod_map_laxmon `{MonSet X, MonSet X', MonSet Y, MonSet Y'}
         {F : X -> X'} {G : Y -> Y'} `{!Functor F, !Functor G, !LaxMon F, !LaxMon G}
  : LaxMon (prod_map F G).
Proof. firstorder. Qed.
Instance prod_map_oplaxmon `{MonSet X, MonSet X', MonSet Y, MonSet Y'}
         {F : X -> X'} {G : Y -> Y'} `{!Functor F, !Functor G, !OplaxMon F, !OplaxMon G}
  : OplaxMon (prod_map F G).
Proof. firstorder. Qed.
Instance prod_map_strongmon `{MonSet X, MonSet X', MonSet Y, MonSet Y'}
         {F : X -> X'} {G : Y -> Y'} `{!Functor F, !Functor G, !StrongMon F, !StrongMon G}
  : StrongMon (prod_map F G).
Proof. firstorder. Qed.
Instance prod_map_continuous `{Proset X, Proset X', Proset Y, Proset Y'} {F : X -> X'}
         {G : Y -> Y'} `{!Functor F, !Functor G, !Continuous F, !Continuous G}
  : Continuous (prod_map F G).
Proof.
  move=> R [A B] J /= /prod_lim_cone /= [L1 L2].
  apply/prod_lim_cone; split; simpl; by apply/distrib_lim_cone.
Qed.
Instance prod_map_cocontinuous `{Proset X, Proset X', Proset Y, Proset Y'} {F : X -> X'}
         {G : Y -> Y'} `{!Functor F, !Functor G, !Cocontinuous F, !Cocontinuous G}
  : Cocontinuous (prod_map F G).
Proof.
  move=> R J [A B] /= /prod_colim_cocone /= [L1 L2].
  apply/prod_colim_cocone; split; simpl.
  - by apply/(distrib_colim_cocone (F:=F)).
  - by apply/(distrib_colim_cocone (F:=G)).
Qed.
Definition r_prod_map {X X' Y Y'} (G : Y -> Y') : (X -> X') -> X * Y -> X' * Y'
  := fun F => prod_map F G.
Arguments r_prod_map {_ _ _ _} _ _ /.
Instance prod_map_functor' `{Proset X, Proset X', Proset Y, Proset Y'}
         {F : X -> X'} `{!Functor F} : Functor (prod_map (B:=Y) (B':=Y') F).
Proof. move=> G G' D [A B] /=; firstorder. Qed.
Instance r_prod_map_functor `{Proset X, Proset X', Proset Y, Proset Y'}
         {G : Y -> Y'} `{!Functor G} : Functor (r_prod_map (X:=X) (X':=X') G).
Proof. move=> F F' D [A B] /=; firstorder. Qed.
Instance prod_map_functor'' `{Proset X, Proset X', Proset Y, Proset Y'}
  : Functor (prod_map (A:=X) (A':=X') (B:=Y) (B':=Y')).
Proof. move=> F F' D G [A B] /=; firstorder. Qed.
Instance curry_functor `{Proset X, Proset Y, Proset Z} {F : X -> Y -> Z}
         `{!Functor F, !forall x, Functor (F x)}
  : Functor (curry F).
Proof. move=> [A B] [A' B'] /= [D1 D2]; setoid_rewrite D2; apply: (fmap' (F:=F) D1). Qed.
Instance uncurry_functor `{Proset X, Proset Y, Proset Z} {F : X * Y -> Z} `{!Functor F}
  : Functor (uncurry F).
Proof.
  move=> A A' D B; change (F (r_pair B A) ⊢ F (r_pair B A')); by setoid_rewrite D.
Qed.
Instance uncurry_functor' `{Proset X, Proset Y, Proset Z}
         {F : X * Y -> Z} `{!Functor F} {A}
  : Functor (uncurry F A).
Proof. move=> B B' D; unfold uncurry; by setoid_rewrite D. Qed.
Instance lim_functor {Y} `{Bicomplete X} : Functor (lim (X:=X) (R:=Y)).
Proof. move=> A B H1; apply/lim_right => y; by setoid_rewrite lim_left. Qed.
Instance lim_laxmon {Y} `{Bicomplete X, !MonSet X} : LaxMon (lim (X:=X) (R:=Y)).
Proof.
  constructor=> [| A B].
  - apply: lim_right => A //.
  - apply: lim_right => y; apply: bimap; by apply/lim_left.
Qed.
Instance colim_functor {Y} `{Bicomplete X} : Functor (colim (X:=X) (R:=Y)).
Proof. move=> A B H1; apply/colim_left => y; by setoid_rewrite <- colim_right. Qed.
Instance colim_oplaxmon {Y} `{Bicomplete X, !MonSet X} : OplaxMon (colim (X:=X) (R:=Y)).
Proof.
  constructor=> [| A B].
  - apply: colim_left => A //.
  - apply: colim_left => y; apply: bimap; by apply/colim_right.
Qed.

Class ClosedMonoidal {X} (R : X -> X -> Prop) (T : X -> X -> X) (H : X -> X -> X) :=
  tensor_hom A B C : R (T A B) C <-> R A (H B C).
Hint Mode ClosedMonoidal ! ! ! - : typeclass_instances.
Hint Mode ClosedMonoidal ! ! - ! : typeclass_instances.
Class LClosedMonoidal {X} (R : X -> X -> Prop) (T : X -> X -> X) (H : X -> X -> X) :=
  l_tensor_hom A B C : R (T B A) C <-> R A (H B C).
Hint Mode LClosedMonoidal ! ! ! - : typeclass_instances.
Hint Mode LClosedMonoidal ! ! - ! : typeclass_instances.
Class ClosedMonSet (X : Type) `{MonSet X} :=
  {internal_hom : X -> X -> X;
   pro_tensor_hom :> ClosedMonoidal pro_le pro_tens internal_hom}.
Hint Mode ClosedMonSet ! - - : typeclass_instances.
Infix "⊸" := internal_hom (at level 40).
Class LClosedMonSet (X : Type) `{MonSet X} :=
  {l_internal_hom : X -> X -> X;
   pro_l_tensor_hom :> LClosedMonoidal pro_le pro_tens l_internal_hom}.
Hint Mode LClosedMonSet ! - - : typeclass_instances.
Infix "⊸̂" := l_internal_hom (at level 40).
Class Frame (X : Type) `{Bicomplete X} :=
  {exponential : X -> X -> X;
   product_exponential A B C : A × B ⊢ C <-> A ⊢ exponential B C}.
Hint Mode Frame ! - - : typeclass_instances.
Instance sym_lclosed `{SymMonSet X, !ClosedMonSet X} : LClosedMonSet X | 2
  := {| l_internal_hom := internal_hom |}.
Proof.
  move=> A B C; split.
  - setoid_rewrite <- (proj2 (pro_sym' _ _)); apply pro_tensor_hom.
  - setoid_rewrite pro_sym'; apply pro_tensor_hom.
Defined.
Polymorphic Instance prop_frame : Frame Prop
  := {| exponential (P Q : Prop) := P -> Q |}.
Proof.
  move=> A B C; split.
  - move=> D H1 H2; apply: D => -[] //.
  - move=> D H1; apply: D; [apply: (H1 true) | apply: (H1 false)].
Qed.

Definition sig_rel {X} (R : X -> X -> Prop) (P : X -> Prop) : sig P -> sig P -> Prop :=
  fun s1 s2 => R (`s1) (`s2).
Arguments sig_rel {_} _ _ !_ !_ /.
Instance sig_reflexive `{Reflexive X R} {P} : Reflexive (sig_rel R P).
Proof. by compute. Qed.
Instance sig_transitive `{Transitive X R} {P} : Transitive (sig_rel R P).
Proof. move=> a b c; unfold sig_rel; by etransitivity. Qed.
Instance sig_symmetric `{Symmetric X R} {P} : Symmetric (sig_rel R P).
Proof. by compute. Qed.
Instance sig_pro `{PreOrder X R} {P} : PreOrder (sig_rel R P)
  := {}.
Instance sig_proset `{Proset X} {P : X -> Prop} : Proset (sig P) :=
  {| pro_le := sig_rel pro_le P |}.
Class MonClosed `{MonSet X} (P : X -> Prop) :=
  {memp_closed : P memp;
   tensor_closed : forall {A B : X}, P A -> P B -> P (A ⊗ B)}.
Hint Mode MonClosed ! - - ! : typeclass_instances.
Definition tens_sig `{MonSet X} (P : X -> Prop) `{!MonClosed P}
           (s1 s2 : sig P) : sig P
  := match s1, s2 with a1 ↾ H1, a2 ↾ H2 => (a1 ⊗ a2) ↾ tensor_closed H1 H2 end.
Arguments tens_sig {_ _ _} P {_} !_ !_ /.
Instance sig_monoidal `{MonSet X} {P : X -> Prop} `{!MonClosed P} :
  Monoidal (sig_rel pro_le P) (tens_sig P) (memp ↾ memp_closed).
Proof.
  constructor; rewrite /core.
  - move=> [A H_A] [A' H_A'] [B H_B] [B' H_B'] /=; apply: bimap.
  - move=> [A H_A] /=; apply: lunit.
  - move=> [A H_A] /=; apply: runit.
  - move=> [A H_A] [B H_B] [C H_C] /=; apply: massoc.
Qed.
Instance sig_sym `{SymMonSet X} {P : X -> Prop} `{!MonClosed P} :
  Sym (sig_rel pro_le P) (tens_sig P).
Proof. move=> [A H_A] [B H_B] /=; apply: sym. Qed.
Instance sig_mset `{MonSet X} {P : X -> Prop} `{!MonClosed P} :
  MonSet (sig P) := {| pro_tens := tens_sig P |}.
Instance sig_smset  `{SymMonSet X} {P : X -> Prop} `{!MonClosed P} :
  SymMonSet (sig P) := {}.
Instance sval_functor `{Proset X} {P : X -> Prop} : Functor (proj1_sig (P:=P)).
Proof. firstorder. Qed.
Instance sval_full `{Proset X} {P : X -> Prop} : Full (proj1_sig (P:=P)).
Proof. firstorder. Qed.
Instance sval_strongmon `{MonSet X} {P : X -> Prop} `{!MonClosed P}
  : StrongMon (proj1_sig (P:=P)).
Proof. constructor; constructor=> // -[A H_A] [B H_B] //. Qed.
Definition pack {A} (P : A -> Prop) (x : A) `{H : P x} : sig P
  := x ↾ H.
Definition pack_ph {A P} (ph : phant (sig P)) (x : A) `{H : P x} : sig P
  := x ↾ H.
Definition restrict_cod {A B} (P : B -> Prop) (f : A -> B) `{H : forall a, P (f a)} : A -> sig P
  := fun a => f a ↾ H a.
Definition restrict_cod_ph {A B} {P : B -> Prop} (ph : phant (sig P)) (f : A -> B)
           `{H : forall a, P (f a)} : A -> sig P
  := fun a => f a ↾ H a.
Notation "x 'packed_with' P" := (pack P x) (at level 20).
Notation "x 'as_a' T" := (pack_ph (Phant T) x : T) (at level 20).
Notation "f ▷ P" := (restrict_cod P f) (at level 20).
Notation "f ↑" := (restrict_cod _ f) (at level 20).
Notation "f ↣ T" := (restrict_cod_ph (Phant T) f : _ -> T) (at level 20).
Arguments pack {_} _ _ {_} /.
Arguments pack_ph {_ _} _ _ {_} /.
Arguments restrict_cod {_ _} _ _ {_} _ /.
Arguments restrict_cod_ph {_ _ _} _ _ {_} _ /.
Instance restrict_cod_functor `{Functor X Y F} {P : Y -> Prop} {H : forall A, P (F A)}
  : Functor (restrict_cod P F (H:=H)).
Proof. firstorder. Qed.
Instance restrict_cod_full `{Full X Y F} {P : Y -> Prop} {H : forall A, P (F A)}
  : Full (restrict_cod P F (H:=H)).
Proof. firstorder. Qed.
Instance restrict_cod_laxmon `{LaxMon X Y F} {P : Y -> Prop}
         `{!MonClosed P} {H : forall A, P (F A)}
  : LaxMon (restrict_cod P F (H:=H)).
Proof. firstorder. Qed.
Instance restrict_cod_oplaxmon `{OplaxMon X Y F} {P : Y -> Prop}
         `{!MonClosed P} {H : forall A, P (F A)}
  : OplaxMon (restrict_cod P F (H:=H)).
Proof. firstorder. Qed.
Instance restrict_cod_strongmon `{StrongMon X Y F} {P : Y -> Prop}
         `{!MonClosed P} {H : forall A, P (F A)}
  : StrongMon (restrict_cod P F (H:=H)).
Proof. firstorder. Qed.
Instance restrict_cod_ph_functor `{Functor X Y F} {P : Y -> Prop} {H : forall A, P (F A)}
         {ph : phant (sig P)}
  : Functor (restrict_cod_ph ph F (H:=H)).
Proof. firstorder. Qed.
Instance restrict_cod_ph_full `{Full X Y F} {P : Y -> Prop} {H : forall A, P (F A)}
         {ph : phant (sig P)}
  : Full (restrict_cod_ph ph F (H:=H)).
Proof. firstorder. Qed.
Instance restrict_cod_ph_laxmon `{LaxMon X Y F} {P : Y -> Prop} {ph : phant (sig P)}
         `{!MonClosed P} {H : forall A, P (F A)}
  : LaxMon (restrict_cod_ph ph F (H:=H)).
Proof. firstorder. Qed.
Instance restrict_cod_ph_oplaxmon `{OplaxMon X Y F} {P : Y -> Prop} {ph : phant (sig P)}
         `{!MonClosed P} {H : forall A, P (F A)}
  : OplaxMon (restrict_cod_ph ph F (H:=H)).
Proof. firstorder. Qed.
Instance restrict_cod_ph_strongmon `{StrongMon X Y F} {P : Y -> Prop} {ph : phant (sig P)}
         `{!MonClosed P} {H : forall A, P (F A)}
  : StrongMon (restrict_cod_ph ph F (H:=H)).
Proof. firstorder. Qed.


(* !!!
   IMPORTANT NOTE: Constraints for adjunctions should generally be placed BEFORE
   Functor, LaxMon, etc constraints (as far as I can tell, so far) because resolving them
   can instantiate evars that would otherwise cause resolution of those other classes
   to fail! Hence, this ordering allows more powerful inference.
   !!! *)
Class Adjoint `{Proset X, Proset Y} (F : X -> Y) (G : Y -> X)
  : Prop := {adj_unit : id ⊢ G ∘ F; adj_counit : F ∘ G ⊢ id}.
(* TODO This is probably worth putting some more careful thought into! *)
Hint Mode Adjoint ! - ! - ! - : typeclass_instances.
Hint Mode Adjoint ! - ! - - ! : typeclass_instances.
Infix "⊣" := Adjoint (no associativity, at level 70).
Lemma transpose : forall `{Adjoint X Y F G, !Functor F, !Functor G}
                    {A : X} {B : Y},
    F A ⊢ B <-> A ⊢ G B.
Proof.
  move=> X Pr_X Y Pr_Y F G Adj Funct_F Funct_G A B; split=> D.
  - setoid_rewrite (adj_unit A); simpl; by apply: fmap'.
  - setoid_rewrite <- (adj_counit B); simpl; by apply: fmap'.
Qed.
Arguments transpose {_ _ _ _ _ _ Adj _ _ _ _} : rename.
(* False!
Lemma transpose_core : forall `{Adjoint X Y F G, !Functor F, !Functor G}
                         {A B},
    F (G A) ⟛ B <-> G A ⟛ G B.
Proof.
  move=> X Pr_X Y Pr_Y F G Adj Funct_F Funct_G A B; split=> E; split.
  - apply transpose, E.
  - case: E => E1 E2; apply transpose.
    setoid_rewrite E2; setoid_rewrite (adj_counit A); apply (adj_counit A).
  - apply transpose, E.
  - case: E => E1 E2; setoid_rewrite <- E2.
Qed.
*)
Lemma transpose_sufficient : forall `{Proset X, Proset Y, !Functor F, !Functor G},
    (forall (A : X) (B : Y), F A ⊢ B <-> A ⊢ G B) -> F ⊣ G.
Proof.
  move=> X Pr_X Y Pr_Y F Funct_F G Funct_G Transpose.
  constructor=> A /=; by apply/Transpose.
Qed.

Instance left_adjoint_cocontinuous `{Proset X, Proset Y} {F : X -> Y} {G : Y -> X}
         `{!F ⊣ G, !Functor F, !Functor G}
  : Cocontinuous F | 2.
Proof.
  move=> R J A [C U]; split=> [r | A' C'] /=; first by apply: fmap'.
  apply/transpose/U => r; apply/transpose/C'.
Qed.
Instance right_adjoint_continuous `{Proset X, Proset Y} {F : X -> Y} {G : Y -> X}
         `{!F ⊣ G, !Functor F, !Functor G}
  : Continuous G | 2.
Proof.
  move=> R J A [C U]; split=> [r | A' C'] /=; first by apply: fmap'.
  apply/transpose/U => r; apply/transpose/C'.
Qed.

Definition universal_left_adjoint `{Proset X, Bicomplete Y} (G : Y -> X) (A : X) : Y
  := lim (fun B : {B0 | A ⊢ G B0} => `B).
Definition universal_right_adjoint `{Bicomplete X, Proset Y} (F : X -> Y) (B : Y) : X
  := colim (fun A : {A0 | F A0 ⊢ B} => `A).
Instance ula_functor `{Proset X, Bicomplete Y} {G : Y -> X}
  : Functor (universal_left_adjoint G).
Proof.
  move=> A B D /=.
  apply/lim_right => -[B0 ?] /=; apply/(lim_left _ (B0 ↾ _)); by etransitivity.
Qed.
Instance ura_functor `{Bicomplete X, Proset Y} {F : X -> Y}
  : Functor (universal_right_adjoint F).
Proof.
  move=> A B D /=.
  apply/colim_left => -[A0 ?] /=; apply/(colim_right _ (A0 ↾ _)); by etransitivity.
Qed.
(* These are not instances because we don't want them to conflict with other ones when we
   give more useful/explicit definitions. But they should be suitable for using to make
   instances very quickly! *)
Lemma universal_adjunction1 : forall `{Proset X, Bicomplete Y}
                                {G : Y -> X} `{!Functor G, !Continuous G},
    universal_left_adjoint G ⊣ G.
Proof.
  move=> X Pr_X Y Pr_Y Comp G Funct_G Cont_G; constructor=> [A | B] /=.
  - case: (create_lim_cone (fun B : {B0 | A ⊢ G B0} => `B)) => C U.
    apply/U => -[B0 ?] //.
    - by apply: (lim_left _ (B ↾ _)).
Qed.
Lemma universal_adjunction2 : forall `{Bicomplete X, Proset Y}
                                {F : X -> Y} `{!Functor F, !Cocontinuous F},
    F ⊣ universal_right_adjoint F.
Proof.
  move=> X Pr_X Comp Y Pr_Y F Funct_F Cont_F; constructor=> [A | B] /=.
  - by apply: (colim_right _ (A ↾ _)).
  - case: (create_colim_cocone (fun B : {A0 | F A0 ⊢ B} => `B)) => C U.
    apply/U => -[A0 ?] //.
Qed.

Instance compose_adjoint `{Proset X, Proset Y, Proset Z'}
         {F : Y -> Z'} {G : Z' -> Y} {F' : X -> Y} {G' : Y -> X}
         `{!F ⊣ G, F' ⊣ G', !Functor F, !Functor G, !Functor F', !Functor G'}
  : F ∘ F' ⊣ G' ∘ G.
Proof. apply/transpose_sufficient => * /=; rewrite 2!transpose //. Qed.
(* TODO Add this to the Hom section.
Instance precomp_adjoint `{Proset X, Proset Y, Proset Y'}
         {F : Y' -> Y} {G : Y -> Y'}
         `{!Functor F, !Functor G, !F ⊣ G}
  : Adjoint (X:=Y -> X) (.∘ F) (.∘ G).
Proof. constructor=> f /=. setoid_rewrite <- adj_counit. firstorder. Qed.
*)
Instance postcomp_adjoint `{Proset X, Proset X'} {Y}
         {F : X -> X'} {G : X' -> X}
         `{!F ⊣ G, !Functor F, !Functor G}
  : Adjoint (X:=Y -> X) (F ∘.) (G ∘.).
Proof. firstorder. Qed.
Instance const_lim_adjoint `{Bicomplete X} {Y} : @const X Y ⊣ lim.
Proof.
  constructor=> [A | P y] /=.
  - apply: lim_right => y; by simpl.
  - apply/lim_left.
Qed.
Instance colim_const_adjoint `{Bicomplete X} {Y} : colim ⊣ @const X Y.
Proof.
  constructor=> [P y | A] /=.
  - apply/colim_right.
  - apply: colim_left => y; by simpl.
Qed.

(* TODO Are these monoidal? Other instances? etc *)
Definition lim' `{Bicomplete X} {Y Y'} (f : Y -> Y') (P : Y -> X) : Y' -> X :=
  fun y' => lim (fun y : sig (fun y0 => f y0 = y') => P (`y)).
Arguments lim' {_ _ _ _ _} f P _ /.
Instance subst_lim'_adjoint `{Bicomplete X} {Y Y'} {f : Y -> Y'} : (.∘ f) ⊣ lim' f.
Proof.
  constructor=> [P y' | P y] /=.
  - apply: lim_right => -[y0 /= <-] //.
  - by apply: (lim_left _ (y ↾ _)).
Qed.
Definition colim' `{Bicomplete X} {Y Y'} (f : Y -> Y') (P : Y -> X) : Y' -> X :=
  fun y' => colim (fun y : sig (fun y0 => f y0 = y') => P (`y)).
Arguments colim' {_ _ _ _ _} f P _ /.
Instance colim'_subst_adjoint `{Bicomplete X} {Y Y'} {f : Y -> Y'} : colim' f ⊣ (.∘ f).
Proof.
  constructor=> [P y | A y'] /=.
  - by apply: (colim_right _ (y ↾ _)).
  - apply: colim_left => -[y0 /= <-] //.
Qed.

Lemma oplax_adjoint_lax : forall `{MonSet X, MonSet Y} {F : X -> Y} {G : Y -> X}
                            `{!F ⊣ G, !Functor F, !Functor G},
    OplaxMon F <-> LaxMon G.
Proof.
  move=> X Pr_X MS_X Y Pr_Y MS_Y F G Adj Funct_F Funct_G.
  split=> [Oplax | Lax]; constructor=> [| A B]; apply/transpose.
  - apply/pres_memp_op.
  - setoid_rewrite pres_tens_op; apply/bimap; apply/adj_counit.
  - apply/pres_memp.
  - setoid_rewrite <- pres_tens; apply/bimap; apply/adj_unit.
Qed.
Corollary adjoint_equivalence_strong : forall `{MonSet X, MonSet Y} {F : X -> Y} {G : Y -> X}
                                         `{!F ⊣ G, !G ⊣ F, !Functor F, !Functor G},
    StrongMon F <-> StrongMon G.
Proof. split; constructor; try apply/oplax_adjoint_lax. Qed.
(* The next one may only work because we are in prosets---I should go check. *)
Corollary triple_strong : forall `{MonSet X, MonSet Y} {F : X -> Y} {G : Y -> X} {H : X -> Y}
                            `{!F ⊣ G, !G ⊣ H, !Functor F, !Functor G, !Functor H},
    TCAnd (OplaxMon F) (LaxMon H) <-> StrongMon G.
Proof. split=> [[Oplax Lax] | Strong]; split; apply/oplax_adjoint_lax. Qed.

Definition r_product `{Bicomplete X} (B : X) : X -> X := fun A => A × B.
Definition r_tensor `{MonSet X} (B : X) : X -> X := fun A => A ⊗ B.
Arguments r_product {_ _ _} _ _ /.
Arguments r_tensor {_ _ _} _ _ /.
Instance product_functor `{Bicomplete X} : Functor (product (X:=X)).
Proof.
  move=> A A' D B.
  apply: product_right.
  - by etransitivity; first apply: product_left1.
  - apply: product_left2.
Qed.
Instance product_functor' `{Bicomplete X} {A : X} : Functor (product A).
Proof.
  move=> B B' D.
  apply: product_right.
  - apply: product_left1.
  - by etransitivity; first apply: product_left2.
Qed.
Instance coproduct_functor `{Bicomplete X} : Functor (coproduct (X:=X)).
Proof.
  move=> A A' D B.
  apply: coproduct_left.
  - by etransitivity; last apply: coproduct_right1.
  - apply: coproduct_right2.
Qed.
Instance coproduct_functor' `{Bicomplete X} {A : X} : Functor (coproduct A).
Proof.
  move=> B B' D.
  apply: coproduct_left.
  - apply: coproduct_right1.
  - by etransitivity; last apply: coproduct_right2.
Qed.
Instance tensor_functor `{MonSet X} : Functor (pro_tens (X:=X)).
Proof. move=> A A' D1 B; by apply: bimap. Qed.
Instance tensor_functor' `{MonSet X} {A : X} : Functor (pro_tens A).
Proof. move=> B B' D /=; by apply: bimap. Qed.
Instance r_product_functor `{Bicomplete X} {B : X} : Functor (r_product B).
Proof.
  move=> A A' D.
  apply: product_right.
  - by etransitivity; first apply: product_left1.
  - apply: product_left2.
Qed.
Instance r_tensor_functor `{MonSet X} {B : X} : Functor (r_tensor B).
Proof. move=> A A' D /=; by apply: bimap. Qed.
Instance exponential_functor `{Frame X} {A : X} : Functor (exponential A).
Proof.
  move=> B B' D.
  apply/product_exponential; setoid_rewrite <- D; by apply/product_exponential.
Qed.
Instance internal_hom_functor `{ClosedMonSet X} {A : X} : Functor (internal_hom A).
Proof.
  move=> B B' D.
  apply/tensor_hom; setoid_rewrite <- D; by apply/tensor_hom.
Qed.
Instance l_internal_hom_functor `{LClosedMonSet X} {A : X} : Functor (l_internal_hom A).
Proof.
  move=> B B' D.
  apply/l_tensor_hom; setoid_rewrite <- D; by apply/l_tensor_hom.
Qed.
Instance r_product_exponential_adjoint `{Frame X} {A : X}
  : r_product A ⊣ exponential A.
Proof. apply transpose_sufficient => * /=; apply/product_exponential. Qed.
Instance r_tensor_internal_hom_adjoint `{ClosedMonSet X} {A : X}
  : r_tensor A ⊣ internal_hom A.
Proof. apply transpose_sufficient => * /=; apply/tensor_hom. Qed.
Instance tensor_l_internal_hom_adjoint `{LClosedMonSet X} {A : X}
  : pro_tens A ⊣ l_internal_hom A.
Proof. apply transpose_sufficient => * /=; apply/l_tensor_hom. Qed.

Definition r_product_exponential_adjoint_sufficient `{Bicomplete X}
           (exponential' : X -> X -> X) `{forall A, Functor (exponential' A)}
           (Adj : forall (A : X), r_product A ⊣ exponential' A) : Frame X
  := {| exponential := exponential'; product_exponential A B C := transpose |}.
Definition r_product_cocontinuous_sufficient `{Bicomplete X}
           `(Cocont : forall A, Cocontinuous (r_product A))
  : Frame X
  := {| exponential A := universal_right_adjoint (r_product A);
        product_exponential A B C :=
          transpose (Adj:=universal_adjunction2 (F:=r_product B)) |}.
Definition r_tensor_internal_hom_adjoint_sufficient `{MonSet X}
           (internal_hom' : X -> X -> X) `{forall A, Functor (internal_hom' A)}
           (Adj : forall (A : X), r_tensor A ⊣ internal_hom' A) : ClosedMonSet X
  := {| internal_hom := internal_hom'; pro_tensor_hom A B C := transpose |}.
Definition r_tensor_cocontinuous_sufficient `{Bicomplete X, !MonSet X}
           `(Cocont : forall A, Cocontinuous (r_tensor A))
  : ClosedMonSet X
  := {| internal_hom A := universal_right_adjoint (r_tensor A);
        pro_tensor_hom A B C :=
          transpose (Adj:=universal_adjunction2 (F:=r_tensor B)) |}.
Definition tensor_l_internal_hom_adjoint_sufficient `{MonSet X}
           (l_internal_hom' : X -> X -> X) `{forall A, Functor (l_internal_hom' A)}
           (Adj : forall (A : X), pro_tens A ⊣ l_internal_hom' A) : LClosedMonSet X
  := {| l_internal_hom := l_internal_hom'; pro_l_tensor_hom A B C := transpose |}.
Definition tensor_cocontinuous_sufficient `{Bicomplete X, !MonSet X}
           `(Cocont : forall A, Cocontinuous (pro_tens A))
  : LClosedMonSet X
  := {| l_internal_hom A := universal_right_adjoint (pro_tens A);
        pro_l_tensor_hom A B C :=
          transpose (Adj:=universal_adjunction2 (F:=pro_tens B)) |}.

Lemma left_adjoint_unique : forall `{Proset X, Proset Y} {F F' : X -> Y} {G : Y -> X}
                              `{!F ⊣ G, !Functor F, !Functor F', !Functor G},
    F' ⊣ G -> F' ⟛ F.
Proof.
  move=> X Pr_X Y Pr_Y F F' G Adj1 Funct_F Funct_F' Funct_G Adj2.
  enough (H : forall (F F' : X -> Y) `{!Functor F, !Functor F'}, F ⊣ G -> F' ⊣ G -> F ⊢ F')
    by by split; apply H.
  move=> {F F' Adj1 Adj2 Funct_F Funct_F'} F F' ? ? Adj1 Adj2 A.
  apply/(transpose (Adj:=Adj1))/(adj_unit (Adjoint:=Adj2)).
Qed.
Lemma right_adjoint_unique : forall `{Proset X, Proset Y} {F : X -> Y} {G G' : Y -> X}
                               `{!F ⊣ G, !Functor F, !Functor G, !Functor G'},
    F ⊣ G' -> G' ⟛ G.
Proof.
  move=> X Pr_X Y Pr_Y F G G' Adj1 Funct_F Funct_G Funct_G' Adj2.
  enough (H : forall (G G' : Y -> X) `{!Functor G, !Functor G'}, F ⊣ G -> F ⊣ G' -> G ⊢ G')
    by by split; apply H.
  move=> {G G' Adj1 Adj2 Funct_G Funct_G'} G G' ? ? Adj1 Adj2 A.
  apply/(transpose (Adj:=Adj2))/(adj_counit (Adjoint:=Adj1)).
Qed.

Record Adjunction {X Y : Type} `{Proset X, Proset Y} : Type :=
  {LeftAdjoint : X -> Y; RightAdjoint : Y -> X;
   LAdj_Functor : Functor LeftAdjoint; RAdj_Functor : Functor RightAdjoint;
   Adjunction_adjoint : LeftAdjoint ⊣ RightAdjoint}.
Arguments Adjunction _ _ {_ _}.
Existing Instances LAdj_Functor RAdj_Functor Adjunction_adjoint | 10.
Definition adj_compose `{Proset X, Proset Y, Proset Z}
        (Adj1 : Adjunction Y Z) (Adj2 : Adjunction X Y) : Adjunction X Z
  := {| LeftAdjoint := LeftAdjoint Adj1 ∘ LeftAdjoint Adj2;
        RightAdjoint := RightAdjoint Adj2 ∘ RightAdjoint Adj1 |}.


Definition Hom (X Y : Type) `{Proset X, Proset Y} : Type :=
  sig (Functor (X:=X) (Y:=Y)).
Definition ap_Hom (X Y : Type) `{Proset X, Proset Y} : Hom X Y -> X -> Y
  := sval.
Arguments ap_Hom {_ _ _ _} !_.
Coercion ap_Hom : Hom >-> Funclass.
Instance ap_Hom_functor `{Proset X, Proset Y} : Functor (ap_Hom (X:=X) (Y:=Y)).
Proof. move=> [F ?] [G ?] //. Qed.
Instance ap_Hom_proper `{Proset X, Proset Y}
  : Proper (pro_le ++> pro_le ++> pro_le) (ap_Hom (X:=X) (Y:=Y)).
Proof. move=> [? ?] [? ?] /= D ? ? D'; by setoid_rewrite D'. Qed.
Instance ap_Hom_proper' `{Proset X, Proset Y}
  : Proper (pro_le --> pro_le --> flip pro_le) (ap_Hom (X:=X) (Y:=Y)).
Proof. move=> [? ?] [? ?] /= D ? ? D' /=; by setoid_rewrite D'. Qed.
Instance functorial `{Proset X, Proset Y} (F : Hom X Y) : Functor F := proj2_sig _.
Definition hom_map : forall `{Proset X, Proset Y} {F F' : Hom X Y} {A},
    F ⊢ F' -> F A ⊢ F' A.
Proof. intros *; apply. Qed.
Definition in_Hom `{Proset X, Proset Y} (F : X -> Y) `{!Functor F} : Hom X Y :=
  F packed_with Functor.
Lemma Hom_core : forall `{Proset X, Proset Y} {F G : Hom X Y}, (forall A, F A ⟛ G A) <-> F ⟛ G.
Proof. compute; firstorder. Qed.
Definition Hom_id `{Proset X} : Hom X X := in_Hom id.
Definition Hom_compose `{Proset X, Proset Y, Proset Z'}
           (F : Hom Y Z') (G : Hom X Y) : Hom X Z' := in_Hom (F ∘ G).
Infix "○" := Hom_compose (at level 40).
Definition Hom_id_lident : forall `{Proset X, Proset Y} {F : Hom X Y},
    Hom_id ○ F ⟛ F.
Proof. by compute. Qed.
Definition Hom_id_rident : forall `{Proset X, Proset Y} {F : Hom X Y},
    F ○ Hom_id ⟛ F.
Proof. by compute. Qed.
Definition Hom_compose_assoc : forall `{Proset X, Proset Y, Proset Z', Proset W}
                                 {F : Hom Z' W} {G : Hom Y Z'} {H : Hom X Y},
    F ○ (G ○ H) ⟛ (F ○ G) ○ H.
Proof. by compute. Qed.

Program Instance Hom_bicomplete `{Proset X, Bicomplete Y} : Bicomplete (Hom X Y)
  := {| lim R J := lim (ap_Hom ∘ J) ↾ _; colim R J := colim (ap_Hom ∘ J) ↾ _ |}.
Next Obligation.
  move=> X Pr_X Y Pr_Y Comp R J A B D /=.
  apply: fmap' => r; by apply: functorial.
Qed.
Next Obligation.
  move=> X Pr_X Y Pr_Y Comp R J; rewrite /lim /=; split=> [r | A' Cone] y /=.
  - apply: lim_left.
  - apply: lim_right => r; apply: Cone.
Qed.
Next Obligation.
  move=> X Pr_X Y Pr_Y Comp R J A B D /=.
  apply: fmap' => r; by apply: functorial.
Qed.
Next Obligation.
  move=> X Pr_X Y Pr_Y Comp R J; rewrite /colim /=; split=> [r | A' Cocone] y /=.
  - rewrite -/ap_Hom; by setoid_rewrite <- colim_right.
  - apply: colim_left => r; apply: Cocone.
Qed.
Instance ap_Hom_bicomplete_continuous `{Proset X, Bicomplete Y}
  : Continuous (ap_Hom (X:=X) (Y:=Y)).
Proof. by apply/distrib_lim_sufficient. Qed.
Instance ap_Hom_bicomplete_cocontinuous `{Proset X, Bicomplete Y}
  : Cocontinuous (ap_Hom (X:=X) (Y:=Y)).
Proof. by apply/distrib_colim_sufficient. Qed.
(* Hmm. *)
Instance flip_functoriality {R} `{Proset Y, Proset X}
         {F : R -> Y -> X} `{!forall r, Functor (F r)}
  : Functor (flip F).
Proof. firstorder. Qed.
Lemma Hom_lim : forall `{Proset Y, Bicomplete X} {R} (J : R -> Hom Y X),
    lim J ⟛ (lim ∘ flip J) as_a Hom Y X.
Proof. by compute. Qed.
Lemma Hom_colim : forall `{Proset Y, Bicomplete X} {R} (J : R -> Hom Y X),
    colim J ⟛ (colim ∘ flip J) as_a Hom Y X.
Proof. by compute. Qed.

Definition Hom_eval_at `{Proset X, Proset Y} (y : Y) : Hom Y X -> X := fun F => F y.
Arguments Hom_eval_at {_ _ _ _} _ _ /.
Instance subst_functor3 `{Proset X, Proset Y} {y : Y}
  : Functor (X:=Hom Y X) (Hom_eval_at y).
Proof. compute; firstorder. Qed.

Definition Endo (X : Type) `{Proset X} := Hom X X.
Identity Coercion endo_to_hom : Endo >-> Hom.
Instance compose_monoidal `{Proset X} :
  Monoidal (pro_le (X:=Endo X)) Hom_compose Hom_id.
Proof.
  constructor.
  - move=> F F' G G' D1 D2 A /=.
    setoid_rewrite D1; by setoid_rewrite D2.
  - by setoid_rewrite Hom_id_lident.
  - by setoid_rewrite Hom_id_rident.
  - by setoid_rewrite Hom_compose_assoc.
Qed.
Instance endo_mset `{Proset X} : MonSet (Endo X)
  := {| pro_tens := Hom_compose |}.


Class Monoid `{MonSet X} (M : X) :=
  {eta : memp ⊢ M; mu : M ⊗ M ⊢ M}.
Hint Mode Monoid ! - - ! : typeclass_instances.
Class Comonoid `{MonSet X} (W : X) :=
  {epsilon : W ⊢ memp; delta : W ⊢ W ⊗ W}.
Hint Mode Comonoid ! - - ! : typeclass_instances.

Instance lax_pres_monoid `{LaxMon X Y F, !Functor F} {M : X} `{!Monoid M}
  : Monoid (F M) | 2.
Proof.
  constructor.
  - setoid_rewrite pres_memp; apply/fmap'/eta.
  - setoid_rewrite pres_tens; apply/fmap'/mu.
Qed.
Instance oplax_pres_comonoid `{OplaxMon X Y F, !Functor F} {W : X} `{!Comonoid W}
  : Comonoid (F W) | 2.
Proof.
  constructor.
  - setoid_rewrite <- pres_memp_op; apply/fmap'/epsilon.
  - setoid_rewrite <- pres_tens_op; apply/fmap'/delta.
Qed.

(* TODO MonClosed for {,co}kleisli and {,co}EM *)
Class Monad `{Proset X} (M : X -> X) `{!Functor M} :=
  {ret : forall A, A ⊢ M A;
   join : forall {A}, M (M A) ⊢ M A}.
Hint Mode Monad ! - ! - : typeclass_instances.
(* TODO Is it worth defining this (and maybe other concepts later too)
        by duality instead? *)
Class Comonad `{Proset X} (W : X -> X) `{!Functor W} :=
  {extract : forall A, W A ⊢ A;
   duplicate : forall {A}, W A ⊢ W (W A)}.
Hint Mode Comonad ! - ! - : typeclass_instances.
Lemma bind : forall `{Monad X M} {P Q},
    P ⊢ M Q -> M P ⊢ M Q.
Proof. move=> *; setoid_rewrite <- join at 2; by apply fmap'. Qed.
Lemma extend : forall `{Comonad X W} {P Q},
    W P ⊢ Q -> W P ⊢ W Q.
Proof. move=> *; setoid_rewrite duplicate at 1; by apply fmap'. Qed.

Lemma monad_monoid : forall `{Proset X} {M : X -> X} `{!Functor M},
    Monad M <-> Monoid (M as_a Endo X).
Proof.
  move=> X ? M ?; split=> ?; constructor.
  - apply: ret.
  - move=> ? /=; apply: join.
  - apply: (eta (M:=M as_a Endo X)).
  - apply: (mu (M:=M as_a Endo X)).
Qed.
Lemma comonad_comonoid : forall `{Proset X} {W : X -> X} `{!Functor W},
    Comonad W <-> Comonoid (W as_a Endo X).
Proof.
  move=> X ? W ?; split=> ?; constructor.
  - apply: extract.
  - move=> ? /=; apply: duplicate.
  - apply: (epsilon (W:=W as_a Endo X)).
  - apply: (delta (W:=W as_a Endo X)).
Qed.
Instance act_monad `{MonSet E, Proset X} {L : E -> Endo X} `{!LaxMon L, !Functor L}
         {M : E} `{!Monoid M}
  : Monad (L M).
Proof. apply monad_monoid; rewrite /= -sig_eta; typeclasses eauto. Qed.
Instance act_comonad `{MonSet E, Proset X} {R : E -> Endo X} `{!OplaxMon R, !Functor R}
         {W : E} `{!Comonoid W}
  : Comonad (R W).
Proof. apply comonad_comonoid; rewrite /= -sig_eta; typeclasses eauto. Qed.

Instance adjunction_monad `{Adjoint X Y F G, !Functor F, !Functor G} : Monad (G ∘ F)
  := {| ret := adj_unit; join _ := fmap' (F:=G) (adj_counit _) |}.
Instance adjunction_comonad `{Adjoint X Y F G, !Functor F, !Functor G} : Comonad (F ∘ G)
  := {| extract := adj_counit; duplicate _ := fmap' (F:=F) (adj_unit _) |}.

Record kleisli {X} (M : X -> X) := Kleisli {get_kleisli : X}.
Add Printing Constructor kleisli.
Arguments Kleisli {_}.
Arguments get_kleisli {_ _}.
Definition pro_kleisli `{Proset X} (M : X -> X) : kleisli M -> kleisli M -> Prop :=
  fun '(Kleisli _ a1) '(Kleisli _ a2) => a1 ⊢ M a2.
Arguments pro_kleisli {X _} M !a1 !a2 /.
Instance pro_kleisli_pro `{Monad X M} : PreOrder (pro_kleisli M).
Proof.
  constructor.
  - move=> [A] /=; apply: ret.
  - move=> [A] [B] [C] /= D1 D2.
    setoid_rewrite D1; setoid_rewrite D2; by setoid_rewrite join.
Qed.
Definition tens_kleisli {X} (M : X -> X) (T : X -> X -> X) (a1 a2 : kleisli M) : kleisli M
  := Kleisli M (T (get_kleisli a1) (get_kleisli a2)).
Arguments tens_kleisli {X} M T !a1 !a2 /.
Instance kleisli_monoidal `{Monad X M, !MonSet X, !LaxMon M} :
  Monoidal (pro_kleisli M) (tens_kleisli M pro_tens) (Kleisli M memp).
Proof.
  constructor.
  - move=> [A] [A'] [B] [B'] /= D1 D2; setoid_rewrite <- pres_tens; by apply: bimap.
  - move=> [A]; split=> /=; [setoid_rewrite (pro_lunit A) |
                           setoid_rewrite <- (proj2 (pro_lunit A))]; apply: ret.
  - move=> [A]; split=> /=; [setoid_rewrite (pro_runit A) |
                           setoid_rewrite <- (proj2 (pro_runit A))]; apply: ret.
  - move=> [A] [B] [C]; split=> /=; [setoid_rewrite (pro_massoc A B C) |
                                   setoid_rewrite <- (proj2 (pro_massoc A B C))];
                                apply: ret.
Qed.
Instance kleisli_sym `{Monad X M, !MonSet X, !SymMonSet X, !LaxMon M} :
  Sym (pro_kleisli M) (tens_kleisli M pro_tens).
Proof. move=> [A] [B] /=; setoid_rewrite pro_sym' at 1; apply: ret. Qed.
Instance kleisli_proset `{Monad X M} : Proset (kleisli M)
  := {| pro_le := pro_kleisli M |}.
Instance kleisli_mset `{Monad X M, !MonSet X, !LaxMon M} :
  MonSet (kleisli M) := {| pro_tens := tens_kleisli M pro_tens |}.
Instance kleisli_smset `{Monad X M, !MonSet X, !SymMonSet X, !LaxMon M} :
  SymMonSet (kleisli M) := {}.
Instance Kleisli_functor `{Monad X M} : Functor (Kleisli M).
Proof. move=> A B D /=; by setoid_rewrite <- ret. Qed.
Instance Kleisli_strongmon `{Monad X M, !MonSet X, !LaxMon M} :
  StrongMon (Kleisli M).
Proof. done. Qed.
Definition dekleisli {X} {M : X -> X} : kleisli M -> X
  := fun '(Kleisli _ A) => M A.
Instance dekleisli_functor `{Monad X M} : Functor (dekleisli (M:=M)).
Proof. move=> [A] [B] /= D; setoid_rewrite D; apply: join. Qed.
Instance dekleisli_laxmon `{Monad X M, !MonSet X, !LaxMon M} :
  LaxMon (dekleisli (M:=M)).
Proof. constructor=> [| [A] [B]] //=; firstorder. Qed.
Instance dekleisli_strongmon `{Monad X M, !MonSet X, !StrongMon M} :
  StrongMon (dekleisli (M:=M)).
Proof. constructor; constructor=> [| [A] [B]] //=; firstorder. Qed.
Instance kleisli_adjunction `{Monad X M} : Kleisli M ⊣ dekleisli.
Proof. apply/transpose_sufficient => A [B] //. Qed.

(* okay, maybe we should just be dualizing *)
Record cokleisli {X} (F : X -> X) := Cokleisli {get_cokleisli : X}.
Add Printing Constructor cokleisli.
Arguments Cokleisli {_}.
Arguments get_cokleisli {_ _}.
Definition pro_cokleisli `{Proset X} (W : X -> X) : cokleisli W -> cokleisli W -> Prop :=
  fun '(Cokleisli _ a1) '(Cokleisli _ a2) => W a1 ⊢ a2.
Arguments pro_cokleisli {X _} W !a1 !a2 /.
Instance pro_cokleisli_pro `{Comonad X W} : PreOrder (pro_cokleisli W).
Proof.
  constructor.
  - move=> [A] /=; apply: extract.
  - move=> [A] [B] [C] /= D1 D2.
    setoid_rewrite duplicate; setoid_rewrite D1; by setoid_rewrite D2.
Qed.
(* TODO Is this monoidal? *)
Instance cokleisli_proset `{Comonad X W} : Proset (cokleisli W)
  := {| pro_le := pro_cokleisli W |}.
Instance Cokleisli_functor `{Comonad X W} : Functor (Cokleisli W).
Proof. move=> A B D /=; by setoid_rewrite extract. Qed.
Definition decokleisli {X} {W : X -> X} : cokleisli W -> X
  := fun '(Cokleisli _ A) => W A.
Instance decokleisli_functor `{Comonad X W} : Functor (decokleisli (W:=W)).
Proof. move=> [A] [B] /= D; setoid_rewrite <- D; apply: duplicate. Qed.
Instance cokleisli_adjunction `{Comonad X W} : decokleisli ⊣ Cokleisli W.
Proof. apply/transpose_sufficient => -[A] B //. Qed.

Class Alg `{Proset X} (F : X -> X) (A : X) : Prop :=
  alg : F A ⊢ A.
Class Coalg `{Proset X} (F : X -> X) (A : X) : Prop :=
  coalg : A ⊢ F A.
Hint Mode Alg ! - ! ! : typeclass_instances.
Hint Mode Coalg ! - ! ! : typeclass_instances.
Arguments alg {_ _ _} _ {_}.
Arguments coalg {_ _ _} _ {_}.
Instance adj_alg_coalg `{Proset X} {F G : X -> X}
         `{!F ⊣ G, !Functor F, !Functor G} `{!Alg F A}
  : Coalg G A | 3.
Proof. by apply transpose. Qed.
Instance adj_coalg_alg `{Proset X} {F G : X -> X}
         `{!F ⊣ G, !Functor F, !Functor G} `{!Coalg G A}
  : Alg F A | 3.
Proof. by apply transpose. Qed.
Instance free_alg `{Monad X M} {A} : Alg M (M A) := join.
Instance cofree_coalg `{Comonad X W} {A} : Coalg W (W A) := duplicate.
Lemma conecessitation : forall `{Alg X F Q, !Functor F} {P},
    P ⊢ Q -> F P ⊢ Q.
Proof.
  move=> X X_Pr F Q Alg_Q Funct_F P.
  setoid_rewrite <- (alg Q) at 2; apply fmap'.
Qed.
Lemma necessitation : forall `{Coalg X F P, !Functor F} {Q},
    P ⊢ Q -> P ⊢ F Q.
Proof.
  move=> X X_Pr F P Coalg_P Funct_F Q.
  setoid_rewrite (coalg P) at 2; apply fmap'.
Qed.

Definition EM `{Proset X} (M : X -> X) : Type :=
  sig (Alg M).
Instance EM_alg `{Proset X} {M : X -> X} (A : EM M) : Alg M (`A) := proj2_sig _.
Definition free_EM `{Monad X M} : X -> EM M := M ↑.
Arguments free_EM {_ _} M {_ _}.
Instance EM_reflection `{Monad X M} : free_EM M ⊣ sval.
Proof. constructor=> [A | [A Alg_A]] /=; [apply: ret | apply: alg]. Qed.

Definition coEM `{Proset X} (W : X -> X) : Type :=
  sig (Coalg W).
Instance coEM_alg `{Proset X} {W : X -> X} (A : coEM W) : Coalg W (`A) := proj2_sig _.
Definition cofree_coEM `{Comonad X W} : X -> coEM W := W ↑.
Arguments cofree_coEM {_ _} W {_ _}.
Instance coEM_coreflection `{Comonad X W} : sval ⊣ cofree_coEM W.
Proof. constructor=> [[A Coalg_A] | A] /=; [apply: coalg | apply: extract]. Qed.

Section AdjZero.
  Context `{Proset X, Proset Y} {F : X -> Y} {G : Y -> X}
          `{!F ⊣ G, !Functor F, !Functor G}.
  Let M : X -> X := G ∘ F.
  Local Notation I := (kleisli M).
  Local Notation T := (EM M).

  Definition from_kleisli : I -> Y := F ∘ dekleisli.
  Definition to_kleisli   : Y -> I := Kleisli M ∘ G.
  Lemma from_kleisli_triangle : G ∘ from_kleisli ⟛ dekleisli.
  Proof.
    apply/pw_core0 => -[A]; rewrite /from_kleisli /=.
    split; [apply (join (M:=M)) | apply (ret (M:=M))].
  Qed.
  Lemma to_kleisli_triangle : to_kleisli ∘ F ⟛ Kleisli M.
  Proof. split=> A //=; setoid_rewrite <- ret; apply (adj_unit A). Qed.
  Lemma from_kleisli_unique : forall `{!Functor FK},
      G ∘ FK ⟛ dekleisli -> FK ⟛ from_kleisli.
  Proof.
  (* TODO *)
  Abort.
End AdjZero.

Section AdjointMW.
  Context `{Proset X} {M : X -> X} {W : X -> X}
          `{Adj : M ⊣ W, !Functor M, !Functor W, !Monad M, !Comonad W}.

  Definition cofree_EM : X -> EM M := W ↑.
  Definition free_coEM : X -> coEM W := M ↑.
  Global Instance EM_coreflection : sval ⊣ cofree_EM.
  Proof.
    constructor=> [[A Coalg_A] | A] /=; [by apply transpose, alg | apply extract].
  Qed.
  Global Instance coEM_reflection : free_coEM ⊣ sval.
  Proof.
    constructor=> [A | [A Alg_A]] /=; [apply ret | by apply transpose, coalg].
  Qed.

  Definition transpose_EM : EM M -> coEM W := sval ↑.
  Definition transpose_coEM : coEM W -> EM M := sval ↑.
  Lemma transpose_EM_inverse : transpose_EM ∘ transpose_coEM ⟛ id.
  Proof. by compute. Qed.
  Lemma transpose_coEM_inverse : transpose_coEM ∘ transpose_EM ⟛ id.
  Proof. by compute. Qed.
  Global Instance transpose_EM_adjunction : transpose_EM ⊣ transpose_coEM.
  Proof. constructor; by compute. Qed.
  Global Instance transpose_coEM_adjunction : transpose_coEM ⊣ transpose_EM.
  Proof. constructor; by compute. Qed.
  Lemma transpose_EM_triple_iso :
    [/\ transpose_EM ∘ free_EM M ⟛ free_coEM, sval ∘ transpose_EM ⟛ sval &
        transpose_EM ∘ cofree_EM ⟛ cofree_coEM W].
    Proof. by compute. Qed.
  Lemma transpose_coEM_triple_iso :
    [/\ transpose_coEM ∘ free_coEM ⟛ free_EM M, sval ∘ transpose_coEM ⟛ sval &
        transpose_coEM ∘ cofree_coEM W ⟛ cofree_EM].
    Proof. by compute. Qed.
End AdjointMW.
Arguments cofree_EM {_ _} M {_ _ _ _ _}.
Arguments free_coEM {_ _ _} W {_ _ _ _}.
(* So any F ⊣ G ⊣ H between X and Y (F : X -> Y) can be resliced into either
   for F and H and same codomain for G, in either of the forms
   free_EM ⊣ sval ⊣ cofree_EM or free_coEM ⊣ sval ⊣ cofree_coEM, which are isomorphic;
   and ... TODO Finish working out the details of this! *)


Local Set Universe Polymorphism.

Definition PSh (X : Type) `{Proset X} : Type := Hom (op X) Prop.
Identity Coercion psh_to_hom : PSh >-> Hom.

Lemma pfunctorial : forall `{Proset X} {F : PSh X} {A B : op X},
    A ⊢ B -> F A -> F B.
Proof. move=> X Pr F A B; apply: (functorial F). Qed.
Lemma psh_map : forall `{Proset X} {F F' : PSh X} {A : op X},
    F ⊢ F' -> F A -> F' A.
Proof. intros *; apply. Qed.
Hint Resolve pfunctorial psh_map.

Program Definition yo `{Proset X} (A : X) : PSh X :=
  fun B => Op A ⊢ B.
Next Obligation. move=> * ? * ?; by etransitivity. Qed.
Lemma yoneda : forall `{Proset X} {A : X} {F : PSh X}, yo A ⊢ F <-> F (Op A).
Proof.
  move=> X ND A F; split.
  - apply; by simpl.
  - move=> FA A' D; by apply: pfunctorial FA.
Qed.
Instance yo_functor `{Proset X} : Functor (yo (X:=X)).
Proof. move=> * [?] /= ?; by etransitivity. Qed.
Instance yo_full `{Proset X} : Full (yo (X:=X)).
Proof. move=> A B /(_ (Op A)) /=; by apply. Qed.
Definition representable `{Proset X} (F : PSh X) : Prop :=
  exists A, F ⟛ yo A.

Program Instance psh_cartesian_closed `{Proset X} : Frame (PSh X)
  := {| exponential F G A := forall B, A ⊢ B -> F B -> G B |}.
Next Obligation. move=> X Pr F G A B D H C D'; apply: H; by etransitivity. Qed.
Next Obligation.
  move=> X Pr F G H /=; split.
  - move=> Uncurried A FA B D GB.
    apply: Uncurried => -[] //=; by apply: pfunctorial.
  - move=> Curried A /= PA; apply: (Curried A) => //.
    + apply: (PA true).
    + apply: (PA false).
Qed.

(*
Definition unyo `{Bicomplete X} (F : PSh X) (G : _) : X
  := colim (fun A : sig F => G (get_op (`A))).
Instance unyo_functor : Functor unyo
*)

Lemma coyoneda : forall `{Proset X} {F : PSh X},
    F ⟛ colim (fun A : sig F => yo (get_op (`A))).
Proof.
  move=> X Pr_X F; split.
  - move=> -[A] H /=; by exists (Op A ↾ H).
  - apply: colim_left => -[[B] /= H]; by apply/yoneda.
Qed.
Definition PShF `{Proset X, Proset Y} (G : X -> Y) (F : PSh X) : PSh Y :=
  colim (fun A : sig F => yo (G (get_op (`A)))).
Instance PShF_functor `{Functor X Y G} : Functor (PShF G).
Proof. move=> F F' D [B] /= [[A FA] /= D']; by exists (A ↾ D _ FA). Qed.
Lemma PShF_yo : forall `{Functor X Y F} {A},
    PShF F (yo A) ⟛ yo (F A).
Proof.
  move=> X Pr_X Y Pr_Y F Funct_F A.
  etransitivity; last by symmetry; apply: coyoneda.
  split.
  - apply: colim_left => -[[B] /= H] [C] /= D; unshelve by eexists (Op (F B) ↾ _).
    simpl; by apply: fmap'.
  - apply: colim_left => -[[B] /= H] [C] /= D; unshelve eexists (Op A ↾ _).
    + done.
    + by etransitivity.
Qed.
Lemma PShF_representable : forall `{Functor X Y G} {F},
    representable F -> representable (PShF G F).
Proof. move=> X Pr_X Y Pr_Y G Funct_G F [B H]; eexists; rewrite H PShF_yo //. Qed.
Instance yo_continuous `{Proset X} : Continuous (yo (X:=X)).
Proof.
  move=> R A J [C U]; split=> [r | A' C'] [B] /=.
  - move=> D; by etransitivity.
  - move=> /yoneda D; apply: U => r; apply/yo_full; etransitivity; firstorder.
Qed.
Lemma yo_colim1 : forall `{Bicomplete X} {Y} {A : Y -> X},
    colim (yo ∘ A) ⊢ yo (colim A).
Proof.
  move=> X Pr_X Comp Y A; apply: colim_left => [B] a /=.
  apply/(fmap' (F:=yo))/colim_right.
Qed.
Lemma yo_colim2 : forall `{Bicomplete X} {Y} {A : Y -> X} {F : PSh X},
    representable F -> colim (yo ∘ A) ⊢ F -> yo (colim A) ⊢ F.
Proof.
  move=> X Pr_X Comp Y A F [B [H H']] D; setoid_rewrite <- H'.
  apply/fmap'/colim_universal => b.
  apply: (unfmap (F:=yo (X:=X))); setoid_rewrite <- H; setoid_rewrite <- D.
  apply: (colim_right (yo ∘ A)).
Qed.
Lemma yo_colim3 : forall `{Bicomplete X} {Y} {A : Y -> X},
    colim (yo ∘ A) ⟛ yo (colim A) <-> representable (colim (yo ∘ A)).
Proof.
  move=> X Pr_X Comp Y A; split=> [| Rep]; first by eexists.
  split; [apply: yo_colim1 | by apply: yo_colim2].
Qed.
Lemma yo_colim4 : forall `{Bicomplete X} {Y} {A : Y -> X} {I} {F : PSh X},
    colim (yo ∘ A) ⊢ yo I -> yo I ⊢ F -> yo (colim A) ⊢ F.
Proof.
  move=> X Pr_X Comp Y A I F D1 D2.
  setoid_rewrite <- D2; apply: yo_colim2; unfold representable; firstorder.
Qed.
Lemma bicomplete_rep : forall `{Bicomplete X} {F : PSh X},
    representable F <-> Continuous F.
Proof.
  move=> X Pr Comp F; split.
  - move=> [A E]; apply/distrib_lim_sufficient => R J.
    rewrite /= /all /=; setoid_rewrite E at 1; setoid_rewrite <- (proj2 E).
    apply: colim_left.
  - move=> Cont; exists (colim (fun A : sig F => get_op (`A))); split.
    + etransitivity; [apply: (proj1 coyoneda) | apply: yo_colim1].
    + apply/yoneda/(distrib_lim (F:=F)) => [[A] FA] //.
Qed.

Program Definition day `{MonSet X} (F G : PSh X) : PSh X :=
  fun A => exists A1 A2, A1 ⊗ A2 ⊢ A /\ F A1 /\ G A2.
Next Obligation.
  move=> X Pr SMS F G A B D [A1] [A2] [Cov_A] [F_A1] G_A2.
  exists A1, A2; intuition by etransitivity.
Qed.
Instance day_monoidal `{MonSet X} : Monoidal pro_le (day (X:=X)) (yo memp).
Proof.
  constructor.
  - move=> F F' G G' D1 D2 A [A1] [A2] [Cov] [F_A1] G_A2 /=.
    exists A1, A2; firstorder using psh_map.
  - move=> F; apply/Hom_core => A /=; split.
    + move=> [A1] [A2] [Cov] [D].
      apply: pfunctorial.
      setoid_rewrite <- Cov; setoid_rewrite <- (pro_lunit A2) at 1; by setoid_rewrite D.
    + exists memp, A; split; last split; [apply pro_lunit | by simpl | done].
  - move=> F; apply/Hom_core => A /=; split.
    + move=> [A1] [A2] [Cov] [F_A1] D.
      apply: pfunctorial F_A1.
      setoid_rewrite <- Cov; setoid_rewrite <- (pro_runit A1) at 1; by setoid_rewrite D.
    + exists A; exists memp; split; last split; [apply pro_runit | by simpl | done].
  - move=> F G H'; apply/Hom_core=> A /=; split.
    + move=> [A1] [A2] [Cov] [F_A1] [A21] [A22] [Cov'] [G_A21] H_A22.
      exists (A1 ⊗ A21), A22; split; last split.
      * setoid_rewrite <- Cov; setoid_rewrite <- pro_massoc; by setoid_rewrite Cov'.
      * by exists A1, A21.
      * done.
    + move=> [A1] [A2] [Cov] [] [A11] [A12] [Cov'] [F_A12] G_A22 H_A2.
      exists A11, (A12 ⊗ A2); split; last split.
      * setoid_rewrite <- Cov; setoid_rewrite pro_massoc; by setoid_rewrite Cov'.
      * done.
      * by exists A12, A2.
Qed.
Instance day_sym `{SymMonSet X} : Sym pro_le (day (X:=X)).
Proof.
  move=> F G A [A1] [A2] [Cov] [F_A1] G_A2.
  exists A2, A1; split; last by done.
  etransitivity; [apply sym | done].
Qed.
Instance psh_mset `{MonSet X} : MonSet (PSh X)
  := {| pro_tens := day |}.
Instance psh_smset `{SymMonSet X} : SymMonSet (PSh X) := {}.
Instance yo_strongmon `{MonSet X} : StrongMon (yo (X:=X)).
Proof.
  constructor; constructor=> [| A B] [C] //=.
  - move=> [A1] [A2] [Cov] [D1] D2; by eapply transitivity, bimap.
  - move=> D; by exists (Op A), (Op B).
Qed.

Program Definition day_hom `{MonSet X} (F G : PSh X) : PSh X :=
  fun A => forall A1 A2, A ⊗ A1 ⊢ A2 -> F A1 -> G A2.
Next Obligation.
  move=> X Pr MS F G A B D H A1 A2 Cov F_A1.
  apply: H.
  - by setoid_rewrite D.
  - done.
Qed.
Instance day_closed_monoidal `{MonSet X}
  : ClosedMonoidal pro_le (day (X:=X)) day_hom.
Proof.
  move=> F G H'; split.
  - move=> Curried A FA A1 A2 Cov G_A1.
    apply: Curried => /=; exists A, A1; by split; last split.
  - move=> Uncurried A /= [A1] [A2] [Cov] [F_A1] G_A2.
      by apply: Uncurried.
Qed.
Program Definition day_l_hom `{MonSet X} (F G : PSh X) : PSh X :=
  fun A => forall A1 A2, A1 ⊗ A ⊢ A2 -> F A1 -> G A2.
Next Obligation.
  move=> X Pr MS F G A B D H A1 A2 Cov F_A1.
  apply: H.
  - by setoid_rewrite D.
  - done.
Qed.
Instance day_l_closed_monoidal `{MonSet X}
  : LClosedMonoidal pro_le (day (X:=X)) day_l_hom.
Proof.
  move=> F G H'; split.
  - move=> Curried A FA A1 A2 Cov G_A1.
    apply: Curried => /=; exists A1, A; by split; last split.
  - move=> Uncurried A /= [A1] [A2] [Cov] [F_A1] G_A2.
      by apply: Uncurried.
Qed.
Instance psh_cmset `{MonSet X} : ClosedMonSet (PSh X) := {}.
Instance psh_lcmset `{MonSet X} : LClosedMonSet (PSh X) := {}.


Lemma act_map : forall `{Proset E, Proset X} (F : E -> Endo X) `{!Functor F} {M N A},
    M ⊢ N -> F M A ⊢ F N A.
Proof. move=> ? ? ? ? F *; by apply: (fmap' (F:=F)). Qed.

(* The left and right regular actegories. *)
Definition lreg `{MonSet X} (A : X) : Endo X := pro_tens A as_a Endo X.
Instance lreg_functor `{MonSet X} : Functor (lreg (X:=X)).
Proof. move=> A A' D B /=; by apply bimap. Qed.
Instance lreg_strongmon `{MonSet X} : StrongMon (lreg (X:=X)).
Proof.
  constructor; constructor=> [A | A B C] /=.
  - apply pro_lunit.
  - apply pro_massoc.
  - apply pro_lunit.
  - apply pro_massoc.
Qed.
Definition sig_lreg `{MonSet X} {P : X -> Prop} `{!MonClosed P} : sig P -> Endo X
  := lreg ∘ sval.
Definition rreg `{MonSet X} (B : X) : Endo X := r_tensor B as_a Endo X.
Instance rreg_functor `{MonSet X} : Functor (rreg (X:=X)).
Proof. move=> B B' D A /=; by apply bimap. Qed.
Instance rreg_strongmon `{SymMonSet X} : StrongMon (rreg (X:=X)).
Proof.
  constructor; constructor=> [A | A B C] /=.
  - apply pro_runit.
  - setoid_rewrite <- pro_massoc; by setoid_rewrite (pro_sym B A).
  - apply pro_runit.
  - setoid_rewrite <- (proj1 (pro_massoc C B A)); by setoid_rewrite (pro_sym A B).
Qed.


Class Profunctor {X Y} `{Proset X, Proset Y} (P : X -> Y -> Prop) :=
  dimap : forall {A' A B B'}, A' ⊢ A -> B ⊢ B' -> P A B -> P A' B'.
Hint Mode Profunctor ! ! - - ! : typeclass_instances.
Definition lmap `{Profunctor X Y P} {A' A : X} {B : Y} (H : A' ⊢ A) : P A B -> P A' B :=
  dimap H (reflexivity _).
Definition rmap `{Profunctor X Y P} {A} {B B'} (H : B ⊢ B') : P A B -> P A B' :=
  dimap (reflexivity _) H.
(* TODO Double check that I haven't spuriously dropped conditions that still exist
        in the proset case!!!! *)
Class Tambara {E X Y} `{!Proset X, !Proset Y} (L : E -> Endo X) (R : E -> Endo Y)
      (P : X -> Y -> Prop) :=
  tamb : forall M {A B}, P A B -> P (L M A) (R M B).
Arguments tamb {_ _ _ _ _} L R {_ _} M {_ _}.
Hint Mode Tambara ! ! ! - - ! ! ! : typeclass_instances.

Lemma prof_core : forall `{Proset X, Proset Y} {P Q : X -> Y -> Prop},
    P ⟛ Q -> (forall A B, P A B <-> Q A B).
Proof. firstorder. Qed.
Lemma prof_core' : forall `{Proset X, Proset Y} {P Q : X -> Y -> Prop},
    P ⟛ Q <-> (forall A B, P A B <-> Q A B).
Proof. firstorder. Qed.

Instance profunctor_proper `{Profunctor X Y P} : Proper (pro_le --> pro_le ++> impl) P.
Proof. move=> ? ? ? ? ? ?; by apply: dimap. Qed.
Instance profunctor_proper' `{Profunctor X Y P}
  : Proper (pro_le ++> pro_le --> flip impl) P.
Proof. move=> ? ? ? ? ? ?; by apply: dimap. Qed.
Instance profunctor_proper_core `{Profunctor X Y P}
  : Proper (core pro_le ==> core pro_le ==> iff) P.
Proof. move=> ? ? [? ?] ? ? [? ?]; split; by apply: dimap. Qed.
Instance profunctor_proper_core' `{Profunctor X Y P}
  : Proper (core pro_le --> core pro_le --> iff) P.
Proof. move=> ? ? [? ?] ? ? [? ?]; split; by apply: dimap. Qed.

(* TODO This is an instance of something more general... *)
Instance tambara_compose `{Tambara E X Y L R P} {E'} {F : E' -> E}
  : Tambara (L ∘ F) (R ∘ F) P | 1.
Proof. move=> M A B /(tamb L R (F M)) //. Qed.
Lemma tambara_triple `{Proset E, Proset E', Proset X, Proset Y}
      {L : E -> Endo X} {R : E -> Endo Y} `{!Functor L, !Functor R}
      {Ψ_L : E' -> E} {Ψ_R : E' -> E} {Φ : E -> E'}
      `{!Φ ⊣ Ψ_L, !Ψ_R ⊣ Φ, !Functor Ψ_L, !Functor Ψ_R, !Functor Φ}
      `{!Tambara (L ∘ Ψ_L) (R ∘ Ψ_R) P, !Profunctor P}
  : Tambara L R P.
Proof.
  move=> M A B /(tamb _ _ (Φ M)); apply dimap; simpl.
  + apply/act_map/adj_unit.
  + apply/act_map/adj_counit.
Qed.

Instance tambara_rreg_rreg `{SymMonSet X} {P : X -> X -> Prop}
         `{!Profunctor P, !Tambara lreg lreg P}
  : Tambara rreg rreg P | 1.
Proof.
  move=> M A B PAB /=.
  apply: dimap; only 1,2: apply: sym.
    by apply: tamb.
Qed.
Instance tambara_lreg_rreg `{SymMonSet X} {P : X -> X -> Prop}
         `{!Profunctor P, !Tambara lreg lreg P}
  : Tambara lreg rreg P.
Proof.
  move=> M A B PAB /=.
  apply: rmap; first by apply: sym.
    by apply: tamb.
Qed.
Instance tambara_rreg_lreg `{SymMonSet X} {P : X -> X -> Prop}
         `{!Profunctor P, !Tambara lreg lreg P}
  : Tambara rreg lreg P.
Proof.
  move=> M A B PAB /=.
  apply: lmap; first by apply: sym.
    by apply: tamb.
Qed.

Instance hom_profunctor `{Proset X} : Profunctor (pro_le (X:=X)).
Proof. move=> A' A B B' H1 H2 H3; by etransitivity; first etransitivity. Qed.
Instance hom_tambara_F_F {E} `{Proset X} {F : E -> Endo X} : Tambara F F pro_le | 0.
Proof. move=> * /=; by apply fmap'. Qed.

Definition first `{SymMonSet X} `{!Profunctor P, !Tambara lreg lreg P}
           F {A B} (m : P A B) : P (A ⊗ F) (B ⊗ F)
  := tamb rreg rreg F m.
Definition second `{MonSet X} `{!Profunctor P, !Tambara lreg lreg P}
           F {A B : X} (m : P A B) : P (F ⊗ A) (F ⊗ B)
  := tamb lreg lreg F m.

Definition conjugate `{Proset E, Proset X, Proset X'}
         (L : E -> Endo X) `{!Functor L}
         (F : X' -> X) (G : X -> X') `{!Functor F, !Functor G}
         (M : E) : Endo X'
  := in_Hom (G ∘ L M ∘ F).
Instance conjugate_functor `{Proset E, Proset X, Proset X'}
         {L : E -> Endo X} `{!Functor L}
         {F : X' -> X} {G : X -> X'} `{!Functor F, !Functor G}
  : Functor (conjugate L F G).
Proof. move=> M1 M2 D B /=; apply/fmap'/act_map/D. Qed.
Instance conjugate_laxmon `{MonSet E, Proset X, Proset X'}
         {L : E -> Endo X} `{!Functor L, !LaxMon L}
         {F : X' -> X} {G : X -> X'} `{!F ⊣ G, !Functor F, !Functor G}
  : LaxMon (conjugate L F G).
Proof.
  constructor=> [B | M1 M2 B] /=.
  - apply transpose; apply: (pres_memp (F:=L)).
  - apply/fmap'.
    setoid_rewrite <- pres_tens; simpl; by apply fmap', transpose.
Qed.
Instance conjugate_oplaxmon `{MonSet E, Proset X, Proset X'}
         {L : E -> Endo X} `{!Functor L, !OplaxMon L}
         {F : X -> X'} {G : X' -> X} `{!F ⊣ G, !Functor F, !Functor G}
  : OplaxMon (conjugate L G F).
Proof.
  constructor=> [B | M1 M2 B] /=.
  - apply transpose; apply: (pres_memp_op (F:=L)).
  - apply/fmap'.
    setoid_rewrite pres_tens_op; simpl; by apply fmap', transpose.
Qed.

Definition pro_compose {X Y Z} (P : X -> Y -> Prop) (Q : Y -> Z -> Prop) : X -> Z -> Prop
  := fun x z => exists y, P x y /\ Q y z.
Arguments pro_compose {_ _ _} P Q _ _ /.
Infix "⦾" := pro_compose (at level 40).
Instance pro_compose_proper {X Y Z}
  : Proper (pro_le (X:=X -> Y -> Prop) ++> pro_le (X:=Y -> Z -> Prop) ++> pro_le) pro_compose.
Proof. move=> P P' D1 Q Q' D2 A C /=. setoid_rewrite D1; by setoid_rewrite D2. Qed.
Instance pro_compose_proper' {X Y Z}
  : Proper (pro_le (X:=X -> Y -> Prop) --> pro_le (X:=Y -> Z -> Prop) --> flip pro_le) pro_compose.
Proof. move=> P P' /= D1 Q Q' /= D2 A C /=; setoid_rewrite D1; by setoid_rewrite D2. Qed.
Instance compose_profunctor `{Proset X, Proset Y, Proset Z}
         {P : X -> Y -> Prop} {Q : Y -> Z -> Prop} `{!Profunctor P, !Profunctor Q}
  : Profunctor (P ⦾ Q) | 0.
Proof.
  move=> A' A C C' D1 D2 [B [PAB QBC]] /=.
  exists B; by split; [setoid_rewrite D1 | setoid_rewrite <- D2].
Qed.
(* TODO See if there's a way to get this to work better without sacrificing the
        right mode on Tambara. *)
Instance compose_tambara `{Tambara E X Y L LR P, Proset Z, !Tambara (Y:=Z) LR R Q}
  : Tambara L R (P ⦾ Q) | 0.
Proof. move=> M A C [B [/(tamb _ _ M) PAB /(tamb _ _ M) QBC]]; by eexists. Qed.
Lemma pro_compose_assoc : forall {X Y Z W} {P : X -> Y -> Prop} {Q : Y -> Z -> Prop} {R : Z -> W -> Prop},
    (P ⦾ Q) ⦾ R ⟛ P ⦾ (Q ⦾ R).
Proof. firstorder. Qed.

Definition star `{Proset X} {Y} (F : Y -> X) : X -> Y -> Prop
  := fun A B => A ⊢ F B.
Definition costar {X} `{Proset Y} (G : X -> Y) : X -> Y -> Prop
  := fun A B => G A ⊢ B.
Arguments star {_ _ _} F _ _ /.
Arguments costar {_ _ _} G _ _ /.
Lemma adj_star_costar : forall `{Adjoint X Y F G, !Functor F, !Functor G},
    costar F ⟛ star G.
Proof. move=> *; apply/prof_core' => ? ?; apply/transpose. Qed.
Instance star_profunctor `{Functor X Y F} : Profunctor (star F).
Proof. move=> A' A B B' D1 D2 /=; by setoid_rewrite D1 at 1; setoid_rewrite D2 at 1. Qed.
Instance costar_profunctor `{Functor X Y F} : Profunctor (costar F).
Proof. move=> A' A B B' D1 D2 /=; by setoid_rewrite D1 at 1; setoid_rewrite D2 at 1. Qed.
Definition starred {X Y Z} (P : X -> Y -> Prop) (F : Z -> Y) : X -> Z -> Prop
  := fun A C => P A (F C).
Definition costarred {X Y Z} (G : X -> Y) (P : Y -> Z -> Prop) : X -> Z -> Prop
  := fun A C => P (G A) C.
Arguments starred {_ _ _} P F _ _ /.
Arguments costarred {_ _ _} G P _ _ /.
Instance starred_profunctor `{Profunctor X Y P, Proset Z} {F : Z -> Y} `{!Functor F}
  : Profunctor (starred P F).
Proof. move=> ? ? ? ? D1 D2 /=; by setoid_rewrite D1; setoid_rewrite D2. Qed.
Instance costarred_profunctor `{Proset X, Profunctor Y Z P} {G : X -> Y} `{!Functor G}
  : Profunctor (costarred G P).
Proof. move=> ? ? ? ? D1 D2 /=; by setoid_rewrite D1; setoid_rewrite D2. Qed.
Lemma star_right : forall `{Profunctor X Y P} {Z} {F : Z -> Y},
    P ⦾ star F ⟛ starred P F.
Proof.
  move=> X Y ? ? P ? Z F; split=> A C /=.
  - move=> [B [PAB BFC]]; by setoid_rewrite <- BFC.
  - move=> PAFC; by exists (F C).
Qed.
Lemma costar_left : forall {X} `{Profunctor Y Z P} {G : X -> Y},
    costar G ⦾ P ⟛ costarred G P.
Proof.
  move=> X Y Z ? ? P ? G; split=> A C /=.
  - move=> [B [GAB PBC]]; by setoid_rewrite GAB.
  - move=> PGAC; by exists (G A).
Qed.
Lemma pro_hom_left_id : forall `{Profunctor X Y P},
    pro_le ⦾ P ⟛ P.
Proof. move=> *; apply: costar_left. Qed.
Lemma pro_hom_right_id : forall `{Profunctor X Y P},
    P ⦾ pro_le ⟛ P.
Proof. move=> *; apply: star_right. Qed.

Class LaxActMorph {E E'} `{Proset X}
      (F : E -> Endo X) (F' : E' -> Endo X) (Φ : E -> E') : Prop :=
  act_triangle : F ⊢ F' ∘ Φ.
Hint Mode LaxActMorph ! ! ! - ! ! ! : typeclass_instances.
Class OplaxActMorph {E E'} `{Proset X}
      (F : E -> Endo X) (F' : E' -> Endo X) (Φ : E -> E') : Prop :=
  act_triangle_op : F' ∘ Φ ⊢ F.
Hint Mode OplaxActMorph ! ! ! - ! ! ! : typeclass_instances.
Class StrongActMorph {E E'} `{Proset X}
      (F : E -> Endo X) (F' : E' -> Endo X) (Φ : E -> E') : Prop :=
  {strong_am_lax :> LaxActMorph F F' Φ;
   strong_am_oplax :> OplaxActMorph F F' Φ}.
Hint Mode StrongActMorph ! ! ! - ! ! ! : typeclass_instances.
Class LaxEquivariant {E} `{Proset X, Proset X'}
      (F : E -> Endo X) (F' : E -> Endo X') (Φ : X -> X') : Prop :=
  act_square M : F' M ∘ Φ ⊢ Φ ∘ F M.
Hint Mode LaxEquivariant ! ! - ! - ! ! ! : typeclass_instances.
Class OplaxEquivariant {E} `{Proset X, Proset X'}
      (F : E -> Endo X) (F' : E -> Endo X') (Φ : X -> X') : Prop :=
  act_square_op M : Φ ∘ F M ⊢ F' M ∘ Φ.
Hint Mode OplaxEquivariant ! ! - ! - ! ! ! : typeclass_instances.
Class StrongEquivariant {E} `{Proset X, Proset X'}
      (F : E -> Endo X) (F' : E -> Endo X') (Φ : X -> X') : Prop :=
  {strong_equ_lax :> LaxEquivariant F F' Φ;
   strong_equ_oplax :> OplaxEquivariant F F' Φ}.
Hint Mode StrongEquivariant ! ! - ! - ! ! ! : typeclass_instances.
Instance adjoint_lax_act_morph `{Adjoint E E' Φ Ψ, Proset X}
         {F : E -> Endo X} `{!Functor F}
  : LaxActMorph F (F ∘ Ψ) Φ.
Proof. move=> M /=; apply/fmap'/adj_unit. Qed.
Instance adjoint_oplax_act_morph `{Adjoint E' E Ψ Φ, Proset X}
         {F : E -> Endo X} `{!Functor F}
  : OplaxActMorph F (F ∘ Ψ) Φ.
Proof. move=> M /=; apply/fmap'/adj_counit. Qed.
Instance adjoint_lax_equivariant `{Proset E, Adjoint X X' Φ Ψ}
         {F : E -> Endo X'} `{!Functor F, !Functor Φ, !Functor Ψ}
  : LaxEquivariant F (conjugate F Φ Ψ) Ψ.
Proof. move=> M A /=; apply/fmap'/fmap'/adj_counit. Qed.
Instance adjoint_oplax_equivariant `{Proset E, Adjoint X X' Φ Ψ}
         {F : E -> Endo X} `{!Functor F, !Functor Ψ, !Functor Φ}
  : OplaxEquivariant F (conjugate F Ψ Φ) Φ.
Proof. move=> M A /=; apply/fmap'/fmap'/adj_unit. Qed.

Instance star_tambara  {E} `{Functor Y X Φ} {L : E -> Endo X} {R : E -> Endo Y}
         `{!LaxEquivariant R L Φ}
  : Tambara L R (star Φ).
Proof. move=> M A B /= /(tamb L L M) D; setoid_rewrite D; apply: act_square. Qed.
Lemma star_tambara_iff : forall {E} `{Functor Y X Φ} {L : E -> Endo X} {R : E -> Endo Y},
    LaxEquivariant R L Φ <-> Tambara L R (star Φ).
Proof.
  move=> E Y ? X ? Φ ? L R; split; first by typeclasses eauto.
  move=> Tamb M B /=; apply: tamb; by simpl.
Qed.
Instance costar_tambara  {E} `{Functor X Y Φ} {L : E -> Endo X} {R : E -> Endo Y}
         `{!OplaxEquivariant L R Φ}
  : Tambara L R (costar Φ).
Proof. move=> M A B /= /(tamb R R M) D; setoid_rewrite <- D; apply: act_square_op. Qed.
Lemma costar_tambara_iff : forall {E} `{Functor X Y Φ} {L : E -> Endo X} {R : E -> Endo Y},
    OplaxEquivariant L R Φ <-> Tambara L R (costar Φ).
Proof.
  move=> E X ? Y ? Φ ? L R; split; first by typeclasses eauto.
  move=> Tamb M B /=; apply: tamb; by simpl.
Qed.

Instance starred_tambara `{Tambara E X Y L LR P, Proset Z} `{!Profunctor P}
         {Φ : Z -> Y} `{!Functor Φ} {R : E -> Endo Z} `{!LaxEquivariant R LR Φ}
  : Tambara L R (starred P Φ).
Proof.
  assert (Tambara L R (P ⦾ star Φ)) by apply: (compose_tambara (LR:=LR)).
    by unfold Tambara; setoid_rewrite <- (prof_core star_right).
Qed.
Instance costarred_tambara `{Proset X, Tambara E Y Z LR R P} `{!Profunctor P}
         {Φ : X -> Y} `{!Functor Φ} {L : E -> Endo X} `{!OplaxEquivariant L LR Φ}
  : Tambara L R (costarred Φ P).
Proof.
  assert (Tambara L R (costar Φ ⦾ P)) by apply: (compose_tambara (LR:=LR)).
    by unfold Tambara; setoid_rewrite <- (prof_core costar_left).
Qed.

(*
Definition Star {X Y' Y} (P : X -> Y -> Prop) (G : Y' -> Y)
  : X -> Y' -> Prop := fun A B => P A (G B).
Instance Star_profunctor `{Profunctor X Y P} `{Proset Y'}
         {G : Y' -> Y} `{!Functor G} : Profunctor (Star P G).
Proof.
  move=> A' A B B' D1 D2.
  rewrite /Star; apply dimap; [done | by apply fmap'].
Qed.
Instance Star_tambara `{Tambara E X Y L R P}
         `{!Proset E, Proset Y', !Functor R, !Profunctor P}
         {F : Y -> Y'} {G : Y' -> Y} `{!F ⊣ G, !Functor F, !Functor G} :
  Tambara L (conjugate R G F) (Star P G).
Proof.
  move=> M A B; rewrite /Star /= => /(tamb _ _ M).
  apply: rmap; apply: (adj_unit _).
Qed.

Definition Costar {X X' Y} `{Proset X'} (F : X -> X') (P : X -> Y -> Prop)
  : X' -> Y -> Prop := fun A B => exists2 _A, P _A B & A ⊢ F _A.
Lemma Costar_adj : forall `{Proset X, Proset X', Proset Y, !F ⊣ G, !Functor F, !Functor G}
                     {P : X -> Y -> Prop} `{!Profunctor P} {A : X'} {B : Y},
    Costar G P A B <-> P (F A) B.
Proof.
  move=> X' Pr_X' X Pr_X Y Pr_Y F G Adj Funct_F Funct_G P Pro A B; split.
  - move=> [_A PAB /(transpose (Adj:=Adj)) D]; by apply: (lmap D).
  - move=> PFAB; exists (F A); [done | apply: (adj_unit (Adjoint:=Adj))].
Qed.
Instance Costar_profunctor `{Profunctor X Y P} `{Proset X'}
         {F : X -> X'} `{!Functor F} : Profunctor (Costar F P).
Proof.
  move=> A' A B B' D1 D2 [_A PAB D3].
    by eexists; [apply: rmap | etransitivity].
Qed.
Instance Costar_tambara `{Tambara E X Y L R P}
         `{!Proset E, Proset X', !Functor L, !Profunctor P}
         {F : X' -> X} {G : X -> X'} `{!F ⊣ G, !Functor F, !Functor G} :
  Tambara (conjugate L F G) R (Costar G P).
Proof.
  move=> M A B; rewrite !Costar_adj /= => /(tamb _ _ M); apply lmap, (adj_counit (F:=F)).
Qed.
*)


(* TODO This is a preorder!!! *)
Definition optic {E} `{Proset X, Proset Y} (L : E -> Endo X) (R : E -> Endo Y)
           (S : X) (T : Y) (A : X) (B : Y) : Prop :=
  forall P, Profunctor P -> Tambara L R P -> P A B -> P S T.

Definition optic_id {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y} {S T}
  : optic L R S T S T
  := fun _ _ _ => id.
(* TODO We could get a (semi?)lattice of kinds of optic like in Haskell... Hmm. *)
Definition optic_compose {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y}
           {S T A B W Z} (l : optic L R S T A B) (m : optic L R A B W Z)
  : optic L R S T W Z
  := fun _ _ _ => l _ _ _ ∘ m _ _ _.
Infix "∘∘" := optic_compose (at level 40).
Definition optic_frame {E} `{Proset X, Proset Y}
           {L : E -> Endo X} {R : E -> Endo Y} {M A B}
  : optic L R (L M A) (R M B) A B
  := fun P _ _ => tamb _ _ (P:=P) M.
Definition first_optic `{SymMonSet X} {P P' Q : X}
  : optic lreg lreg (P ⊗ Q) (P' ⊗ Q) P P'
  := fun _ _ _ => first _.
Definition second_optic `{MonSet X} {P Q Q' : X}
  : optic lreg lreg (P ⊗ Q) (P ⊗ Q') Q Q'
  := fun _ _ _ => second _.

(* TODO This is a crappy name. *)
Lemma yoneda_ish :
  forall `{Proset E, Proset E', Proset X, Proset Y}
    {L : E -> Endo X} {R : E -> Endo Y} {L' : E' -> Endo X} {R' : E' -> Endo Y},
    (forall {S T A B}, optic L R S T A B -> optic L' R' S T A B) <->
    (forall `{!Profunctor P}, Tambara L' R' P -> Tambara L R P).
Proof.
  split.
  - move=> Conv P Pro Tamb M A B.
    apply: (Conv _ _ _ _ optic_frame).
  - move=> Inst S T A B l P Pro Tamb /l //.
Qed.

Definition opticP1 {E} `{Proset X, Proset Y} (L : E -> Endo X) (R : E -> Endo Y)
           (A : X) (B : Y) (S : X) (T : Y) : Prop := optic L R S T A B.
Instance optic_profunctor1 {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y}
         {A B} : Profunctor (opticP1 L R A B).
Proof.
  move=> S' S T T' D1 D2 l P Pro Tamb PAB.
  apply: (dimap D1 D2 _); by apply: l.
Qed.
Instance optic_tambara1 {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y}
         {A B} : Tambara L R (opticP1 L R A B).
Proof. move=> M S T l P Pro Tamb /l /(tamb _ _ M) //. Qed.

Definition opticP2 {E} `{Proset X, Proset Y} (L : E -> Endo X) (R : E -> Endo Y)
           (S : X) (T : Y) (B : Y) (A : X) : Prop := optic L R S T A B.
Instance optic_profunctor2 {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y}
  {S T} : Profunctor (opticP2 L R S T).
Proof.
  move=> B B' A' A D1 D2 l P Pro Tamb PAB.
  apply: l; by apply: (dimap D2 D1 _).
Qed.

Definition smap {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y} {S' S T A B}
  : S' ⊢ S -> optic L R S T A B -> optic L R S' T A B := lmap (P:=opticP1 L R A B).
Definition tmap {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y} {S T T' A B}
  : T ⊢ T' -> optic L R S T A B -> optic L R S T' A B := rmap (P:=opticP1 L R A B).
Definition amap {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y} {S T A A' B}
  : A ⊢ A' -> optic L R S T A B -> optic L R S T A' B := rmap (P:=opticP2 L R S T).
Definition bmap {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y} {S T A B' B}
  : B' ⊢ B -> optic L R S T A B -> optic L R S T A B' := lmap (P:=opticP2 L R S T).

(* wait, some kind of weird mixture of lax and oplax going on here *)
Definition lax_action_2sided_morph `{Proset E, Proset E', Proset X, Proset X'}
           (Φ : E -> E') (F : X -> X') (L : E -> Endo X) (L' : E' -> Endo X') : Prop
  := forall M, F ∘ L M ⊢ L' (Φ M) ∘ F.
Definition oplax_action_2sided_morph `{Proset E, Proset E', Proset Y, Proset Y'}
           (Φ : E -> E') (G : Y -> Y') (R : E -> Endo Y) (R' : E' -> Endo Y') : Prop
  := forall M, R' (Φ M) ∘ G ⊢ G ∘ R M.
Definition exymap `{Proset E, Proset E', Proset X, Proset X', Proset Y, Proset Y'}
           {L : E -> Endo X} {L' : E' -> Endo X'} {R : E -> Endo Y} {R' : E' -> Endo Y'}
           (Φ : E -> E') (F : X -> X') (G : Y -> Y') `{!Functor Φ, !Functor F, !Functor G}
  : lax_action_2sided_morph Φ F L L' -> oplax_action_2sided_morph Φ G R R' ->
    forall {S T A B}, optic L R S T A B -> optic L' R' (F S) (G T) (F A) (G B).
Proof.
  move=> LaxAct OplaxAct S T A B l P Pro Tamb PFAGB; apply l, PFAGB.
  - move=> {l PFAGB A B} A' A B B' D1 D2.
    apply dimap; by apply fmap'.
  - move=> {l PFAGB A B} M A B /(tamb _ _ (Φ M)).
    apply: dimap; [apply: LaxAct | apply: OplaxAct].
Qed.

Definition emap `{Proset E, Proset E', Proset X, Proset Y}
           {L : E -> Endo X} {L' : E' -> Endo X} {R : E -> Endo Y} {R' : E' -> Endo Y}
           (Φ : E -> E') `{!Functor Φ, !LaxActMorph L L' Φ, !OplaxActMorph R R' Φ}
  : forall {S T A B}, optic L R S T A B -> optic L' R' S T A B.
Proof. apply: exymap; [apply: act_triangle | apply: act_triangle_op]. Qed.
Definition xmap `{Proset E, Proset X, Proset X', Proset Y}
           {L : E -> Endo X} {L' : E -> Endo X'} {R : E -> Endo Y}
           (Φ : X -> X') `{!Functor Φ, !OplaxEquivariant L L' Φ}
  : forall {S T A B}, optic L R S T A B -> optic L' R (Φ S) T (Φ A) B.
Proof. apply: (exymap id); [apply: act_square_op | by hnf]. Qed.
Definition ymap `{Proset E, Proset X, Proset Y, Proset Y'}
           {L : E -> Endo X} {R : E -> Endo Y} {R' : E -> Endo Y'}
           (Φ : Y -> Y') `{!Functor Φ, !LaxEquivariant R R' Φ}
  : forall {S T A B}, optic L R S T A B -> optic L R' S (Φ T) A (Φ B).
Proof. apply: (exymap id); [by hnf | apply: act_square]. Qed.

Definition emap_adj `{Proset E, Proset E', Proset X, Proset Y}
           {L : E -> Endo X} {R : E -> Endo Y} `{!Functor L, !Functor R}
           (Ψ_L : E' -> E) (Ψ_R : E' -> E) {Φ : E -> E'}
           `{!Φ ⊣ Ψ_L, !Ψ_R ⊣ Φ, !Functor Ψ_L, !Functor Ψ_R, !Functor Φ}
  : forall {S T A B}, optic L R S T A B -> optic (L ∘ Ψ_L) (R ∘ Ψ_R) S T A B.
Proof. apply: emap. Qed.
Definition xmap_adj `{Proset E, Proset X, Proset X', Proset Y}
           {L : E -> Endo X} {R : E -> Endo Y} `{!Functor L, !Functor R}
           (Φ : X -> X') {Ψ : X' -> X} `{!Φ ⊣ Ψ, !Functor Φ, !Functor Ψ}
  : forall {S T A B}, optic L R S T A B -> optic (conjugate L Ψ Φ) R (Φ S) T (Φ A) B.
Proof. apply: xmap. Qed.
Definition ymap_adj `{Proset E, Proset X, Proset Y, Proset Y'}
           {L : E -> Endo X} {R : E -> Endo Y} `{!Functor L, !Functor R}
           (Φ : Y -> Y') {Ψ : Y' -> Y} `{!Ψ ⊣ Φ, !Functor Φ, !Functor Ψ}
  : forall {S T A B}, optic L R S T A B -> optic L (conjugate R Ψ Φ) S (Φ T) A (Φ B).
Proof. apply: ymap. Qed.

Definition eunmap {E E'} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y}
           {F : E' -> E} {S T A B} (l : optic (L ∘ F) (R ∘ F) S T A B)
  : optic L R S T A B
  := fun _ _ _ => l _ _ _.

Definition eoptic {E} `{Proset X, Proset Y} (L : E -> Endo X) (R : E -> Endo Y)
           (A : X) (B : Y) : X -> Y -> Prop
  := fun S T => exists2 M, S ⊢ L M A & R M B ⊢ T.
Instance eoptic_profunctor {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y}
         {A B} : Profunctor (eoptic L R A B).
Proof.
  move=> S' S T T' D1 D2 [M Open Close].
  exists M; by etransitivity.
Qed.
Instance eoptic_tambara `{MonSet E, Proset X, Proset Y}
         {L : E -> Endo X} {R : E -> Endo Y}
         `{!Functor L, !Functor R, !LaxMon L, !OplaxMon R} {A B}
  : Tambara L R (eoptic L R A B).
Proof.
  move=> M S T [M' Open Close].
  exists (M ⊗ M').
  - setoid_rewrite <- pres_tens; by apply: functorial.
  - setoid_rewrite pres_tens_op; simpl; by apply: functorial.
Qed.

Definition optic_to_eoptic `{MonSet E, Proset X, Proset Y}
           {L : E -> Endo X} {R : E -> Endo Y}
           `{!Functor L, !Functor R, !LaxMon L, !OplaxMon R} {S T A B}
           (l : optic L R S T A B)
  : eoptic L R A B S T.
Proof. apply: l; exists memp; [apply: (pres_memp (F:=L)) |
                          apply: (pres_memp_op (F:=R))]. Qed.
Definition eoptic_to_optic {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y}
           {S T A B} (e : eoptic L R A B S T)
  : optic L R S T A B.
Proof.
  move: e => [M Open Close] P Pro Tamb PAB.
  apply (dimap Open Close), (tamb L R), PAB.
Qed.
Definition optic_iff_eoptic :
  forall `{MonSet E, Proset X, Proset Y}
    {L : E -> Endo X} {R : E -> Endo Y}
    `{!Functor L, !Functor R, !LaxMon L, !OplaxMon R} {S T A B},
    optic L R S T A B <-> eoptic L R A B S T.
Proof. move=> *; split; [apply: optic_to_eoptic | apply: eoptic_to_optic]. Qed.

(* Maybe I should learn how Yoneda reduction works. *)
Theorem optic_concrete :
  forall `{MonSet E, Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y}
    `{!Functor L, !Functor R, !LaxMon L, !OplaxMon R} {B}
    R_B `{!Hom_eval_at B ∘ R ⊣ R_B, !Functor R_B} {S T A},
    optic L R S T A B <-> S ⊢ L (R_B T) A.
Proof.
  move=> E Pr_E MS X Pr_X Y Pr_Y L R Funct_L Funct_R Lax Oplax B R_B Adj Funct_R_B S T A.
  rewrite optic_iff_eoptic /eoptic; split.
  - move=> [M Open /(transpose (Adj:=Adj)) Close].
    setoid_rewrite Open; by setoid_rewrite Close.
  - move=> D.
    eexists; last apply/(transpose (Adj:=Adj)); done.
Qed.
Arguments optic_concrete {_ _ _ _ _ _ _ _ _ _ _ _ _ _} R_B Adj {_ _ _ _} : rename.


Class Quantale (X : Type)
      `{!Proset X, !Bicomplete X, !MonSet X, !ClosedMonSet X, !LClosedMonSet X}.
Hint Mode Quantale ! - - - - - : typeclass_instances.
Instance quantale_def
         `{Proset X, !Bicomplete X, !MonSet X, !ClosedMonSet X, !LClosedMonSet X}
  : Quantale X.

Lemma quantale_tensor_prop :
  forall `{Quantale X} {P Q}, embed_prop (X:=X) P ⊗ embed_prop Q ⟛ embed_prop (P ∧ Q).
Proof.
  move=> X Pr Comp MS CMS LCMS Quant P Q; split.
  - rewrite tensor_hom; apply: embed_prop_left => H1; rewrite -tensor_hom.
    rewrite l_tensor_hom; apply: embed_prop_left => H2; rewrite -l_tensor_hom.
      by apply: embed_prop_right.
  - apply: embed_prop_left => -[H1 H2].
    transitivity (⊤ ⊗ ⊤ : X). (* wat *)
    + setoid_rewrite <- terminal_right at 2; by setoid_rewrite <- (proj2 (pro_lunit _)).
    + apply: bimap; by apply: embed_prop_right.
Qed.
(* TODO Put these in the right place. *)
Instance embed_prop_functor `{Bicomplete X} : Functor (embed_prop (X:=X)).
Proof. move=> P Q D; apply: embed_prop_left => ?; by apply/embed_prop_right/D. Qed.
Lemma mon_modus_ponens : forall `{ClosedMonSet X} {A B : X},
    (A ⊸ B) ⊗ A ⊢ B.
Proof. move=> X Pr MS CM A B; by rewrite tensor_hom. Qed.
Lemma l_mon_modus_ponens : forall `{LClosedMonSet X} {A B : X},
    A ⊗ (A ⊸̂ B) ⊢ B.
Proof. move=> X Pr MS CM A B; by rewrite l_tensor_hom. Qed.
Lemma prop_loop :
  forall `{Frame X} {P Q : X} {R : Prop},
    (P ⊢ embed_prop R) -> (R -> P ⊢ Q) -> P ⊢ Q.
Proof.
  move=> X Pr Comp CCl P Q R D1 D2.
  setoid_rewrite (product_right D1 (reflexivity _)).
  apply/product_exponential/embed_prop_left => H.
  apply/product_exponential; setoid_rewrite product_left2; by apply: D2.
Qed.

Definition mon_embed_prop `{MonSet X, !Bicomplete X} (P : Prop) : X :=
  colim (fun H : P => memp).
Arguments mon_embed_prop : simpl never.
Lemma mon_embed_prop_left : forall `{MonSet X, !Bicomplete X} {P : Prop} {Q : X},
    (P -> memp ⊢ Q) -> mon_embed_prop P ⊢ Q.
Proof. move=> X Pr Comp P Q D; by apply: colim_left. Qed.
Lemma mon_embed_prop_right : forall `{MonSet X, !Bicomplete X} {P : X} {Q : Prop},
    Q -> P ⊢ memp -> P ⊢ mon_embed_prop Q.
Proof. move=> X Pr MS Comp P Q H D; by setoid_rewrite <- colim_right. Qed.
Instance mon_embed_prop_functor `{MonSet X, !Bicomplete X}
  : Functor (mon_embed_prop (X:=X)).
Proof.
  move=> P Q D.
  apply: mon_embed_prop_left => ?; apply/mon_embed_prop_right.
  - by apply: D.
  - reflexivity.
Qed.

Lemma mon_tensor_prop :
  forall `{ClosedMonSet X, !Bicomplete X} {P Q},
    mon_embed_prop (X:=X) P ⊗ mon_embed_prop Q ⟛ mon_embed_prop (P ∧ Q).
Proof.
  move=> X Pr MS CMS Comp P Q; split.
  - rewrite tensor_hom; apply: mon_embed_prop_left => H1; rewrite -tensor_hom pro_lunit.
      by apply: fmap'.
  - apply: mon_embed_prop_left => -[H1 H2].
    rewrite -(pro_runit memp); apply: bimap; by apply: mon_embed_prop_right.
Qed.

Lemma mon_prop_loop :
  forall `{ClosedMonSet X, !Bicomplete X} {P Q R : X} {S : Prop},
    (P ⊢ mon_embed_prop S) -> (S -> Q ⊢ R) -> P ⊗ Q ⊢ R.
Proof.
  move=> X Pr MS CM Comp P Q R S D1 D2.
  setoid_rewrite D1; apply/tensor_hom/mon_embed_prop_left => H.
  apply/tensor_hom; rewrite pro_lunit; by apply: D2.
Qed.
Lemma l_mon_prop_loop :
  forall `{LClosedMonSet X, !Bicomplete X} {P Q R : X} {S : Prop},
    (Q ⊢ mon_embed_prop S) -> (S -> P ⊢ R) -> P ⊗ Q ⊢ R.
Proof.
  move=> X Pr MS CM Comp P Q R S D1 D2.
  setoid_rewrite D1; apply/l_tensor_hom/mon_embed_prop_left => H.
  apply/l_tensor_hom; rewrite pro_runit; by apply: D2.
Qed.


(* TODO eoptic instance? *)
Class ProRepresentable {X Y} `{!Proset X, !Proset Y}
      (P : X -> Y -> Prop) `{!Profunctor P} : Prop :=
  represent B : exists P_B, forall A, P A B <-> A ⊢ P_B.
Hint Mode ProRepresentable ! ! - - ! - : typeclass_instances.
Class ProCorepresentable {X Y} `{!Proset X, !Proset Y}
      (P : X -> Y -> Prop) `{!Profunctor P} : Prop :=
  corepresent A : exists P_A, forall B, P A B <-> P_A ⊢ B.
Hint Mode ProCorepresentable ! ! - - ! - : typeclass_instances.
Program Definition pro_psh {X Y} (P : X -> Y -> Prop) `{Profunctor X Y P} (B : Y) : PSh X :=
  fun A => P (get_op A) B.
Next Obligation. move=> X Y P Pr_X Pr_Y Pro B [A] [A'] /= D; apply: lmap D. Qed.
Program Definition pro_copsh {X Y} (P : X -> Y -> Prop) `{Profunctor X Y P} (A : op X)
  : PSh (op Y) :=
  fun B => P (get_op A) (get_op (get_op B)).
Next Obligation. move=> X Y P Pr_X Pr_Y Pro [A] [[B]] [[B']] /= D; apply: rmap D. Qed.
Instance pro_psh_functor `{Profunctor X Y P} : Functor (pro_psh P).
Proof. move=> B B' /= D [A] /=; apply: rmap D. Qed.
Instance pro_copsh_functor `{Profunctor X Y P} : Functor (pro_copsh P).
Proof. move=> [A] [A'] /= D [[B]] /=; apply: lmap D. Qed.
Lemma pro_yoneda : forall `{Profunctor X Y P} {A B},
    P A B <-> yo A ⊢ pro_psh P B.
Proof. move=> *; rewrite yoneda //. Qed.
Lemma pro_yoneda_op : forall `{Profunctor X Y P} {A B},
    P A B <-> yo (Op B) ⊢ pro_copsh P (Op A).
Proof. move=> *; rewrite yoneda //. Qed.
Lemma prorepresentable_alt : forall `{Profunctor X Y P},
    ProRepresentable P <-> forall B, representable (pro_psh P B).
Proof.
  move=> X Y Pr_X Pr_Y P Pro.
  apply: forall_proper => B; apply: exist_proper => P_B; rewrite -Hom_core /=.
  split=> [H [A] | H A] /=; [firstorder | apply: (H (Op A))].
Qed.
Lemma procorepresentable_alt : forall `{Profunctor X Y P},
    ProCorepresentable P <-> forall B, representable (pro_copsh P (Op B)).
Proof.
  move=> X Y Pr_X Pr_Y P Pro.
  apply: forall_proper => A; split=> [[P_A Repr_A] | [[P_A] /Hom_core E]].
  - exists (Op P_A); apply/Hom_core => -[[?]] /=; apply: Repr_A.
  - exists P_A => B; apply: (E (Op (Op B))).
Qed.

Instance hom_representable `{Proset X} : ProRepresentable (pro_le (X:=X)).
Proof. firstorder. Qed.
Instance hom_corepresentable `{Proset X} : ProCorepresentable (pro_le (X:=X)).
Proof. firstorder. Qed.
Instance compose_representable `{ProRepresentable X Y P, Proset Z}
         {Q : Y -> Z -> Prop} `{!Profunctor Q, !ProRepresentable Q}
  : ProRepresentable (P ⦾ Q).
Proof.
  move=> C.
  case: (represent C) => P_C Repr_C; case: (represent P_C) => P_P_C Repr_P_C /=.
  setoid_rewrite Repr_C; setoid_rewrite (prof_core star_right); eauto.
Qed.
Instance compose_corepresentable `{ProCorepresentable X Y P, Proset Z}
         {Q : Y -> Z -> Prop} `{!Profunctor Q, !ProCorepresentable Q}
  : ProCorepresentable (P ⦾ Q).
Proof.
  move=> C.
  case: (corepresent C) => P_C Repr_C; case: (corepresent P_C) => P_P_C Repr_P_C /=.
  setoid_rewrite Repr_C; setoid_rewrite (prof_core costar_left); eauto.
Qed.
Instance star_representable `{Functor Y X F} : ProRepresentable (star F).
Proof. move=> B; rewrite /star; eauto. Qed.
Instance costar_corepresentable `{Functor X Y G} : ProCorepresentable (costar G).
Proof. move=> A; rewrite /costar; eauto. Qed.
Instance starred_representable `{ProRepresentable X Y P} `{Proset Z}
         {F : Z -> Y} `{!Functor F}
  : ProRepresentable (starred P F).
Proof.
  assert (ProRepresentable (P ⦾ star F)) by typeclasses eauto.
    by unfold ProRepresentable; setoid_rewrite <- (prof_core star_right) at 1.
Qed.
Instance costarred_representable `{Proset X} `{ProCorepresentable Y Z P}
         {G : X -> Y} `{!Functor G}
  : ProCorepresentable (costarred G P).
Proof.
  assert (ProCorepresentable (costar G ⦾ P)) by typeclasses eauto.
    by unfold ProCorepresentable; setoid_rewrite <- (prof_core costar_left) at 1.
Qed.

(* TODO Maybe this shouldn't be an 'Instance'---it doesn't seem terribly inferrable. *)
Instance optic_representable1 `{MonSet E, Proset X, Proset Y}
         {L : E -> Endo X} {R : E -> Endo Y}
         `{!Functor L, !Functor R, !LaxMon L, !OplaxMon R} {A B}
         {R_B : Y -> E} `{!Hom_eval_at B ∘ R ⊣ R_B, !Functor R_B}
  : ProRepresentable (opticP1 L R A B).
Proof. move=> T; eexists=> S; by apply optic_concrete. Qed.

(* Coyoneda facts in disguise! *)
Definition extractable `{Bicomplete X, Proset Y} (P : X -> Y -> Prop) `{!Profunctor P} : Prop
  := forall T (A : T -> X) B, (forall y, P (A y) B) -> P (colim A) B.
Definition coextractable `{Proset X, Bicomplete Y} (P : X -> Y -> Prop) `{!Profunctor P} : Prop
  := forall T A (B : T -> Y), (forall y, P A (B y)) -> P A (lim B).

Lemma representable_extractable :
  forall `{Bicomplete X, Proset Y} {P : X -> Y -> Prop} `{!Profunctor P, !ProRepresentable P},
    extractable P.
Proof.
  move=> X Pr_X Comp Y Pr_Y P Pro Rep T A B PAB.
  case: (represent (P:=P) B) => P_B Repr_B.
  apply Repr_B, colim_left => x; apply Repr_B, PAB.
Qed.
Lemma extractable_representable :
  forall `{Bicomplete X, Proset Y} {P : X -> Y -> Prop} `{!Profunctor P},
    extractable P -> ProRepresentable P.
Proof.
  move=> X Pr_X Comp Y Pr_Y P Pro Extr B.
  exists (colim (fun A : sig (pro_psh P B ∘ Op) => `A)) => A; split.
  - move=> PAB; rewrite /compose /=; by apply: (colim_right _ (A ↾ PAB)).
  - move=> D; apply (lmap D), Extr => -[A'] //.
Qed.
Lemma extractable_iff_representable :
  forall `{Bicomplete X, Proset Y} {P : X -> Y -> Prop} `{!Profunctor P},
    extractable P <-> ProRepresentable P.
Proof. move=> *; split; [apply: extractable_representable |
                        apply @representable_extractable]. Qed.

Lemma corepresentable_coextractable :
  forall `{Proset X, Bicomplete Y} {P : X -> Y -> Prop} `{!Profunctor P, !ProCorepresentable P},
    coextractable P.
Proof.
  move=> X Pr_X Y Pr_Y Comp P Pro Corep T A B PAB.
  case: (corepresent (P:=P) A) => P_A Repr_A.
  apply Repr_A, lim_right => x; apply Repr_A, PAB.
Qed.
Lemma coextractable_corepresentable :
  forall `{Proset X, Bicomplete Y} {P : X -> Y -> Prop} `{!Profunctor P},
    coextractable P -> ProCorepresentable P.
Proof.
  move=> X Pr_X Y Pr_Y Comp P Pro Extr A.
  exists (lim (fun B : sig (pro_copsh P (Op A) ∘ Op) => get_op (`B))) => B; split.
  - move=> PAB; rewrite /compose /=; by apply: (lim_left _ (Op B ↾ PAB)).
  - move=> D; apply (rmap D), Extr => -[B'] //.
Qed.
Lemma coextractable_iff_corepresentable :
  forall `{Proset X, Bicomplete Y} {P : X -> Y -> Prop} `{!Profunctor P},
    coextractable P <-> ProCorepresentable P.
Proof. move=> *; split; [apply: coextractable_corepresentable |
                        apply @corepresentable_coextractable]. Qed.

Definition rep_optic {E} `{Proset X, Proset Y} (L : E -> Endo X) (R : E -> Endo Y)
           (S : X) (T : Y) (A : X) (B : Y) : Prop :=
  forall P `(!Profunctor P), Tambara L R P -> ProRepresentable P -> P A B -> P S T.
Definition extend_rep
      `{MonSet E, Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y}
      `{!Functor L, !Functor R, !LaxMon L, !OplaxMon R} {B}
      (R_B : Y -> E) `{!Hom_eval_at B ∘ R ⊣ R_B, !Functor R_B} {S T A}
      (l : rep_optic L R S T A B) : optic L R S T A B
  := l (opticP1 L R A B) _ _ _ optic_id.

Instance PShF_cocont `{Proset X, Proset Y} {F : X -> Y} `{!Functor F}
  : Cocontinuous (PShF F).
Proof.
  apply/distrib_colim_sufficient => R J B /= [[A [r J_r_A]] /= D].
    by exists r, (A ↾ J_r_A).
Qed.
Lemma PShF_id : forall `{Proset X}, PShF (@id X) ⟛ id.
Proof. move=> X ?; apply/pw_core' => F /=; symmetry; apply: coyoneda. Qed.
Lemma PShF_compose : forall `{Proset X, Proset Y, Proset Z}
                       {F : Y -> Z} `{!Functor F} {G : X -> Y},
    PShF (F ∘ G) ⟛ PShF F ∘ PShF G.
Proof.
  move=> X ? Y ? Z ? F ? G; apply/pw_core' => H.
  simpl; unfold PShF at 1 2; rewrite /compose; split.
  - apply/colim_left => -[[A] H_A]; apply/yoneda; simpl.
    unshelve eexists (Op (G A) ↾ _); last by done.
    simpl; by (exists (Op A ↾ H_A)).
  - apply/colim_left => -[[B] [[[A] H_A] D]]; apply/yoneda; simpl in D |- *.
    exists (Op A ↾ H_A); simpl; by apply fmap'.
Qed.
Instance PShF_functor' `{Proset X, Proset Y} : Functor (PShF (X:=X) (Y:=Y)).
Proof. move=> F F' D G [B] /=; by setoid_rewrite D. Qed.
Definition psh_act {E} `{Proset X} (F : E -> Endo X) : E -> Endo (PSh X)
  := (PShF ∘ F) ↑.
Instance psh_act_laxmon `{MonSet E, Proset X} {F : E -> Endo X} `{!Functor F, !LaxMon F}
  : LaxMon (psh_act F).
Proof.
  constructor=> [| M N] /=.
  - setoid_rewrite <- pres_memp; rewrite PShF_id //.
  - setoid_rewrite <- pres_tens; rewrite /pro_tens /= PShF_compose //.
Qed.
Instance psh_act_oplaxmon `{MonSet E, Proset X}
         {F : E -> Endo X} `{!Functor F, !OplaxMon F}
  : OplaxMon (psh_act F).
Proof.
  constructor=> [| M N] /=.
  - setoid_rewrite pres_memp_op; rewrite PShF_id //.
  - setoid_rewrite pres_tens_op; rewrite /pro_tens /= PShF_compose //.
Qed.
Instance psh_act_strongmon `{MonSet E, Proset X}
         {F : E -> Endo X} `{!Functor F, !StrongMon F}
  : StrongMon (psh_act F).
Proof. constructor; typeclasses eauto. Qed.

(* TODO Move these *)
Instance Op_proper : forall `{Proset X}, Proper (pro_le --> pro_le) Op.
Proof. firstorder. Qed.
Instance Op_proper' : forall `{Proset X}, Proper (pro_le ++> flip pro_le) Op.
Proof. firstorder. Qed.
Instance get_op_proper : forall `{Proset X}, Proper (pro_le --> pro_le) get_op.
Proof. firstorder. Qed.
Instance get_op_proper' : forall `{Proset X}, Proper (pro_le ++> flip pro_le) get_op.
Proof. firstorder. Qed.

Lemma tambara_alt :
  forall {E} `{Proset X, Proset Y}
    {L : E -> Endo X} {R : E -> Endo Y}
    {P : X -> Y -> Prop} `{!Profunctor P},
    Tambara L R P <-> LaxEquivariant R (psh_act L) (pro_psh P).
Proof.
  move=> E X ? Y ? L R P ?; split.
  - move=> Tamb M B [A] /= [[[A'] /= /(tamb _ _ M) PA'B] D]; by setoid_rewrite D.
  - move=> LaxACM M A B.
    rewrite pro_yoneda [R in _ -> R]pro_yoneda => /(fmap' (F:=PShF (L M))).
    rewrite PShF_yo => D; setoid_rewrite D; apply: (act_square (Φ:=pro_psh P)).
Qed.


Record irel (X Y : Type) := Irel {ap_irel :> Y -> Y -> X}.
Add Printing Constructor irel.
Arguments Irel {_ _}.
Arguments ap_irel {_ _}.
Definition pro_irel {X} (R : X -> X -> Prop) (Y : Type) : irel X Y -> irel X Y -> Prop :=
  fun R1 R2 => pointwise_relation Y (pointwise_relation Y R) (ap_irel R1) (ap_irel R2).
Arguments pro_irel {X} R Y !a1 !a2 /.
Instance pro_irel_pro `{PreOrder X R} {Y} : PreOrder (pro_irel R Y).
Proof.
  constructor; first by firstorder.
  move=> [x] [y] [z] /=; etransitivity; eauto.
Qed.
Instance irel_proset `{Proset X} {Y} : Proset (irel X Y)
  := {| pro_le := pro_irel pro_le Y |}.
Lemma irel_core : forall `{Proset X} {Y} {R S : irel X Y},
    R ⟛ S <-> (forall rho rho', R rho rho' ⟛ S rho rho').
Proof. firstorder. Qed.
Instance ap_irel_functor `{Proset X} {Y} : Functor (ap_irel (X:=X) (Y:=Y)).
Proof. move=> [R] [S] //. Qed.
Instance ap_irel_proper `{Proset X} {Y}
  : Proper (pro_le ++> (=) ==> (=) ==> pro_le) (ap_irel (X:=X) (Y:=Y)).
Proof. move=> [R] [R'] /= D ? _ <- ? _ <-; apply: D. Qed.
Instance ap_irel_proper' `{Proset X} {Y}
  : Proper (pro_le --> (=) ==> (=) ==> flip pro_le) (ap_irel (X:=X) (Y:=Y)).
Proof. move=> [R] [R'] /= D ? _ <- ? _ <-; apply: D. Qed.
Instance Irel_functor `{Proset X} {Y} : Functor (Irel (X:=X) (Y:=Y)).
Proof. move=> R S //. Qed.
Instance Irel_proper `{Proset X} {Y}
  : Proper (pro_le ++> (=) ==> (=) ==> pro_le) (Irel (X:=X) (Y:=Y)).
Proof. move=> R R' /= D ? _ <- ? _ <-; apply: D. Qed.
Instance Irel_proper' `{Proset X} {Y}
  : Proper (pro_le --> (=) ==> (=) ==> flip pro_le) (Irel (X:=X) (Y:=Y)).
Proof. move=> R R' /= D ? _ <- ? _ <-; apply: D. Qed.
Instance irel_adjunction1 `{Proset X} {Y} : Irel (X:=X) (Y:=Y) ⊣ ap_irel.
Proof. constructor=> [R | [R]] //=. Qed.
Instance irel_adjunction2 `{Proset X} {Y} : ap_irel (X:=X) (Y:=Y) ⊣ Irel.
Proof. constructor=> [[R] | R] //=. Qed.
(*
Instance irel_core' `{Proset X} {Y}
: subrelation (core (pro_le (X:=Y -> X))) (pointwise_relation Y (core (pro_le (X:=X)))).
Proof. firstorder. Qed.
 *)
Definition tens_irel `{MonSet X, !Bicomplete X} (Y : Type) (R1 R2 : irel X Y)
  : irel X Y :=
  Irel (fun rho rho'' => colim (fun rho' : Y => R1 rho rho' ⊗ R2 rho' rho'')).
Arguments tens_irel {X _ _ _} Y !R1 !R2 /.
Instance irel_monoidal `{Quantale X} {Y}
  : Monoidal (pro_le (X:=irel X Y)) (tens_irel Y)
             (Irel (fun a b => mon_embed_prop (a = b))).
Proof.
  constructor.
  - move=> [R] [R'] [S] [S'] /= D1 D2 rho rho''.
    apply: fmap' => rho'; apply/bimap; firstorder.
  - move=> [R]; apply/irel_core=> rho rho' /=; split.
    + apply: colim_left => rho''.
      apply/tensor_hom/mon_embed_prop_left => <-.
      apply/tensor_hom; rewrite pro_lunit //.
    + setoid_rewrite <- colim_right.
      setoid_rewrite <- (mon_embed_prop_right erefl (reflexivity _)).
        by setoid_rewrite <- (proj2 (pro_lunit _)).
  - move=> [R]; apply/irel_core=> rho rho' /=; split.
    + apply: colim_left=> rho''.
      apply/l_tensor_hom/mon_embed_prop_left => <-.
      apply/l_tensor_hom; rewrite pro_runit //.
    + setoid_rewrite <- colim_right.
      setoid_rewrite <- (mon_embed_prop_right erefl (reflexivity _)).
        by setoid_rewrite <- (proj2 (pro_runit _)).
  - move=> [R] [S] [T]; apply/irel_core=> rho0 rho /=; split.
    + apply: colim_left=> rho1; setoid_rewrite distrib_colim; apply: colim_left=> rho2 /=.
      rewrite pro_massoc; setoid_rewrite <- colim_right; apply: bimap; last by done.
        by setoid_rewrite <- colim_right.
    + apply: colim_left=> rho2; rewrite -/(r_tensor _ _).
      setoid_rewrite (distrib_colim (F:=r_tensor _)).
      simpl; apply: colim_left=> rho1 /=; rewrite -pro_massoc.
      setoid_rewrite <- colim_right; apply: bimap; first by done.
        by setoid_rewrite <- colim_right.
Qed.
Instance irel_mset `{Quantale X} {Y}
  : MonSet (irel X Y)
  := {| pro_tens := tens_irel Y |}.
Program Instance irel_bicomplete `{Bicomplete X} {Y} : Bicomplete (irel X Y)
  := {| lim R J := Irel (lim (ap_irel ∘ J)); colim R J := Irel (colim (ap_irel ∘ J)) |}.
Next Obligation.
  move=> X Pr_X Comp Y R J; rewrite /lim /=; split=> [r | A' Cone] rho rho' /=.
  - apply: lim_left.
  - apply: lim_right => r; apply: Cone.
Qed.
Next Obligation.
  move=> X Pr_X Comp Y R J; rewrite /colim /=; split=> [r | A' Cocone] rho rho' /=.
  - rewrite -/ap_Hom; by setoid_rewrite <- colim_right.
  - apply: colim_left => r; apply: Cocone.
Qed.
(* TODO Maybe define the actual internal homs directly? *)
Instance irel_tensor_cocont `{Quantale X} {Y} {R : irel X Y}
  : Cocontinuous (pro_tens R).
Proof.
  apply/distrib_colim_sufficient => R' J rho rho'' /=.
  apply/colim_left => rho'; rewrite distrib_colim; apply/colim_left => r /=.
  setoid_rewrite <- colim_right; by setoid_rewrite <- colim_right.
Qed.
Instance irel_r_tensor_cocont `{Quantale X} {Y} {R : irel X Y}
  : Cocontinuous (r_tensor R).
Proof.
  apply/distrib_colim_sufficient => R' J rho rho'' /=.
  apply/colim_left => rho'; setoid_rewrite (distrib_colim (F:=r_tensor _)).
  apply/colim_left => r /=; setoid_rewrite <- colim_right.
    by setoid_rewrite <- colim_right.
Qed.
Instance irel_cmset `{Quantale X} {Y} : ClosedMonSet (irel X Y)
  := r_tensor_cocontinuous_sufficient _.
Instance irel_lcmset `{Quantale X} {Y} : LClosedMonSet (irel X Y)
  := tensor_cocontinuous_sufficient _.
Instance irel_quantale `{Quantale X} {Y} : Quantale (irel X Y).

(* TODO "basechange" is a misnomer *)
Definition irel_basechange {X X' Y} (f : X -> X') : irel X Y -> irel X' Y
  := Irel ∘ compose (compose f) ∘ ap_irel.
(* TODO This should be inferrable. *)
Instance irel_basechange_full `{Proset X, Proset X'} {F : X -> X'} `{!Full F} {Y}
  : Full (irel_basechange (Y:=Y) F).
Proof. move=> R S /= D rho rho'; apply/(unfmap (F:=F))/D. Qed.
Instance irel_basechange_laxmon `{Quantale X, Quantale X'}
         {F : X -> X'} `{!Functor F, !LaxMon F} {Y}
  : LaxMon (irel_basechange (Y:=Y) F).
Proof.
  constructor=> [| [R] [S]] rho rho' //=.
  - apply/colim_left=> <- //.
    setoid_rewrite (pres_memp (F:=F)); apply/fmap'; by setoid_rewrite <- colim_right.
  - apply/colim_left=> rho''.
    setoid_rewrite pres_tens; apply/fmap'; by setoid_rewrite <- colim_right.
Qed.
(* TODO Figure out whether cocontinuity of F is actually necessary. *)
Instance irel_basechange_oplaxmon `{Quantale X, Quantale X'}
         {F : X -> X'} `{!Functor F, !OplaxMon F, !Cocontinuous F} {Y}
  : OplaxMon (irel_basechange (Y:=Y) F).
Proof.
  constructor=> [| [R] [S]] rho rho'' //=.
  - rewrite distrib_colim; apply: fmap' => rho' /=; apply: pres_memp_op.
  - rewrite distrib_colim; apply: fmap' => rho' /=; apply: pres_tens_op.
Qed.
Instance irel_basechange_strongmon `{Quantale X, Quantale X'}
         {F : X -> X'} `{!Functor F, !StrongMon F, !Cocontinuous F} {Y}
  : StrongMon (irel_basechange (Y:=Y) F).
Proof. constructor; typeclasses eauto. Qed.

Definition rel_act `{MonSet X, !Bicomplete X} (Y : Type) (R : irel X Y)
  : Endo (Y -> X)
  := (fun Q rho => colim (fun rho' : Y => R rho rho' ⊗ Q rho')) as_a Endo (Y -> X).
Instance rel_act_functor `{MonSet X, !Bicomplete X} {Y}
  : Functor (rel_act (X:=X) Y).
Proof.
  move=> A B D P /= rho.
  apply: fmap' => rho'; apply/bimap; firstorder.
Qed.
Instance rel_act_strongmon `{Quantale X} {Y}
  : StrongMon (rel_act (X:=X) Y).
Proof.
  constructor; constructor.
  - move=> P rho /=; setoid_rewrite <- colim_right.
    setoid_rewrite <- mon_embed_prop_right; only 2-3: reflexivity.
      by setoid_rewrite <- (proj2 (pro_lunit _)).
  - move=> [R] [S] P rho /=.
    apply: colim_left => ?; setoid_rewrite distrib_colim; apply: colim_left => ? /=.
    setoid_rewrite <- colim_right; rewrite -[colim _ ⊗ _]/(r_tensor _ _).
    setoid_rewrite <- (proj2 (distrib_colim (F:=r_tensor _) _)).
    setoid_rewrite <- colim_right; rewrite pro_massoc //=.
  - move=> P rho /=; apply: colim_left => rho'.
    apply/(mon_prop_loop (reflexivity _)) => <- //.
  - move=> [R] [S] P rho /=.
    apply: colim_left => ?; rewrite -[colim _ ⊗ _]/(r_tensor _ _).
    rewrite (distrib_colim (F:=r_tensor _)); apply: colim_left => ? /=.
    setoid_rewrite <- colim_right; setoid_rewrite <- colim_right; rewrite -pro_massoc //=.
Qed.

(* TODO Move these *)
Instance internal_hom_proper `{ClosedMonSet X}
  : Proper (pro_le --> pro_le ++> pro_le) (internal_hom (X:=X)).
Proof.
  move=> A A' /= D1 B B' D2; rewrite -tensor_hom.
  setoid_rewrite D1; setoid_rewrite D2; rewrite tensor_hom //.
Qed.
Instance internal_hom_proper' `{ClosedMonSet X}
  : Proper (pro_le ++> pro_le --> flip pro_le) (internal_hom (X:=X)).
Proof. move=> ? ? ? ? ? ? /=; by apply: internal_hom_proper. Qed.
Instance l_internal_hom_proper `{LClosedMonSet X}
  : Proper (pro_le --> pro_le ++> pro_le) (l_internal_hom (X:=X)).
Proof.
  move=> A A' /= D1 B B' D2; rewrite -l_tensor_hom.
  setoid_rewrite D1; setoid_rewrite D2; rewrite l_tensor_hom //.
Qed.
Instance l_internal_hom_proper' `{LClosedMonSet X}
  : Proper (pro_le ++> pro_le --> flip pro_le) (l_internal_hom (X:=X)).
Proof. move=> ? ? ? ? ? ? /=; by apply: l_internal_hom_proper. Qed.

Definition rel_act_op `{ClosedMonSet X, !Bicomplete X} (Y : Type) (R : op (irel X Y))
  : Endo (Y -> X)
  := (fun Q rho => lim (fun rho' : Y => get_op R rho rho' ⊸ Q rho')) as_a Endo (Y -> X).
Instance rel_act_op_functor `{ClosedMonSet X, !Bicomplete X} {Y}
  : Functor (rel_act_op (X:=X) Y).
Proof. move=> [A] [B] /= D P rho; apply: fmap' => rho'; by setoid_rewrite D. Qed.
Instance rel_act_op_strongmon `{Quantale X} {Y}
  : StrongMon (rel_act_op (X:=X) Y).
Proof.
  constructor; constructor=> [| [R] [S]] P rho /=.
  - apply/lim_right => rho'; apply/tensor_hom/l_mon_prop_loop => // <- //.
  - apply/lim_right => rho''; apply/tensor_hom/l_tensor_hom/colim_left => rho'.
    do 2 setoid_rewrite lim_left.
    rewrite -l_tensor_hom pro_massoc; by do 2 setoid_rewrite mon_modus_ponens.
  - setoid_rewrite lim_left; rewrite -[R in R ⊢ _]pro_runit.
    setoid_rewrite (mon_embed_prop_right erefl (reflexivity _)).
    apply/mon_modus_ponens.
  - apply/lim_right => rho'; apply/tensor_hom/lim_right => rho''.
    setoid_rewrite lim_left; setoid_rewrite <- colim_right.
    rewrite -tensor_hom -pro_massoc; by setoid_rewrite mon_modus_ponens.
Qed.


(* TODO Put this in the right place. *)
Instance pw_closed_monoidal {Y} `{ClosedMonoidal X R T H} :
  ClosedMonoidal (pointwise_relation Y R) (fun f g y => T (f y) (g y))
                 (fun f g y => H (f y) (g y)).
Proof. firstorder. Qed.
Instance pw_l_closed_monoidal {Y} `{LClosedMonoidal X R T H} :
  LClosedMonoidal (pointwise_relation Y R) (fun f g y => T (f y) (g y))
                  (fun f g y => H (f y) (g y)).
Proof. firstorder. Qed.
Instance pw_cmset {Y} `{ClosedMonSet X} : ClosedMonSet (Y -> X) :=
  {| internal_hom := fun f g y => internal_hom (f y) (g y) |}.
Instance pw_lcmset {Y} `{LClosedMonSet X} : LClosedMonSet (Y -> X) :=
  {| l_internal_hom := fun f g y => l_internal_hom (f y) (g y) |}.
Instance pw_quantale `{Quantale X} {Y} : Quantale (Y -> X).

Example rel_act_optic : forall `{Quantale X} {R} {S T A B : R -> X},
    optic (rel_act R) (rel_act R) S T A B <->
    forall rho : R, S rho ⊢ colim (fun rho' : R => (B rho' ⊸ T rho) ⊗ A rho').
Proof.
  move=> X Pr Comp MS CMS LCMS Quant R S T A B.
  pose proof (_ : compose (colim ∘ r_tensor B) ∘ ap_irel ⊣ _) as Adj.
  rewrite optic_concrete //.
Qed.

(* TODO Move these to the right place. *)
Instance flip_functor {Y1 Y2} `{Proset X} : Functor (@flip Y1 Y2 X).
Proof. firstorder. Qed.
Instance flip_strongmon {Y1 Y2} `{MonSet X} : StrongMon (@flip Y1 Y2 X).
Proof. constructor; by constructor. Qed.
Instance flip_adjunction {Y1 Y2} `{Proset X} : @flip Y1 Y2 X ⊣ flip.
Proof. done. Qed.
(* TODO This could stand to compile faster; can I optimize at all? *)
Example rel_act_optic' :
  forall `{Quantale X} {R Y} {S : R -> X} {T : Y -> R -> X} {A : R -> X} {B : Y -> R -> X},
    optic (rel_act R) (conjugate (rel_act R) flip flip ∘ irel_basechange const)
          S T A B <->
    forall rho : R, S rho ⊢ colim (fun rho' : R => lim (fun y => B y rho' ⊸ T y rho) ⊗ A rho').
Proof.
  move=> X Pr Comp MS CMS LCMS Quant R Y S T A B.
  pose proof (_ : flip ∘ compose (colim ∘ flip ∘ r_tensor B ∘ flip) ∘
                       ap_irel ∘ irel_basechange const ⊣ _) as Adj.
  rewrite optic_concrete //.
Qed.

(* TODO Do we want to support alternate monoidal structures? *)
(* TODO Fiddle with the constraints. *)
(*
Class InternalReflexive `{MonSet X} {Y} (R : Y -> Y -> X) : Prop :=
  internal_reflexivity : forall rho, memp ⊢ R rho rho.
Hint Mode InternalReflexive ! - - ! ! : typeclass_instances.
Class InternalTransitive `{MonSet X} {Y} (R : Y -> Y -> X) : Prop :=
  internal_transitivity : forall rho1 rho2 rho3, R rho1 rho2 ⊗ R rho2 rho3 ⊢ R rho1 rho3.
Hint Mode InternalTransitive ! - - ! ! : typeclass_instances.
Class InternalPreOrder `{MonSet X} {Y} (R : Y -> Y -> X) : Prop :=
  {IPrO_IReflexive :> InternalReflexive R;
   IPrO_ITransitive :> InternalTransitive R}.
Hint Mode InternalPreOrder ! - - ! ! : typeclass_instances.
*)
Lemma monoid_ipro : forall `{Quantale X} {Y} {R : Y -> Y -> X},
    Monoid (Irel R) <-> (forall rho, memp ⊢ R rho rho) /\
                      (forall rho1 rho2 rho3, R rho1 rho2 ⊗ R rho2 rho3 ⊢ R rho1 rho3).
Proof.
  move=> X ? ? ? ? ? ? Y R; split=> [Mon | [IRefl ITrans]]; split.
  - move=> rho.
      by etransitivity; [apply: mon_embed_prop_right | apply: (eta (M:=Irel R))].
  - move=> rho1 rho2 rho3.
    etransitivity; last by apply: (mu (M:=Irel R)).
      by setoid_rewrite <- colim_right.
  - move=> rho rho' /=; apply: mon_embed_prop_left => <- //.
  - move=> rho1 rho3 /=; apply: colim_left => rho2 //.
Qed.
Definition mon_embed_rel (X : Type) `{MonSet X, !Bicomplete X} {Y} (R : Y -> Y -> Prop)
  : Y -> Y -> X
  := fun a b => mon_embed_prop (R a b).
Arguments mon_embed_rel _ {_ _ _ _} _ _ _ /.
Definition prop_irel (X : Type) `{MonSet X, !Bicomplete X} {Y} : (Y -> Y -> Prop) -> irel X Y
  := Irel ∘ mon_embed_rel X.
Arguments prop_irel _ {_ _ _ _} _ /.
Instance prop_irel_monoid `{Quantale X, PreOrder Y R}
  : Monoid (prop_irel X R).
Proof.
  constructor.
  - move=> rho rho' /=; apply/fmap' => <- //.
  - move=> rho1 rho3 /=; apply/colim_left => rho2.
    rewrite mon_tensor_prop; apply: fmap' => -[? ?]; by etransitivity.
Qed.

Definition box `{Quantale X} {Y} (Reach : irel X Y) : (Y -> X) -> Y -> X
  := rel_act_op _ (Op Reach).
Arguments box {_ _ _ _ _ _ _ _} Reach P /.
(* TODO Move this *)
Instance op_comonoid `{Monoid X M} : Comonoid (Op M).
Proof. firstorder. Qed.
Instance op_monoid `{Comonoid X W} : Monoid (Op W).
Proof. firstorder. Qed.
(* box is a comonad if Reach is a preorder *)
Example box_comonad `{Quantale X} {Y} {Reach : irel X Y}
         `{!Monoid Reach} : Comonad (box Reach).
Proof. typeclasses eauto. Qed.

(* Coalgebras of box *)
Notation Hereditary Reach := (Coalg (box Reach)).
Definition hpred `{Quantale X} {Y} (Reach : irel X Y) : Type := coEM (box Reach).
Lemma Hereditary_prop_irel : forall `{Quantale X} {Y} {Reach : Y -> Y -> Prop} {P : Y -> X},
    Hereditary (prop_irel X Reach) P <->
    forall {rho rho'}, Reach rho rho' -> P rho ⊢ P rho'.
Proof.
  move=> X Pr Comp MS CMS LCMS Quant Y Reach P; split.
  - move=> Hered rho rho' ReachStep.
    setoid_rewrite (Hered rho); setoid_rewrite lim_left; rewrite -[_ ⊸ _]pro_runit /=.
    setoid_rewrite (mon_embed_prop_right ReachStep (reflexivity _)).
    apply/mon_modus_ponens.
  - move=> Hered' rho; apply/lim_right => rho'.
    apply/tensor_hom/l_tensor_hom/mon_embed_prop_left => H.
    apply/l_tensor_hom; rewrite pro_runit; by apply/Hered'.
Qed.
(* TODO Move this *)
Instance mon_embed_prop_comonoid `{MonSet X, !Bicomplete X} {P : Prop}
  : Comonoid (mon_embed_prop P).
Proof.
  constructor.
  - apply: mon_embed_prop_left => //.
  - apply: mon_embed_prop_left => PP.
    setoid_rewrite <- mon_embed_prop_right; try first [assumption | reflexivity].
    rewrite pro_runit //.
Qed.
(* Can we loosen the sym constraint at all? *)
Instance Hereditary_monclosed `{Quantale X, !SymMonSet X} {Y} {Reach : irel X Y}
         `{!forall rho1 rho2, Comonoid (Reach rho1 rho2)}
  : MonClosed (Hereditary Reach).
Proof.
  constructor.
  - move=> rho /=; apply/lim_right => rho'.
    setoid_rewrite epsilon; rewrite -tensor_hom pro_lunit //.
  - move=> P Q Hered1 Hered2 rho /=; apply/lim_right => rho'.
    rewrite /pro_tens /=; setoid_rewrite (Hered1 rho); setoid_rewrite (Hered2 rho).
    simpl; setoid_rewrite (lim_left _ rho').
    rewrite -tensor_hom -pro_massoc; setoid_rewrite delta at 3.
    setoid_rewrite pro_massoc at 2; setoid_rewrite mon_modus_ponens.
    rewrite [Q rho' ⊗ _]pro_sym' pro_massoc; by setoid_rewrite mon_modus_ponens.
Qed.
(* In any case, it's still true for prop_irel even if the quantale is non-commutative. *)
Instance Hereditary_prop_irel_monclosed `{Quantale X} {Y} {Reach : Y -> Y -> Prop}
  : MonClosed (Hereditary (prop_irel X Reach)).
Proof.
  constructor.
  - by apply/Hereditary_prop_irel.
  - move=> P Q /Hereditary_prop_irel Hered1 /Hereditary_prop_irel Hered2.
    apply/Hereditary_prop_irel => rho rho' ReachStep; apply: bimap; auto.
Qed.

Instance mon_embed_prop_hereditary `{Quantale X} {Y} {Reach : irel X Y}
         `{!forall rho1 rho2, Comonoid (Reach rho1 rho2)} {P : Prop}
  : Hereditary Reach (mon_embed_prop P).
Proof.
  move=> rho; apply/lim_right => rho' /=; apply/tensor_hom/mon_prop_loop; first by compute.
  move=> PP; apply/mon_embed_prop_right => //; apply/epsilon.
Qed.
Instance embed_prop_hereditary `{Quantale X} {Y} {Reach : irel X Y} {P : Prop}
  : Hereditary Reach (embed_prop P).
Proof.
  move=> *; apply/lim_right => rho' /=; apply: colim_left => ? /=; apply/tensor_hom.
  setoid_rewrite <- colim_right; [| done]; apply/lim_right => -[].
Qed.

Definition dia `{MonSet X, !Bicomplete X} {Y} (Reach : irel X Y) : (Y -> X) -> Y -> X
  := rel_act _ Reach.
Arguments dia {_ _ _ _ _} _ _ /.
(* box is a monad if Reach is a preorder *)
Example dia_monad `{Quantale X} {Y} {Reach : irel X Y}
        `{!Monoid Reach} : Monad (dia Reach).
Proof. typeclasses eauto. Qed.

(* TODO *Should* I move this? *)
(* TODO Set up "flipped monoidal product". *)
Definition irel_flip {X Y} : irel X Y -> irel X Y := Irel ∘ flip ∘ ap_irel.
Instance flip_monoid `{Quantale X, !SymMonSet X} {Y} {R : irel X Y} `{Mon : !Monoid R}
  : Monoid (irel_flip R).
Proof.
  move: R Mon => [R] /monoid_ipro [IRefl ITrans]; apply/monoid_ipro; split.
  - move=> rho //=.
  - move=> rho1 rho2 rho3 /=; rewrite pro_sym' //.
Qed.
Instance flip_comonoid `{Quantale X, !SymMonSet X} {Y} {R : irel X Y}
         `{Comon : !Comonoid R}
  : Comonoid (irel_flip R).
Proof.
  constructor.
  - move=> rho rho'; setoid_rewrite (epsilon (W:=R)); simpl; apply/fmap' => ? //.
  - move=> rho rho'' /=; setoid_rewrite (delta (W:=R)) at 1; simpl; apply/fmap' => rho'.
    apply: pro_sym.
Qed.
Example dia_monad_flip `{Quantale X, !SymMonSet X} {Y} {Reach : irel X Y}
        `{!Monoid Reach} : Monad (dia (irel_flip Reach)).
Proof. typeclasses eauto. Qed.
(* TODO Reformulate this in terms of flipped monoidal product. *)
Instance dia_box_adjoint `{Quantale X, !SymMonSet X} {Y} {Reach : irel X Y} :
  dia (irel_flip Reach) ⊣ box Reach.
Proof.
  apply transpose_sufficient => P Q /=; rewrite /pro_le /= /pointwise_relation.
  setoid_rewrite (transpose (Adj:=colim_const_adjoint)).
  setoid_rewrite <- (transpose (Adj:=const_lim_adjoint)).
  rewrite /pro_le /= /pointwise_relation /=.
  setoid_rewrite <- (transpose (Adj:=r_tensor_internal_hom_adjoint)); simpl.
  setoid_rewrite pro_sym' at 1; firstorder.
Qed.

(* Algebas of dia *)
(* ...are the same as coalgebras of box, because we have an adjunction!
      So there is an adjoint triple. *)
Definition free_hpred `{Quantale X, !SymMonSet X} {Y} {Reach : irel X Y} `{!Monoid Reach}
  : (Y -> X) -> hpred Reach
  := free_coEM _.
Definition cofree_hpred `{Quantale X} {Y} {Reach : irel X Y} `{!Monoid Reach}
  : (Y -> X) -> hpred Reach
  := cofree_coEM _.
Arguments free_hpred {_ _ _ _ _ _ _ _ _} Reach {_}.
Arguments cofree_hpred {_ _ _ _ _ _ _ _} Reach {_}.

