Require Import VST.floyd.proofauto.
From Coq Require Import ssreflect ssrfun ssrbool.
Require Import stdpp.tactics.

Require Import Optics.

Open Scope logic.

Local Hint Mode NatDed ! : typeclass_instances.

(* flipped instances due to https://gitlab.mpi-sws.org/iris/stdpp/issues/58 *)
Local Instance andp_proper `{NatDed X} : Proper (derives ++> derives ++> derives) andp.
Proof. move=> ? ? ? ? ? ?; by apply: andp_derives. Qed.
Local Instance orp_proper `{NatDed X} : Proper (derives ++> derives ++> derives) orp.
Proof. move=> ? ? ? ? ? ?; by apply: orp_derives. Qed.
Local Instance exp_proper `{NatDed X} {Y} :
  Proper (pointwise_relation Y derives ++> derives) exp.
Proof. move=> ? ? ?; by apply: exp_derives. Qed.
Local Instance allp_proper `{NatDed X} {Y} :
  Proper (pointwise_relation Y derives ++> derives) allp.
Proof. move=> ? ? ?; by apply: allp_derives. Qed.
Local Instance imp_proper `{NatDed X} : Proper (derives --> derives ++> derives) imp.
Proof. move=> ? ? ? ? ? ?; by apply: imp_derives. Qed.
Local Instance prop_proper `{NatDed X} : Proper (impl ++> derives) prop.
Proof. move=> ? ? ?; by apply: prop_derives. Qed.
Local Instance sepcon_proper `{SepLog X} :
  Proper (derives ++> derives ++> derives) sepcon.
Proof. move=> ? ? ? ? ? ?; by apply: sepcon_derives. Qed.
Local Instance wand_proper `{SepLog X} : Proper (derives --> derives ++> derives) wand.
Proof. move=> ? ? ? ? ? ?; by apply: wand_derives. Qed.
(* not sure we have anything comparable for ewand? *)
Local Instance andp_proper' `{NatDed X}
  : Proper (derives --> derives --> flip derives) andp.
Proof. move=> ? ? ? ? ? ?; by apply: andp_derives. Qed.
Local Instance orp_proper' `{NatDed X}
  : Proper (derives --> derives --> flip derives) orp.
Proof. move=> ? ? ? ? ? ?; by apply: orp_derives. Qed.
Local Instance exp_proper' `{NatDed X} {Y} :
  Proper (pointwise_relation Y derives --> flip derives) exp.
Proof. move=> ? ? ?; by apply: exp_derives. Qed.
Local Instance allp_proper' `{NatDed X} {Y} :
  Proper (pointwise_relation Y derives --> flip derives) allp.
Proof. move=> ? ? ?; by apply: allp_derives. Qed.
Local Instance imp_proper' `{NatDed X}
  : Proper (derives ++> derives --> flip derives) imp.
Proof. move=> ? ? ? ? ? ?; by apply: imp_derives. Qed.
Local Instance prop_proper' `{NatDed X} : Proper (impl --> flip derives) prop.
Proof. move=> ? ? ?; by apply: prop_derives. Qed.
Local Instance sepcon_proper' `{SepLog X} :
  Proper (derives --> derives --> flip derives) sepcon.
Proof. move=> ? ? ? ? ? ?; by apply: sepcon_derives. Qed.
Local Instance wand_proper' `{SepLog X}
  : Proper (derives ++> derives --> flip derives) wand.
Proof. move=> ? ? ? ? ? ?; by apply: wand_derives. Qed.

Instance derives_pro `{NatDed X} : PreOrder (derives (A:=X))
  := {| PreOrder_Reflexive := derives_refl; PreOrder_Transitive := derives_trans |}.
Instance natded_proset `{NatDed X} : Proset X | 2 := {| pro_le := derives |}.
Instance natded_antisymmetric `{NatDed X} : Antisymmetric X (=) derives := pred_ext.
Instance andp_monoidal `{NatDed X}
  : Monoidal derives (andp (A:=X)) TT.
Proof.
  constructor; [move=> *; by apply andp_derives | | |].
  all: move=> *; rewrite ?TT_andp ?andp_TT ?andp_assoc //.
Qed.
Instance andp_sym `{NatDed X} : Sym derives (andp (A:=X)).
Proof. move=> A B; rewrite andp_comm //. Qed.
Instance sepcon_monoidal `{ClassicalSep X}
  : Monoidal derives (sepcon (A:=X)) emp.
Proof.
  constructor; [apply sepcon_derives | | |].
  all: move=> *; rewrite ?emp_sepcon ?sepcon_emp ?sepcon_assoc //.
Qed.
Instance sepcon_sym `{ClassicalSep X} : Sym derives (sepcon (A:=X)).
Proof. move=> A B; rewrite sepcon_comm //. Qed.
Instance classicalsep_mset `{ClassicalSep X} : MonSet X | 2
  := {| pro_tens := sepcon |}.
Instance classicalsep_smset `{ClassicalSep X} : SymMonSet X | 2 := {}.
Fact lift_natded_instance_defeq_coherence : forall `{NatDed X} {Y},
    natded_proset (X:=Y -> X) = @pw_proset Y X natded_proset.
Proof. reflexivity. Qed.
Program Instance natded_bicomplete `{NatDed X} : Bicomplete X | 2 :=
  {| lim R J := @allp X _ R J; colim R J := exp J |}.
Next Obligation.
  move=> X ND R J; split=> [r | A' C] /=.
  - by apply: allp_left.
  - by apply: allp_right.
Qed.
Next Obligation.
  move=> X ND R J; split=> [r | A' C] /=.
  - by apply: exp_right.
  - by apply: exp_left.
Qed.
Instance classicalsep_monoidal `{ClassicalSep X}
  : ClosedMonoidal pro_le (sepcon (A:=X)) wand.
Proof. move=> P Q R; rewrite /pro_le /= wand_sepcon_adjoint //. Qed.
Instance seplog_cmset `{ClassicalSep X} : ClosedMonSet X := {}.
Instance classicalsep_quantale `{ClassicalSep X} : Quantale X.


Definition diff_supp {A} (S : ident -> Prop) (rho rho' : Map.t A) : Prop
  := forall i, S i \/ Map.get rho i = Map.get rho' i.
Inductive te_diff_supp (S : ident -> Prop) : environ -> environ -> Prop :=
| LocalsChanged : forall {ge ve te te'},
    diff_supp S te te' -> te_diff_supp S (mkEnviron ge ve te) (mkEnviron ge ve te').
Hint Constructors te_diff_supp.
Instance diff_supp_pro {A} {S : ident -> Prop} : Equivalence (@diff_supp A S).
Proof.
  constructor.
  - compute; auto.
  - move=> rho rho' DS i; case: (DS i); auto.
  - move=> rho1 rho2 rho3 DS DS' i.
    move: (DS i) (DS' i) => [Si | ->] [Si' | <-] //; auto.
Qed.
Instance te_diff_supp_pro {S : ident -> Prop} : Equivalence (te_diff_supp S).
Proof.
  constructor.
  - move=> []; by constructor.
  - move=> _ _ [ge ve te te'] //.
  - move=> _ _ rho [ge ve te te'] DS; inversion_clear 1.
    constructor; by etransitivity.
Qed.
Lemma te_diff_supp_spec : forall {S P},
    closed_wrt_vars S P <-> Hereditary (prop_irel mpred (te_diff_supp S)) P.
Proof.
  move=> S P; split.
  - move=> CWV.
    apply Hereditary_prop_irel => _ _ [ge ve te te'] /(CWV (mkEnviron _ _ _) te') -> //.
  - move=> /Hereditary_prop_irel Hered [ge ve te] /= te' DS.
    apply pred_ext; by apply Hered.
Qed.
Definition unchanged_modvars (c : statement) : irel mpred environ :=
  prop_irel mpred (te_diff_supp (modifiedvars c)).
Notation closed_wrt_modvars' c := (Hereditary (unchanged_modvars c)).
Corollary closed_wrt_modvars_alt : forall {c P},
    closed_wrt_modvars c P <-> closed_wrt_modvars' c P.
Proof. move=> *; apply te_diff_supp_spec. Qed.

Definition frameable (c : statement) : Type := hpred (unchanged_modvars c).

(* TODO Maybe I want this to be using Delta somehow? *)
Definition semaxP {CS : compspecs} {Espec : OracleKind}
           (Delta : tycontext) (c : statement)
  : (environ -> mpred) -> (environ -> mpred) -> Prop :=
  fun P Q => semax Delta P c (normal_ret_assert Q).

Instance semax_profunctor {CS Espec Delta c} :
  Profunctor (@semaxP CS Espec Delta c).
Proof.
  move=> P' P Q Q' H1 H2 Step.
  apply/semax_pre_post'/Step; by apply/derives_ENTAIL.
Qed.

Instance semax_tambara_frameable {CS Espec Delta c} :
  Tambara (E:=frameable c) sig_lreg sig_lreg (@semaxP CS Espec Delta c).
Proof.
  move=> [F CWM] P Q Step; change (semaxP Delta c (F * P) (F * Q)).
  apply closed_wrt_modvars_alt in CWM.
  rewrite /semaxP 2![F * _]sepcon_comm -frame_normal.
    by apply/semax_frame.
Qed.
Instance semax_tambara_lreg_F {CS Espec Delta c} `{Proset X} {F : X -> environ -> mpred}
         `{!Functor F} {H : forall A, closed_wrt_modvars' c (F A)}
  : Tambara (lreg ∘ F) (lreg ∘ F) (@semaxP CS Espec Delta c).
Proof.
  change (lreg ∘ F) with (sig_lreg ∘ F ↣ frameable c).
  typeclasses eauto.
Qed.

Instance semax_representable {CS Espec Delta c}
  : ProRepresentable (@semaxP CS Espec Delta c).
Proof. apply extractable_representable => T A B; apply semax_extract_exists. Qed.


Example ramif_optic : forall `{ClassicalSep X} {S T A B : X},
    optic lreg lreg S T A B <-> S |-- A * (B -* T).
Proof.
  move=> X ND SL CL S T A B.
  rewrite optic_concrete /= sepcon_comm //.
Qed.

Example sig_ramif_optic : forall {c} {S T A B : environ -> mpred},
    optic (E:=frameable c) sig_lreg sig_lreg S T A B <->
    S |-- A * box (unchanged_modvars c) (B -* T).
Proof.
  move=> c S T A B.
  pose proof (_ : r_tensor B ∘ (sval : frameable c -> _) ⊣ _) as Adj.
  rewrite optic_concrete sepcon_comm //.
Qed.

Example ramifQ_optic : forall `{ClassicalSep X} {R} {S : X} {T : R -> X} {A : X} {B : R -> X},
    optic (lreg ∘ allp) lreg S T A B <-> S |-- A * allp (B -* T).
Proof.
  move=> X ND SL CL R S T A B.
  change (@allp _ _ _) with (@lim X _ _ R).
  rewrite optic_concrete /= sepcon_comm //.
Qed.
(* TODO Write a tactic for using optic_concrete. *)
Example ramifQ_optic' : forall `{ClassicalSep X} {R} {S : X} {T : R -> X} {A : X} {B : R -> X},
    optic lreg (lreg ∘ const) S T A B <-> S |-- A * allp (B -* T).
Proof.
  move=> X ND SL CL R S T A B.
  pose proof (_ : r_tensor B ∘ const ⊣ _) as Adj.
  rewrite optic_concrete /= sepcon_comm //.
Qed.

Example ramifQ_optic'' : forall `{ClassicalSep X} {R} {S T A B : R -> X},
    optic (lreg ∘ const) (lreg ∘ const) S T A B <-> S |-- A * const (allp (B -* T)).
Proof.
  move=> X ND SL CL R S T A B.
  pose proof (_ : r_tensor B ∘ const ⊣ _) as Adj.
  rewrite optic_concrete sepcon_comm //.
Qed.

Example ramifQ_optic2 : forall `{ClassicalSep X} {R} {S : R -> X} {T : X} {A : R -> X} {B : X},
    optic lreg (lreg ∘ exp) S T A B <-> S |-- A * const (B -* T).
Proof.
  move=> X ND SL CL R S T A B.
  change (@exp _ _ _) with (@colim X _ _ R).
  pose proof (_ : r_tensor B ∘ colim (R:=R) ⊣ _) as Adj.
  rewrite optic_concrete sepcon_comm //.
Qed.

Section Localization.
  Context {CS : compspecs} {Espec : OracleKind} {Delta : tycontext}
          {Y : Type} (c : statement).
  Definition localP := starred (semaxP Delta c) (colim (R:=Y)).

  Definition empred : Type := environ -> mpred.
  Let L : frameable c -> Endo empred := sig_lreg.
  Let R : frameable c -> Endo (Y -> empred) := lreg ∘ const ∘ sval.
  (* conjugate sig_lreg lim const. *)
  Definition localization : empred -> (Y -> empred) -> empred -> (Y -> empred) -> Prop
    := optic L R.

  Let L' : empred -> Endo empred := lreg ∘ box (unchanged_modvars c).
  Let R' : empred -> Endo (Y -> empred) := lreg ∘ const ∘ dia (unchanged_modvars c).
  Definition localization' : empred -> (Y -> empred) -> empred -> (Y -> empred) -> Prop
    := optic L' R'.

  (* (* wtf universe error?!?!? *)
  Lemma add_prime : forall {S T A B},
      localization S T A B -> localization' S T A B.
  Proof.
    rewrite /localization' -[L']/(L ∘ cofree_coEM _) -[R']/(R ∘ free_coEM _).
    move=> *; by apply: emap_adj.
  Qed.
   *)

  Lemma localization_intro : forall {S T A B M},
      closed_wrt_modvars c M ->
      S |-- A * M -> M |-- allp (B -* T) ->
      localization S T A B.
  Proof.
    move=> S T A B M /closed_wrt_modvars_alt CWM Open Close.
    apply eoptic_to_optic; exists (M ↾ CWM).
    - by rewrite sepcon_comm in Open.
    - simpl; apply/(wand_sepcon_adjoint _ _ T) => rho.
      setoid_rewrite Close; by apply/allp_left.
  Qed.

  Instance localP_tambara : Tambara L R localP.
  Proof.
    unshelve eapply starred_tambara; try typeclasses eauto.
    2: apply: semax_tambara_frameable.
    move=> [M ?] P rho; rewrite /= /pro_le /pro_tens /= /pro_tens /= exp_sepcon2 //.
  Qed.
  Instance localP_tambara' : Tambara L' R' localP.
  Proof.
    unshelve eapply starred_tambara; try typeclasses eauto.
    2: apply: semax_tambara_lreg_F.
    move=> M P rho; rewrite /= /pro_le /pro_tens /= /pro_tens /= /pro_tens /=.
    Intros y; Exists y rho; cancel; allp_left rho.
    Exists (reflexivity (R:=te_diff_supp (modifiedvars c)) rho).
    rewrite emp_sepcon -[L in L |-- _]emp_sepcon; apply modus_ponens_wand'.
      by Exists (reflexivity (R:=te_diff_supp (modifiedvars c)) rho).
  Qed.

  Lemma localization_elim : forall {S T A B},
      localization S T A B ->
      semax Delta A c (normal_ret_assert (exp B)) ->
      semax Delta S c (normal_ret_assert (exp T)).
  Proof.
    move=> S T A B.
    change (semax Delta ?W c (normal_ret_assert (exp ?Z))) with (localP W Z).
    apply.
  Qed.

  Instance localization_representable {A B} : ProRepresentable (opticP1 L R A B).
  Proof.
    pose proof (_ : r_tensor B ∘ const ∘ proj1_sig (P:=closed_wrt_modvars' c) ⊣ _) as Adj.
    rewrite -[L in L ⊣ _]/(Hom_eval_at B ∘ R) in Adj.
    typeclasses eauto.
  Qed.

  Fact localization_extract_exists : forall {V} {S T A B},
      (forall x : V, localization (S x) T A B) -> localization (exp S) T A B.
  Proof. intros *; apply: (representable_extractable (P:=opticP1 L R _ _)). Qed.
End Localization.

Local Instance exp_proper'' `{NatDed X} {Y} :
  Proper (pointwise_relation Y (=) ==> (=)) exp.
Proof. move=> ? ? ?; by apply: exp_congr. Qed.

Section Framing.
  Context {R Y : Type}.

  Definition mframing (S T A B : R -> mpred) : Prop
    := optic (rel_act R) (rel_act R) S T A B.
  Definition framing
             (S : R -> mpred) (T : Y -> R -> mpred) (A : R -> mpred) (B : Y -> R -> mpred) : Prop
    := optic (rel_act R) (conjugate (rel_act R) flip flip ∘ irel_basechange const)
             S T A B.
  Fact mframing_concrete : forall {S T A B},
      mframing S T A B <->
      forall rho : R, S rho |-- EX rho' : R, A rho' * (B rho' -* T rho).
  Proof.
    move=> *; split; setoid_rewrite sepcon_comm => ?;
      by apply/(rel_act_optic (X:=mpred)).
  Qed.
  Fact framing_concrete : forall {S T A B},
      framing S T A B <->
      forall rho : R, S rho |-- EX rho' : R, A rho' * (ALL y : Y, B y rho' -* T y rho).
  Proof.
    move=> *; split; setoid_rewrite sepcon_comm => ?;
      by apply/(rel_act_optic' (X:=mpred)).
  Qed.

  Definition opticP3 {E} `{Proset X, Proset Y} (L : E -> Endo X) (R : E -> Endo Y)
             (S A : X) (B T : Y) : Prop := optic L R S T A B.
  Instance optic_profunctor3 {E} `{Proset X, Proset Y} {L : E -> Endo X} {R : E -> Endo Y}
           {S A} : Profunctor (opticP3 L R S A).
  Proof.
    move=> B' B T T' D1 D2 l P Pro Tamb /(rmap D1) PAB.
    apply: (rmap D2); by apply: l.
  Qed.
  (* hrmmmmm *)
  Instance optic_tambara3 `{SymMonSet E, Proset X, Proset Y}
           {L : E -> Endo X} {R' : E -> Endo Y} `{!Functor R', !StrongMon R'}
           {S A} : Tambara R' R' (opticP3 L R' S A).
  Proof.
    move=> M B T l P Pro Tamb; apply (l (starred P (R' M))).
    - typeclasses eauto.
    - move=> M' A' B'; rewrite /starred => /(tamb _ _ M'); apply: rmap.
      enough (R' M' ⊗ R' M ⊢ R' M ⊗ R' M') by firstorder.
      setoid_rewrite pres_tens at 1.
      setoid_rewrite <- pres_tens_op.
      rewrite pro_sym' //.
  Qed.

  (* hrmmmmmmmmmmmm *)
  Definition framingP3 (S A : R -> mpred) (B T : Y -> R -> mpred) : Prop
    := framing S T A B.
  Definition precomp_act `{Proset X} {Y} : (Y -> Y) -> Endo (Y -> X)
    := flip compose ↑.
  Instance framing_tambara3_precomp2 {S A}
    : Tambara precomp_act precomp_act (framingP3 S A).
  Proof.
    move=> M B T; rewrite /framingP3 2!framing_concrete => l r /=.
    setoid_rewrite l; apply/exp_derives => ?; apply/sepcon_derives => //.
    apply/allp_right => ?; by apply/allp_left.
  Qed.
End Framing.
(* ...but what if we want to change Y? guess maybe we *do* need 2-actegories!?!?! *)
Lemma refine_framing : forall {R Y Y'} {S T A B} (f : Y' -> Y) (l : @framing R Y S T A B),
    @framing R Y' S (T ∘ f) A (B ∘ f).
Proof.
  move=> R Y Y' S T A B f; rewrite !framing_concrete => l r /=.
  setoid_rewrite l; apply/exp_derives => ?; apply/sepcon_derives => //.
  apply/allp_right => ?; by apply/allp_left.
Qed.
Notation "l1 ∘∘∘ l2" := (refine_framing _ l1 ∘∘ l2) (at level 40).



(* Worked example *)
Require tree.
Module TreeExample.
  Import tree.
  Local Instance CompSpecs : compspecs. make_compspecs prog. Defined.
  Definition Vprog : varspecs. mk_varspecs prog. Defined.

  Definition NVal : type := Tstruct _NodeVal noattr.
  Definition Node : type := tptr NVal.

  Inductive tree := leaf | branch (v : Z) (l r : tree).
  Fixpoint repr_tree (t : tree) (p : val) : mpred :=
    match t with
    | leaf => !! (p = nullval)
    | branch v l r => EX p_l : val, EX p_r : val,
      data_at Ews NVal (Vint (Int.repr v), (p_l, p_r)) p *
      repr_tree l p_l * repr_tree r p_r
    end.

  Lemma left_branch : forall {v : Z} {l r : tree},
      framing (repr_tree (branch v l r)) (fun l' => repr_tree (branch v l' r))
              (repr_tree l)              repr_tree.
  Proof.
    move=> v l r /=; apply/framing_concrete => p.
    apply/exp_left => p_l; apply/exp_left => p_r; apply/(exp_right p_l); cancel.
    apply/allp_right => l'; rewrite -wand_sepcon_adjoint.
    apply/exp_right/exp_right; cancel.
  Qed.
  Lemma right_branch : forall {v : Z} {l r : tree},
      framing (repr_tree (branch v l r)) (fun r' => repr_tree (branch v l r'))
              (repr_tree r)              repr_tree.
  Proof.
    move=> v l r /=; apply/framing_concrete => p.
    apply/exp_left => p_l; apply/exp_left => p_r; apply/(exp_right p_r); cancel.
    apply/allp_right => r'; rewrite -wand_sepcon_adjoint.
    apply/exp_right/exp_right; cancel.
  Qed.

  Inductive path_to (v : Z) : tree -> Type :=
  | here : forall l r, path_to v (branch v l r)
  | in_left  : forall v' l r, path_to v l -> path_to v (branch v' l r)
  | in_right : forall v' l r, path_to v r -> path_to v (branch v' l r).

  Fixpoint replace (v : Z) (t : tree) (c : path_to v t) (s : tree) : tree :=
    match c with
    | here l r => s
    | in_left v' l r c' => branch v' (replace v l c' s) r
    | in_right v' l r c' => branch v' l (replace v r c' s)
    end.

  Lemma path_to_framing : forall {v t} (c : path_to v t),
      framing (repr_tree t) (fun s => repr_tree (replace v t c s))
              (EX l : _, EX r : _, repr_tree (branch v l r)) repr_tree.
  Proof.
    move=> v t; elim: t /.
    - move=> l r; apply: smap optic_id => p; by apply/exp_right/exp_right.
    - move=> v' l r c IH; refine (left_branch ∘∘∘ IH).
    - move=> v' l r c IH; refine (right_branch ∘∘∘ IH).
  Qed.
End TreeExample.
