From Coq Require Import ssreflect ssrfun ssrbool.
Require Export stdpp.list.
Require Export stdpp.relations.


(* why aren't these in the stdlib... *)
Instance: Proper (pointwise_relation A impl ==> pointwise_relation A impl ==> impl) (ex2 (A:=A)). Proof. firstorder. Qed.
Instance: Proper (pointwise_relation A (flip impl) ==> pointwise_relation A (flip impl) ==> (flip impl)) (ex2 (A:=A)). Proof. firstorder. Qed.
Instance: Proper (pointwise_relation A iff ==> pointwise_relation A iff ==> iff) (ex2 (A:=A)). Proof. firstorder. Qed.
Instance: Proper (impl ==> impl ==> impl ==> impl) and3. Proof. firstorder. Qed.
Instance: Proper (flip impl ==> flip impl ==> flip impl ==> flip impl) and3. Proof. firstorder. Qed.
Instance: Proper (iff ==> iff ==> iff ==> iff) and3. Proof. firstorder. Qed.
Instance: Proper (impl ==> impl ==> impl ==> impl) or3. Proof. firstorder. Qed.
Instance: Proper (flip impl ==> flip impl ==> flip impl ==> flip impl) or3. Proof. firstorder. Qed.
Instance: Proper (iff ==> iff ==> iff ==> iff) or3. Proof. firstorder. Qed.


(* TODO WHY AREN'T MY INSTANCES WORKING *)
Section GeneralRels.
  (* someday we'll probably need relations between different types... *)
  Context {A : Type}.

  Definition rel_compose (R S : relation A) : relation A :=
    fun x z => exists2 y, R x y & S y z.
  Infix ">>" := rel_compose (at level 60, right associativity).
  Infix "≡ᵣ" := relation_equivalence (at level 70, no associativity).
  Infix "⊆ᵣ" := subrelation (at level 70, no associativity).
  Infix "∪ᵣ" := relation_disjunction (at level 50).
  Notation "(>>)" := rel_compose (only parsing).
  Notation "(≡ᵣ)" := relation_equivalence (only parsing).
  Notation "(⊆ᵣ)" := subrelation (only parsing).
  Notation "(∪ᵣ)" := relation_disjunction (only parsing).

  Definition flips {R : relation A} {x y} (Step : R x y) : flip R y x := Step.
  Definition step_compose {R S : relation A} {x y z} (Step : R x y) (Step' : S y z)
    : (R >> S) x z := ex_intro2 _ _ _ Step Step'.
  Infix ">>>" := step_compose (at level 60, right associativity).
  Definition rel_inl {x y} {R : relation A} (S : relation A) (Step : R x y)
    : (R ∪ᵣ S) x y := or_introl Step.
  Definition rel_inr {x y} (R : relation A) {S : relation A} (Step : S x y)
    : (R ∪ᵣ S) x y := or_intror Step.

  Lemma flip_flip : forall {R : relation A}, flip (flip R) ≡ᵣ R.
  Proof. firstorder. Qed.
  Lemma flip_inj : forall {R S : relation A}, flip R ≡ᵣ flip S <-> R ≡ᵣ S.
  Proof. firstorder. Qed.
  Lemma flip_compose : forall {R S}, flip (R >> S) ≡ᵣ flip S >> flip R.
  Proof. firstorder. Qed.
  Lemma flip_compare : forall {R S : relation A}, R ⊆ᵣ S <-> flip R ⊆ᵣ flip S.
  Proof. firstorder. Qed.
  Global Instance flip_proper : Proper ((≡ᵣ) ==> (≡ᵣ)) (@flip A A Prop).
  Proof. firstorder. Qed.
  Global Instance rel_compose_proper : Proper ((≡ᵣ) ==> (≡ᵣ) ==> (≡ᵣ)) (>>).
  Proof.
    move=> R R' E_R S S' E_S x z; split.
    - move=> [y /E_R H /E_S H']; firstorder.
    - move=> [y /E_R H /E_S H']; firstorder.
  Qed.
  Global Instance rel_compose_mono1 : Proper ((⊆ᵣ) ==> (≡ᵣ) ==> (⊆ᵣ)) (>>).
  Proof. firstorder. Qed.
  Global Instance rel_compose_mono2 : Proper ((≡ᵣ) ==> (⊆ᵣ) ==> (⊆ᵣ)) (>>).
  Proof. firstorder. Qed.

  Global Instance subrelation_proper : Proper ((≡ᵣ) ==> (≡ᵣ) ==> iff) (@subrelation _A).
  Proof.
    move=> _A R R' E_R S S' E_S; split;
            move=> Sub x y /E_R /Sub /E_S //.
  Qed.

  Global Instance Ur_proper : Proper ((≡ᵣ) ==> (≡ᵣ) ==> (≡ᵣ)) (@relation_disjunction _A).
  Proof. firstorder. Qed.
  Global Instance Ur_mono1 : Proper ((⊆ᵣ) ==> (≡ᵣ) ==> (⊆ᵣ)) (@relation_disjunction _A).
  Proof. firstorder. Qed.
  Global Instance Ur_mono2 : Proper ((≡ᵣ) ==> (⊆ᵣ) ==> (⊆ᵣ)) (@relation_disjunction _A).
  Proof. firstorder. Qed.
  Lemma Ur_comm : forall {R S : relation A}, R ∪ᵣ S ≡ᵣ S ∪ᵣ R.
  Proof. firstorder. Qed.

  (* figure out something about the relationship between [sub]commute and Proper *)
  Definition commute : relation (relation A) := fun R S => R >> S ≡ᵣ S >> R.
  Definition subcommute : relation (relation A) := fun R S => R >> S ⊆ᵣ S >> R.

  Global Instance commute_proper : Proper ((≡ᵣ) ==> (≡ᵣ) ==> iff) commute.
  Proof. move=> R R' E_R S S' E_S; rewrite /commute E_R E_S //. Qed.
  Global Instance subcommute_proper : Proper ((≡ᵣ) ==> (≡ᵣ) ==> iff) subcommute.
  Proof. move=> R R' E_R S S' E_S; rewrite /subcommute E_R E_S //. Qed.

  Global Instance commute_refl : Reflexive commute.
  Proof. firstorder. Qed.
  Global Instance commute_sym : Symmetric commute.
  Proof. move=> *. rewrite /commute. by symmetry. Qed.
  Lemma subcommute2 : forall {R S}, commute R S <-> subcommute R S /\ subcommute S R.
  Proof. firstorder. Qed.
  Lemma subcommute_weaken : forall {R S}, commute R S -> subcommute R S.
  Proof. firstorder. Qed.
  Lemma subcommute_flip : forall {R S}, subcommute R S <-> subcommute (flip S) (flip R).
  Proof. move=> R S. rewrite /subcommute -2!flip_compose -flip_compare //. Qed.
  Lemma commute_flip : forall {R S}, commute R S <-> commute (flip R) (flip S).
  Proof. move=> R S. rewrite /commute -2!flip_compose flip_inj //. Qed.

  Definition simulation (T T' R : relation A) : Prop :=
    forall x y, R x y -> forall x', T x x' -> exists2 y', R x' y' & T' y y'.
  Definition bisimulation (T T' R : relation A) : Prop :=
    simulation T T' R /\ simulation T T' (flip R).

  Lemma commut_alt : forall {R S}, commut _ R S <-> subcommute S R.
  Proof.
    move=> R S; split.
    - move=> Commut x z [y H_S H_R].
      move: (Commut _ _ H_R _ H_S); firstorder.
    - move=> Sub z y H_R x H_S.
      case: (Sub x z); firstorder.
  Qed.

  Lemma simulation_alt : forall {T T' R}, simulation T T' R <-> flip R >> T ⊆ᵣ T' >> flip R.
  Proof.
    move=> T T' R; split.
    - move=> Sim x z [y /= H H'].
      move: (Sim _ _ H _ H') => [y' K K']; firstorder.
    - move=> Sub x y E x' Step.
      move: (Sub _ _ (flips E >>> Step)) => [y' H H'].
      firstorder.
  Qed.

  Lemma po_squish : forall {R : relation A} `{!Reflexive R, !Transitive R},
      R >> R ≡ᵣ R.
  Proof. firstorder. Qed.

  Global Instance rtc_mono : Proper ((⊆ᵣ) ==> (⊆ᵣ)) (@rtc A).
  Proof. move=> R R' Sub x y; by apply: rtc_subrel. Qed.
  Global Instance rtc_proper : Proper ((≡ᵣ) ==> (≡ᵣ)) (@rtc A).
  Proof.
    move=> R R' /PartialOrder_inverse [/= Sub1 Sub2].
    apply antisymmetry; apply: rtc_mono.
  Qed.

  Lemma rtc_universal : forall {R S : relation A} `{!Reflexive S, !Transitive S},
      R ⊆ᵣ S -> rtc R ⊆ᵣ S.
  Proof.
    move=> R S Refl Trans Sub x y.
    elim: x y /.
    - done.
    - by etransitivity; first apply Sub.
  Qed.

  Lemma rtc_redundant : forall {R : relation A} `{!Reflexive R, !Transitive R},
      rtc R ≡ᵣ R.
  Proof.
    move=> R Refl Trans.
    apply antisymmetry.
    - apply: rtc_universal.
    - apply: rtc_once.
  Qed.

  Lemma rtc_unfold : forall {R}, rtc R ≡ᵣ (=) ∪ᵣ (R >> rtc R).
  Proof.
    move=> R x z; split.
    - case: x z / => *.
      + by left.
      + by right; econstructor.
    - move=> [-> | [y Step Steps]]; [done | by econstructor].
  Qed.

  Lemma rtc_flip : forall {R : relation A}, rtc (flip R) ≡ᵣ flip (rtc R).
  Proof. move=> R x y; split; elim; firstorder; by apply: rtc_r. Qed.

  Lemma rtc_subcommute1 : forall {R S}, subcommute R S -> subcommute (rtc R) S.
  Proof.
    move=> R S SubComm x z [y RSteps SStep].
    elim/@rtc_ind_r: y / RSteps z SStep; first by eexists.
    move=> _y y RSteps RStep IH z SStep.
    move: (RStep >>> SStep) => /SubComm [y' /IH [x' SStep' RSteps'] RStep'].
    eexists; [done | by apply: rtc_r].
  Qed.
  Lemma rtc_subcommute2 : forall {R S}, subcommute R S -> subcommute R (rtc S).
  Proof.
    move=> R S /subcommute_flip /rtc_subcommute1.
    rewrite rtc_flip -subcommute_flip //.
  Qed.
  Lemma rtc_commute1 : forall {R S}, commute R S -> commute (rtc R) S.
  Proof.
    move=> R S /subcommute2 [/rtc_subcommute1 SubComm1 /rtc_subcommute2 SubComm2].
    apply/subcommute2; split=> //.
  Qed.
  Lemma rtc_commute2 : forall {R S}, commute R S -> commute R (rtc S).
  Proof.
    move=> R S /subcommute2 [/rtc_subcommute2 SubComm1 /rtc_subcommute1 SubComm2].
    apply/subcommute2; split=> //.
  Qed.

  (* TODO Can I get this to accomodate the use case with uv and ur bends? *)
  Lemma rtc_ell : forall {R S},
      subcommute R S -> rtc (R ∪ᵣ S) ≡ᵣ rtc S >> rtc R.
  Proof.
    move=> R S SubComm x z; split.
    - elim: x z /; first by eexists.
      move=> x x' z [RStep | SStep] _ [y SSteps RSteps].
      + move: (RStep >>> SSteps) => /(rtc_subcommute2 SubComm) [_y SSteps' RStep'].
        eexists; [done | by econstructor].
      + eexists; [by econstructor | done].
    - move=> [y RSteps SSteps].
      transitivity y.
      + apply: (rtc_subrel S); firstorder.
      + apply: (rtc_subrel R); firstorder.
  Qed.
End GeneralRels.
Infix ">>" := rel_compose (at level 60).
Infix "≡ᵣ" := relation_equivalence (at level 70, no associativity).
Infix "⊆ᵣ" := subrelation (at level 70, no associativity).
Infix "∪ᵣ" := relation_disjunction (at level 50).
Notation "(>>)" := rel_compose (only parsing).
Notation "(≡ᵣ)" := relation_equivalence (only parsing).
Notation "(⊆ᵣ)" := subrelation (only parsing).
Notation "(∪ᵣ)" := relation_disjunction (only parsing).
Infix ">>>" := step_compose (at level 60).

Instance rel_compose_assoc : Assoc (@relation_equivalence A) (>>).
Proof. firstorder. Qed.
Instance rel_compose_left_id : LeftId (@relation_equivalence A) (=) (>>).
Proof. firstorder. Qed.
Instance rel_compose_right_id : RightId (@relation_equivalence A) (=) (>>).
Proof. firstorder. Qed.
Existing Instance rel_compose_proper.
Existing Instance rel_compose_mono1.
Existing Instance rel_compose_mono2.


Lemma Forall_middle : forall {A} {P : A -> Prop} {l1 x l2},
    Forall P (l1 ++ [x] ++ l2) -> P x.
Proof. move=> A P l1 x l2 /Forall_app [_ /Forall_cons [//]]. Qed.
Lemma Forall2_middle : forall {A} {R : A -> A -> Prop} {l1 x y l2},
    Forall2 R (l1 ++ [x] ++ l2) (l1 ++ [y] ++ l2) -> R x y.
Proof.
  move=> A R l1 x y l2 /(Forall2_app_inv _ _ _ _ _ erefl)
          [_ /(Forall2_cons_inv _ _ _ _) [//]].
Qed.

Section ListRels.
  Context {A : Type}.

  (* element-element relation *)
  Definition ee_relation := A -> A -> Prop.
  (* element-list relation *)
  Definition el_relation := A -> list A -> Prop.
  (* list-element relation
  Definition le_relation := list A -> A -> Prop. *)
  (* list-list relation *)
  Definition ll_relation := list A -> list A -> Prop.

  Context {R_ee : ee_relation} {R_el : el_relation} {R_ll : ll_relation}.

  (* TODO these can probably be unified *)
  (* TODO Proper instances for the type constructors *)
  Inductive ee_contextual : ll_relation :=
  | ee_Contextual : forall l1 x y l2,
      R_ee x y -> ee_contextual (l1 ++ [x] ++ l2) (l1 ++ [y] ++ l2).
  Inductive el_contextual : ll_relation :=
  | el_Contextual : forall l1 x l l2,
      R_el x l -> el_contextual (l1 ++ [x] ++ l2) (l1 ++ l ++ l2).
  Inductive ll_contextual : ll_relation :=
  | ll_Contextual : forall l1 l l' l2,
      R_ll l l' -> ll_contextual (l1 ++ l ++ l2) (l1 ++ l' ++ l2).
  Hint Constructors ee_contextual el_contextual ll_contextual.

  Definition ee_beginning : forall x y l2, R_ee x y -> ee_contextual (x :: l2) (y :: l2)
    := ee_Contextual [].
  Definition el_beginning : forall x l l2, R_el x l -> el_contextual (x :: l2) (l ++ l2)
    := el_Contextual [].
  Definition ll_beginning : forall l l' l2, R_ll l l' -> ll_contextual (l ++ l2) (l' ++ l2)
    := ll_Contextual [].
  Lemma ee_end : forall l1 x y, R_ee x y -> ee_contextual (l1 ++ [x]) (l1 ++ [y]).
  Proof. by constructor. Qed.
  Lemma el_end : forall l1 x l, R_el x l -> el_contextual (l1 ++ [x]) (l1 ++ l).
  Proof. move=> *. rewrite -[R in el_contextual _ R]app_nil_r -app_assoc //. Qed.
  Lemma ll_end : forall l1 l l', R_ll l l' -> ll_contextual (l1 ++ l) (l1 ++ l').
  Proof.
    move=> *.
    rewrite -[L in ll_contextual L _]app_nil_r -[R in ll_contextual _ R]app_nil_r
            -!app_assoc //.
  Qed.
  Hint Resolve ee_beginning el_beginning ll_beginning ee_end el_end ll_end.

  Lemma ee_inv :
    ee_contextual ≡ᵣ fun k k' => exists l1 x y l2,
                         k = l1 ++ [x] ++ l2 /\ k' = l1 ++ [y] ++ l2 /\ R_ee x y.
  Proof.
    move=> k k'; split.
    - case: k k' / => *.
      repeat eexists; eauto.
    - firstorder.
  Qed.
  Lemma el_inv :
    el_contextual ≡ᵣ fun k k' => exists l1 x l l2,
                         k = l1 ++ [x] ++ l2 /\ k' = l1 ++ l ++ l2 /\ R_el x l.
  Proof.
    move=> k k'; split.
    - case: k k' / => *.
      repeat eexists; eauto.
    - firstorder.
  Qed.
  Lemma ll_inv :
    ll_contextual ≡ᵣ fun k k' => exists l1 l l' l2,
                         k = l1 ++ l ++ l2 /\ k' = l1 ++ l' ++ l2 /\ R_ll l l'.
  Proof.
    move=> k k'; split.
    - case: k k' / => *.
      repeat eexists; eauto.
    - firstorder.
  Qed.

  Lemma ee_nil1 : forall {h}, ee_contextual [] h <-> False.
  Proof.
    move=> h; split; last by done.
    move=> /ee_inv [l1 [x [y [l2 [/(app_cons_not_nil _ _ _) //]]]]].
  Qed.
  Lemma ee_nil2 : forall {h}, ee_contextual h [] <-> False.
  Proof.
    move=> h; split; last by done.
    move=> /ee_inv [l1 [x [y [l2 [_ [/(app_cons_not_nil _ _ _) //]]]]]].
  Qed.
  Lemma el_nil : forall {h}, el_contextual [] h <-> False.
  Proof.
    move=> h; split; last by done.
    move=> /el_inv [l1 [x [l [l2 [/(app_cons_not_nil _ _ _) //]]]]].
  Qed.
  Hint Resolve ee_nil1 ee_nil2 el_nil.

  Lemma ee_singleton1 : forall {x h}, ee_contextual [x] h <-> exists2 y, h = [y] & R_ee x y.
  Proof.
    move=> x h; split.
    - move=> /ee_inv [l1 [x' [y [l2 [E [-> Step]]]]]].
      move: l1 l2 E Step => [| ? [| ? ?]] [| ? ?] // [<-]; eauto.
    - move=> [y -> Step]; by apply: ee_beginning.
  Qed.
  Lemma ee_singleton2 : forall {h y}, ee_contextual h [y] <-> exists2 x, h = [x] & R_ee x y.
  Proof.
    move=> h y; split.
    - move=> /ee_inv [l1 [x [y' [l2 [-> [E Step]]]]]].
      move: l1 l2 E Step => [| ? [| ? ?]] [| ? ?] // [<-]; eauto.
    - move=> [x -> Step]; by apply: ee_beginning.
  Qed.
  Lemma el_singleton : forall {x h}, el_contextual [x] h <-> R_el x h.
  Proof.
    move=> x h; split.
    - move=> /el_inv [l1 [x' [y [l2 [E [-> Step]]]]]].
      move: l1 l2 E Step => [| ? [| ? ?]] [| ? ?] // [<-].
      rewrite app_nil_r //.
    - rewrite -{2}[h]app_nil_r.
      apply: el_beginning.
  Qed.
  Hint Resolve el_singleton.

  Global Instance cons_ee1 : Proper (R_ee ==> (=) ==> ee_contextual) cons.
  Proof. move=> x y Step l _ <-. apply/ee_beginning/Step. Qed.
  Global Instance cons_ee2 : Proper ((=) ==> ee_contextual ==> ee_contextual) cons.
  Proof. move=> x _ <- _ _ [l1 x' y l2 Step]. rewrite !app_comm_cons //. Qed.
  Lemma cons_el1 : forall {x l}, R_el x l -> forall {k}, el_contextual (x :: k) (l ++ k).
  Proof. move=> x l Step k. apply/el_beginning/Step. Qed.
  Global Instance cons_el2 : Proper ((=) ==> el_contextual ==> el_contextual) cons.
  Proof. move=> x _ <- _ _ [l1 x' y l2 Step]. rewrite !app_comm_cons //. Qed.
  Global Instance cons_ll : Proper ((=) ==> ll_contextual ==> ll_contextual) cons.
  Proof. move=> x _ <- _ _ [l1 x' y l2 Step]. rewrite !app_comm_cons //. Qed.
  Global Instance app_ee1 : Proper (ee_contextual ==> (=) ==> ee_contextual) app.
  Proof. move=> _ _ [l1 x y l2 Step] k2 _ <-. rewrite -!app_assoc //. Qed.
  Global Instance app_ee2 : Proper ((=) ==> ee_contextual ==> ee_contextual) app.
  Proof. move=> k1 _ <- _ _ [l1 x y l2 Step]. rewrite ![k1 ++ l1 ++ _]app_assoc //. Qed.
  Global Instance app_el1 : Proper (el_contextual ==> (=) ==> el_contextual) app.
  Proof. move=> _ _ [l1 x y l2 Step] k2 _ <-. rewrite -!app_assoc //. Qed.
  Global Instance app_el2 : Proper ((=) ==> el_contextual ==> el_contextual) app.
  Proof. move=> k1 _ <- _ _ [l1 x l l2 Step]. rewrite ![k1 ++ l1 ++ _]app_assoc //. Qed.
  Global Instance app_ll1 : Proper (ll_contextual ==> (=) ==> ll_contextual) app.
  Proof. move=> _ _ [l1 l l' l2 Step] k2 _ <-. rewrite -!app_assoc //. Qed.
  Global Instance app_ll2 : Proper ((=) ==> ll_contextual ==> ll_contextual) app.
  Proof. move=> k1 _ <- _ _ [l1 l l' l2 Step]. rewrite ![k1 ++ l1 ++ _]app_assoc //. Qed.

  Definition ee_contextuals : relation (list A) := rtc ee_contextual.
  Definition el_contextuals : relation (list A) := rtc el_contextual.
  Definition ll_contextuals : relation (list A) := rtc ll_contextual.

  Lemma ees_nil1 : forall {h}, ee_contextuals [] h <-> h = [].
  Proof.
    move=> h; split; last by move=> ->.
    move=> /(rtc_inv _ _) [] // [? [/ee_nil1 //]].
  Qed.
  Lemma ees_nil2 : forall {h}, ee_contextuals h [] <-> h = [].
  Proof.
    move=> h; split; last by move=> ->.
    move=> /(rtc_inv_r _ _) [] // [? [_ /ee_nil2 //]].
  Qed.
  Lemma els_nil : forall {h}, el_contextuals [] h <-> h = [].
  Proof.
    move=> h; split; last by move=> ->.
    move=> /(rtc_inv _ _) [] // [? [/el_nil //]].
  Qed.

  (*
  Lemma ees_singleton1 : forall {x h},
      ee_contextuals [x] h <-> exists2 y, h = [y] & rtc R_ee x y.
  Proof.
    move=> x h; split.
    - move=> /ee_inv [l1 [x' [y [l2 [Step [E ->]]]]]].
      move: l1 l2 E Step => [| ? [| ? ?]] [| ? ?] // [<-]; eauto.
    - move=> [y -> Step]; by apply: ee_beginning.
  Qed.
  Lemma ee_singleton2 : forall {h y},
      ee_contextuals h [y] <-> exists2 x, h = [x] & rtc R_ee x y.
  Proof.
    move=> h y; split.
    - move=> /ee_inv [l1 [x [y' [l2 [Step [-> E]]]]]].
      move: l1 l2 E Step => [| ? [| ? ?]] [| ? ?] // [<-]; eauto.
    - move=> [x -> Step]; by apply: ee_beginning.
  Qed.
   *)

  Global Instance cons_ees :
    Proper (rtc R_ee ==> ee_contextuals ==> ee_contextuals) cons.
  Proof.
    move=> x y.
    elim: x y /.
    - move=> x l l' CSteps.
      elim: l l' / CSteps => // l l' l'' CStep CSteps.
        by apply/rtc_l/cons_ee2.
    - move=> x x' y Step _ IH l l' CSteps.
        by econstructor; [apply: cons_ee1 | apply: IH].
  Qed.
  (* Hmm... *)
  Global Instance cons_els : Proper ((=) ==> el_contextuals ==> el_contextuals) cons.
  Proof.
    move=> x _ <- l l' CSteps.
    elim: l l' / CSteps => // l l' l'' CStep CSteps.
      by apply/rtc_l/cons_el2.
  Qed.
  Global Instance cons_lls : Proper ((=) ==> ll_contextuals ==> ll_contextuals) cons.
  Proof.
    move=> x _ <- l l' CSteps.
    elim: l l' / CSteps => // l l' l'' CStep CSteps.
      by apply/rtc_l/cons_ll.
  Qed.
  Global Instance app_ees :
    Proper (ee_contextuals ==> ee_contextuals ==> ee_contextuals) app.
  Proof.
    move=> k1 k1'; elim: k1 k1' /.
    - move=> k1 k2 k2'; elim: k2 k2' /.
      + reflexivity.
      + move=> k2 k2' k2'' CStep CSteps.
          by apply/rtc_l/app_ee2.
    - move=> k1 k1' k1'' CStep CSteps IH k2 k2' /IH.
        by apply/rtc_l/app_ee1.
  Qed.
  Global Instance app_els :
    Proper (el_contextuals ==> el_contextuals ==> el_contextuals) app.
  Proof.
    move=> k1 k1'; elim: k1 k1' /.
    - move=> k1 k2 k2'; elim: k2 k2' /.
      + reflexivity.
      + move=> k2 k2' k2'' CStep CSteps.
          by apply/rtc_l/app_el2.
    - move=> k1 k1' k1'' CStep CSteps IH k2 k2' /IH.
        by apply/rtc_l/app_el1.
  Qed.
  Global Instance app_lls :
    Proper (ll_contextuals ==> ll_contextuals ==> ll_contextuals) app.
  Proof.
    move=> k1 k1'; elim: k1 k1' /.
    - move=> k1 k2 k2'; elim: k2 k2' /.
      + reflexivity.
      + move=> k2 k2' k2'' CStep CSteps.
          by apply/rtc_l/app_ll2.
    - move=> k1 k1' k1'' CStep CSteps IH k2 k2' /IH.
        by apply/rtc_l/app_ll1.
  Qed.

  (* TODO facts about commutation and getting info out from
     multiple of these holding at once *)

  Global Instance ee_Symmetric `{!Symmetric R_ee} : Symmetric ee_contextual.
  Proof. move=> _ _ [l1 x y l2 Step] //. Qed.
  Global Instance ll_Symmetric `{!Symmetric R_ll} : Symmetric ll_contextual.
  Proof. move=> _ _ [l1 l l' l2 Step] //. Qed.

  Lemma ee_cons1 : forall {x k h},
      ee_contextual (x :: k) h <->
      (exists2 y, h = y :: k & R_ee x y) \/
      (exists2 k', h = x :: k' & ee_contextual k k').
  Proof.
    move=> x k h; split.
    - move=> /ee_inv => -[[| ? l1]] [x'] [y] [l2] /= [[<- ->]] [->] Step; first by eauto.
      right; by eexists.
    - move=> [[y -> Step] | [k' -> CStep]].
      + by apply: ee_beginning.
      + by apply: cons_ee2.
  Qed.
  Lemma ee_cons2 : forall {k y h'},
      ee_contextual k (y :: h') <->
      (exists2 x, k = x :: h' & R_ee x y) \/
      (exists2 h, k = y :: h & ee_contextual h h').
  Proof.
    move=> k y h'; split.
    - move=> /ee_inv => -[[| ? l1]] [x] [y'] [l2] /= [->] [[<- ->]] Step; first by eauto.
      right; by eexists.
    - move=> [[x -> Step] | [h -> CStep]].
      + by apply: ee_beginning.
      + by apply: cons_ee2.
  Qed.
  Lemma el_cons : forall {x k h},
      el_contextual (x :: k) h <->
      (exists2 l, h = l ++ k & R_el x l) \/
      (exists2 k', h = x :: k' & el_contextual k k').
  Proof.
    move=> x k h; split.
    - move=> /el_inv => -[[| ? l1]] [x'] [y] [l2] /= [[<- ->]] [->] Step; first by eauto.
      right; by eexists.
    - move=> [[y -> Step] | [k' -> CStep]].
      + by apply: el_beginning.
      + by apply: cons_el2.
  Qed.

  Lemma ee_app1 : forall {k1 k2 h},
      ee_contextual (k1 ++ k2) h <->
      (exists2 h1, h = h1 ++ k2 & ee_contextual k1 h1) \/
      (exists2 h2, h = k1 ++ h2 & ee_contextual k2 h2).
  Proof.
    move=> k1 k2 h.
    split.
    - move=> /ee_inv [l1 [x [y [l2 [E [-> Step]]]]]].
      move: E => /app_eq_inv [[[| ? j] /= [->]] | [j [->]]] => [<- | [<- ->] | ->].
      + right; eexists.
        * rewrite app_nil_r //.
        * apply/ee_beginning/Step.
      + left; eexists.
        * rewrite cons_middle !app_assoc -[_ ++ j]app_assoc //.
        * by constructor.
      + right; eexists.
        * rewrite -app_assoc //.
        * auto.
    - move=> [[h1 -> Step] | [h2 -> Step]].
      + by apply: app_ee1.
      + by apply: app_ee2.
  Qed.
  Lemma el_app1 : forall {k1 k2 h},
      el_contextual (k1 ++ k2) h <->
      (exists2 h1, h = h1 ++ k2 & el_contextual k1 h1) \/
      (exists2 h2, h = k1 ++ h2 & el_contextual k2 h2).
  Proof.
    move=> k1 k2 h.
    split.
    - move=> /el_inv [l1 [x [y [l2 [E [-> Step]]]]]].
      move: E => /app_eq_inv [[[| ? j] /= [->]] | [j [->]]] => [<- | [<- ->] | ->].
      + right; eexists.
        * rewrite app_nil_r //.
        * apply/el_beginning/Step.
      + left; eexists.
        * rewrite !app_assoc -[_ ++ j]app_assoc //.
        * by constructor.
      + right; eexists.
        * rewrite -app_assoc //.
        * auto.
    - move=> [[h1 -> Step] | [h2 -> Step]].
      + by apply: app_el1.
      + by apply: app_el2.
  Qed.

  (* TODO distinction between what kind of step is being contextually taken? *)
  Lemma ll_app1 : forall {k1 k2 h},
      ll_contextual (k1 ++ k2) h <->
      [\/ exists2 h1, h = h1 ++ k2 & ll_contextual k1 h1,
       exists2 h2, h = k1 ++ h2 & ll_contextual k2 h2 |
       exists k11 k12 k21 k22 l', [/\ k1 = k11 ++ k12, k2 = k21 ++ k22,
                              h = k11 ++ l' ++ k22 & R_ll (k12 ++ k21) l']].
  Proof.
    move=> k1 k2 h; split.
    - move=> /ll_inv [l1 [l [l' [l2 [E [-> Step]]]]]].
      move: E => /app_eq_inv [[j [->]] | [j [-> ->]]].
      + move=> /app_eq_inv [[j' [E ->]] | [j' [-> ->]]].
        * constructor 3; eexists _, _, _, _, _; constructor; only 1-3: done.
          rewrite -E //.
        * constructor 1; eexists.
          -- rewrite !app_assoc -[_ ++ j']app_assoc //.
          -- by constructor.
      + constructor 2; eexists.
        * rewrite -app_assoc //.
        * by constructor.
    - move=> [[h1 -> CStep] | [h2 -> CStep] | [k11 [k12 [k21 [k22 [l' [-> -> -> Step]]]]]]].
      + by apply: app_ll1.
      + by apply: app_ll2.
      + rewrite -app_assoc [k12 ++ _]app_assoc.
          by constructor.
  Qed.
  Lemma el_app2 : forall {k h1 h2},
      el_contextual k (h1 ++ h2) <->
      [\/ exists2 k1, k = k1 ++ h2 & el_contextual k1 h1,
       exists2 k2, k = h1 ++ k2 & el_contextual k2 h2 |
       exists h11 h12 h21 h22 x, [/\ h1 = h11 ++ h12, h2 = h21 ++ h22,
                             k = h11 ++ [x] ++ h22 & R_el x (h12 ++ h21)]].
    move=> k1 k2 h; split.
    - move=> /el_inv [l1 [x [l [l2 [-> [E Step]]]]]].
      move: E => /app_eq_inv [[j [->]] | [j [-> ->]]].
      + move=> /app_eq_inv [[j' [E ->]] | [j' [-> ->]]].
        * constructor 3; eexists _, _, _, _, _; constructor; only 1-3: done.
          rewrite -E //.
        * constructor 1; eexists.
          -- rewrite !app_assoc -[_ ++ j']app_assoc //.
          -- by constructor.
      + constructor 2; eexists.
        * rewrite -app_assoc //.
        * by constructor.
    - move=> [[h1 -> CStep] | [h2 -> CStep] | [k11 [k12 [k21 [k22 [l' [-> -> -> Step]]]]]]].
      + by apply: app_el1.
      + by apply: app_el2.
      + rewrite -app_assoc [k12 ++ _]app_assoc.
          by constructor.
  Qed.

  (*
  Global Instance ee_Transitive `{!Transitive R_ee} : Transitive ee_contextual.
  Proof.
    move=> k j h [l1 x y l2 Step] /ee_app1 [[h1] | [h2 -> /ee_app1 [[h2'] | [h2']]]].
   *)

  Lemma ee_perm : commute (≡ₚ) ee_contextual.
  Proof.
    move=> k h; split=> [[j Perm CStep] | [j [l1 x y l2 Step]]].
    - move: CStep Perm => [l1 x y l2 Step].
      rewrite -Permutation_middle => /(Permutation_cons_inv _ _ _) [k1 [k2 [-> Perm]]].
      eexists; first by econstructor.
      rewrite -!Permutation_middle; by f_equiv.
    - rewrite -Permutation_middle => /Permutation_sym /(Permutation_cons_inv _ _ _)
                                     [h1 [h2 [-> Perm]]].
      eexists; last by constructor.
      rewrite -!Permutation_middle; by f_equiv.
  Qed.
  Lemma ees_perm : commute (≡ₚ) ee_contextuals.
  Proof. apply/rtc_commute2/ee_perm. Qed.

  Lemma el_perm : subcommute (≡ₚ) el_contextual.
  Proof.
    move=> k h [j Perm CStep].
    move: CStep Perm => [l1 x l l2 Step].
    rewrite -Permutation_middle => /(Permutation_cons_inv _ _ _) [k1 [k2 [-> Perm]]].
    eexists; first by econstructor.
    rewrite !app_assoc [k1 ++ _]Permutation_app_comm
            [l1 ++ _]Permutation_app_comm -!app_assoc; by f_equiv.
  Qed.
  Lemma els_perm : subcommute (≡ₚ) el_contextuals.
  Proof. apply/rtc_subcommute2/el_perm. Qed.

  Instance Forall2_proper : Proper ((≡ᵣ) ==> (≡ᵣ)) (@Forall2 A A).
  Proof.
    move=> R R' E; elim.
    - split; by inversion 1.
    - move=> x l IH l'.
      split; inversion 1; firstorder.
  Qed.

  Lemma ee_contextuals_alt : ee_contextuals ≡ᵣ Forall2 (rtc R_ee).
  Proof.
    move=> k h; split.
    - elim: k h / => // k k' h [l1 x y l2 Step] _.
      apply transitivity; apply/Forall2_app/Forall2_app => //.
        by repeat econstructor.
    - elim: k h / => // x y l l' Steps _ CSteps.
        by apply: cons_ees.
  Qed.
End ListRels.
Arguments ee_relation : clear implicits.
Arguments el_relation : clear implicits.
Arguments ll_relation : clear implicits.
Arguments ee_contextual {_}.
Arguments el_contextual {_}.
Arguments ll_contextual {_}.
Arguments ee_contextuals {_}.
Arguments el_contextuals {_}.
Arguments ll_contextuals {_}.

Instance ee_mono : Proper ((⊆ᵣ) ==> (⊆ᵣ)) (@ee_contextual A).
Proof.
  move=> A R R' Sub _ _ [l1 x y l2 Step].
  apply/ee_Contextual/Sub/Step.
Qed.
Instance ll_mono : Proper ((⊆ᵣ) ==> (⊆ᵣ)) (@ll_contextual A).
Proof.
  move=> A R R' Sub _ _ [l1 l l' l2 Step].
  apply/ll_Contextual/Sub/Step.
Qed.

Lemma ee_flip : forall {A} {R : ee_relation A},
    flip (ee_contextual R) ≡ᵣ ee_contextual (flip R).
Proof. move=> A R k h; split=> -[l1 x y l2 Step]; by constructor. Qed.
Lemma ll_flip : forall {A} {R : ll_relation A},
    flip (ll_contextual R) ≡ᵣ ll_contextual (flip R).
Proof. move=> A R k h; split=> -[l1 x y l2 Step]; by constructor. Qed.

Lemma ee_app2 : forall {A} {R : ee_relation A} {k h1 h2},
    ee_contextual R k (h1 ++ h2) <->
    (exists2 k1, k = k1 ++ h2 & ee_contextual R k1 h1) \/
    (exists2 k2, k = h1 ++ k2 & ee_contextual R k2 h2).
Proof.
  move=> A R k h1 h2.
  rewrite [L in L <-> _](ee_flip _ _) ee_app1.
    by setoid_rewrite <- (fun a b => ee_flip a b).
Qed.
Lemma ll_app2 : forall {A} {R : ll_relation A} {k h1 h2},
    ll_contextual R k (h1 ++ h2) <->
    [\/ exists2 k1, k = k1 ++ h2 & ll_contextual R k1 h1,
     exists2 k2, k = h1 ++ k2 & ll_contextual R k2 h2 |
     exists h11 h12 h21 h22 l, [/\ h1 = h11 ++ h12, h2 = h21 ++ h22,
                           k = h11 ++ l ++ h22 & R l (h12 ++ h21)]].
Proof.
  move=> A R k h1 h2.
  rewrite [L in L <-> _](ll_flip _ _) ll_app1.
    by setoid_rewrite <- (fun a b => ll_flip a b).
Qed.

Lemma ee_map_fwd : forall {A B} {R : ee_relation A} {S : ee_relation B} {f k h},
    Forall2 (fun x y => R x y -> S (f x) (f y)) k h ->
    ee_contextual R k h -> ee_contextual S (f <$> k) (f <$> h).
Proof.
  move=> A B R S f k h Resp CStep.
  move: CStep Resp => [l1 x y l2 Step] /Forall2_middle /(_ Step) Step'.
  rewrite !fmap_app //.
Qed.
Lemma el_map_fwd : forall {A B} {R : el_relation A} {S : el_relation B} {f k h},
    Forall (fun x => forall l, R x l -> S (f x) (f <$> l)) k ->
    el_contextual R k h -> el_contextual S (f <$> k) (f <$> h).
Proof.
  move=> A B R S f k h Resp CStep.
  move: CStep Resp => [l1 x l l2 Step] /Forall_middle /(_ _ Step) Step'.
  rewrite !fmap_app //.
Qed.
Lemma ll_map_fwd : forall {A B} {R : ll_relation A} {S : ll_relation B} {f k h},
    (* technically we only need it on sublists, but... TODO I guess *)
    Proper (R ==> S) (fmap f) ->
    ll_contextual R k h -> ll_contextual S (f <$> k) (f <$> h).
Proof.
  move=> A B R S f k h Resp CStep.
  move: CStep Resp => [l1 l l' l2 Step] /(_ _ _ Step).
  rewrite !fmap_app //.
Qed.
(* TODO bwd versions? *)
(* TODO more of these? *)
Lemma ee_map1_bwd : forall {A B} {R : ee_relation A} {S : ee_relation B} {f k h},
    Forall (fun x => forall y, S (f x) y -> exists2 _y, y = f _y & R x _y) k ->
    ee_contextual S (f <$> k) h -> exists2 _h, h = f <$> _h & ee_contextual R k _h.
Proof.
  move=> A B R S f k h Lift CStep.
  move: CStep Lift => /ee_inv [l1] [x] [y] [l2]
                      [/fmap_app_inv [_l1] [[| _x _l2]] [->] [// [-> ->]] ->]
                      [->] Step /Forall_middle /(_ _ Step) [_y -> Step'].
  eexists; last by constructor.
  rewrite !fmap_app //.
Qed.
Lemma el_map1_bwd : forall {A B} {R : el_relation A} {S : el_relation B} {f k h},
    Forall (fun x => forall l, S (f x) l -> exists2 _l, l = f <$> _l & R x _l) k ->
    el_contextual S (f <$> k) h -> exists2 _h, h = f <$> _h & el_contextual R k _h.
  move=> A B R S f k h Lift CStep.
  move: CStep Lift => /el_inv [l1] [x] [y] [l2]
                      [/fmap_app_inv [_l1] [[| _x _l2]] [->] [// [-> ->]] ->]
                      [->] Step /Forall_middle /(_ _ Step) [_y -> Step'].
  eexists; last by constructor.
  rewrite !fmap_app //.
Qed.

Definition parallel {A} (R S : ll_relation A) : ll_relation A :=
  fun k k' => exists l1 l1' l2 l2', [/\ k = l1 ++ l2, k' = l1' ++ l2', R l1 l1' & S l2 l2'].
Lemma ee_comp : forall {A} {R S : ee_relation A},
    ee_contextual R >> ee_contextual S ≡ᵣ
                  parallel (ee_contextual R) (ee_contextual S) ∪ᵣ
                  parallel (ee_contextual S) (ee_contextual R) ∪ᵣ
                  ee_contextual (R >> S).
Proof.
  move=> A R S k h; split.
  - move=> [j [l1 x y l2 RStep]] /ee_app1 [[h1] | [h2 -> /ee_app1 [[h2'] | [h2']]]].
    + move=> -> [m1 x' y' m2 SStep].
      left; right; eexists _, _, _, _; split=> //.
        by apply: ee_beginning.
    + move=> -> /ee_singleton1 [z -> SStep].
      right; constructor; by eexists.
    + move=> -> [m1 x' y' m2 SStep].
      rewrite ![l1 ++ _ ++ _]app_assoc.
      left; left; eexists _, _, _, _; split=> //.
  - move=> [[] [l1] [l1'] [l2] [l2'] [-> -> Step1 Step2] | [l1 x z l2 [y RStep SStep]]].
    + by eexists; [apply: app_ee1 | apply: app_ee2].
    + by eexists; [apply: app_ee2 | apply: app_ee1].
    + by eexists.
Qed.

Lemma ee_subcommute : forall {A} {R S : ee_relation A},
    subcommute R S -> subcommute (ee_contextual R) (ee_contextual S).
Proof.
  move=> A R S SubComm.
  rewrite /subcommute !ee_comp (Ur_proper _ _ Ur_comm _ _ (reflexivity _)).
  apply Ur_mono2; first by done.
    by apply ee_mono.
Qed.
Lemma ee_commute : forall {A} {R S : ee_relation A},
    commute R S -> commute (ee_contextual R) (ee_contextual S).
Proof.
  move=> A R S /subcommute2 [SubComm1 SubComm2].
  apply/subcommute2.
  split; by apply: ee_subcommute.
Qed.

Inductive swap2 {A} : ll_relation A :=
| Swap2 : forall x y, swap2 [x; y] [y; x].
Instance swap2_sym : Symmetric (@swap2 A).
Hint Constructors swap2.
Proof. move=> A _ _ [? ?]. constructor. Qed.
Definition swap {A} : ll_relation A := ll_contextual swap2.
Definition swaps {A} : ll_relation A := ll_contextuals swap2.
Typeclasses Transparent swap swaps.

(* TODO Maybe specialize ll_app{1,2} to swap? *)
Lemma swap_app1 : forall {A} {k1 k2 h : list A},
    swap (k1 ++ k2) h <->
    [\/ exists2 h1, h = h1 ++ k2 & swap k1 h1,
     exists2 h2, h = k1 ++ h2 & swap k2 h2 |
     exists k1' x y k2', [/\ k1 = k1' ++ [x], k2 = y :: k2' & h = k1' ++ [y; x] ++ k2']].
Proof.
  move=> A k1 k2 h.
  rewrite /swap ll_app1; split; move=> [||].
  1,2,4,5: by firstorder.
  - move=> [k11] [k12] [k21] [k22] [l'] [-> -> ->].
    case E: _ _ / => [x y].
    move: k12 E => [| ? [| ? [| ? ?]]] // => [| [-> ->] | [-> -> ->]].
    + rewrite app_nil_l app_nil_r => ->.
      constructor 2; eexists; [done | by apply: ll_beginning].
    + constructor 3; repeat eexists.
    + constructor 1; eexists.
      * rewrite app_assoc //.
      * by apply: ll_end.
  - move=> [k1'] [x] [y] [k2'] [-> -> ->].
    constructor 3; by repeat eexists.
Qed.
Lemma swap_app2 : forall {A} {k h1 h2 : list A},
    swap k (h1 ++ h2) <->
    [\/ exists2 k1, k = k1 ++ h2 & swap k1 h1,
     exists2 k2, k = h1 ++ k2 & swap k2 h2 |
     exists h1' x y h2', [/\ h1 = h1' ++ [x], h2 = y :: h2' & k = h1' ++ [y; x] ++ h2']].
Proof.
  move=> A k h1 h2.
  rewrite /swap ll_app2; split; move=> [||].
  1,2,4,5: by firstorder.
  - move=> [h11] [h12] [h21] [h22] [l] [-> -> ->].
    case E: _ / => [x y].
    move: h12 E => [| ? [| ? [| ? ?]]] // => [| [-> ->] | [-> -> ->]].
    + rewrite app_nil_l app_nil_r => ->.
      constructor 2; eexists; [done | by apply: ll_beginning].
    + constructor 3; repeat eexists.
    + constructor 1; eexists.
      * rewrite app_assoc //.
      * by apply: ll_end.
  - move=> [h1'] [x] [y] [h2'] [-> -> ->].
    constructor 3; by repeat eexists.
Qed.

Lemma swap_commute : forall {A} {R : ee_relation A},
    commute (ee_contextual R) swap.
Proof.
  move=> A R k h; split=> -[j].
  - move=> [l1 x y l2 Step] /swap_app1 [[h1 -> Swap] | [h2 ->] |
                                       [k1'] [x'] [y'] [k2'] [-> [<- <-] ->]].
    + eexists; [by apply: app_ll1 | done].
    + move=> /ll_inv [[| ? l1']] [l] [l'] [l2']. (* [[x' y']] [[<- ->]] -> /=. *)
      * move=> [E] [->] Swap; move: Swap E => [x' y'] [<- ->]; eexists.
        -- by apply: (ll_Contextual l1 [x; y'] _ l2').
        -- rewrite /= !cons_middle ![l1 ++ _ ++ _]app_assoc //.
      * move=> [[<- ->]] [->]; eexists.
        -- rewrite 2!app_assoc //.
        -- rewrite -!app_assoc //.
    + eexists.
      * rewrite -app_assoc [[x'] ++ _]app_assoc //.
      * done.
  - move=> [l1 l l' l2 [x y]] /ee_app1
      [[h1 ->] | [h2 -> /ee_app1 [[h1' ->] | [h2' ->]]]] CStep.
    + eexists; [by apply: app_ee1 | done].
    + change [y; x] with ([y] ++ [x]) in CStep.
      move: CStep => /ee_app1 [[h1'' -> /ee_singleton1 [y' -> Step]] |
                              [h2' -> /ee_singleton1 [x' -> Step]]];
                      (eexists; last by done).
      * rewrite ![l1 ++ [_; _] ++ _]cons_middle ![l1 ++ _ ++ _]app_assoc //.
      * done.
    + eexists; last by done.
      apply: app_ee2; first by done.
        by apply: app_ee2.
Qed.

Lemma swap_perm : forall {A}, @swap A ⊆ᵣ Permutation.
Proof.
  move=> A _ _ [l1 l l' l2 [x y]].
  rewrite Permutation_swap //.
Qed.

Lemma swaps_alt : forall {A}, @swaps A ≡ᵣ Permutation.
Proof.
  move=> A.
  apply antisymmetry; first by apply rtc_universal, swap_perm.
  move=> k h; elim: k h /.
  - reflexivity.
  - move=> ? ? ? _ E.
      by apply: cons_lls.
  - move=> x y l; by apply/rtc_once/(ll_beginning [y; x] [x; y] l).
  - by etransitivity.
Qed.
