From Coq Require Import ssreflect ssrfun ssrbool.
(*
Require Import stdpp.pmap.
Require Import stdpp.mapset.
 *)
Require Import BinNums.
Require Import stdpp.pmap.
Require Import stdpp.gmultiset.
Require Import Common.

Definition name : Type := positive.
Inductive ty : Type :=
| atom : name -> ty
| neg_atom : name -> ty
| one : ty
| perp : ty
| tens : ty -> ty -> ty
| par : ty -> ty -> ty.

Coercion atom : name >-> ty.

Instance decide_ty_eq : Decision (A = B :> ty).
Proof. rewrite /Decision. repeat decide equality. Qed.

Fixpoint neg (T : ty) : ty :=
  match T with
  | atom x => neg_atom x
  | neg_atom x => atom x
  | one => perp
  | perp => one
  | tens A B => par (neg A) (neg B)
  | par A B => tens (neg A) (neg B)
  end.
Lemma neg_involution : forall A, neg (neg A) = A.
Proof. elim => //= ? -> ? -> //. Qed.
Hint Resolve neg_involution.

Definition lto (A B : ty) : ty := par (neg A) B.

Inductive pexp : Type := Pexp : list coeq -> list term -> pexp
with coeq : Type := Coeq : term -> term -> coeq
with term : Type :=
     | var : name -> term
     | star : term
     | costar : term
     | ttens : term -> term -> term
     | tpar : term -> term -> term.
Notation soln := (list coeq).

Coercion var : name >-> term.

Definition lhs : coeq -> term := fun '(Coeq t u) => t.
Definition rhs : coeq -> term := fun '(Coeq t u) => u.

Definition coeqs : pexp -> soln := fun '(Pexp Θ ts) => Θ.
Definition terms : pexp -> list term := fun '(Pexp Θ ts) => ts.

Instance decide_term_eq : Decision (t = u :> term).
Proof. rewrite /Decision. repeat decide equality. Qed.
Instance decide_coeq_eq : Decision (C = D :> coeq).
Proof. rewrite /Decision. repeat decide equality. Qed.
Instance decide_pexp_eq : Decision (p = q :> pexp).
Proof. rewrite /Decision. repeat decide equality. Qed.

Inductive coeq_equiv_ (C : coeq) : coeq -> Prop :=
| coeq_refl : coeq_equiv_ C C
| coeq_swap : coeq_equiv_ C (Coeq (rhs C) (lhs C)).
Hint Constructors coeq_equiv_.
Instance coeq_equiv : Equiv coeq := coeq_equiv_.
Instance coeq_equiv_sym : Symmetric (≡@{coeq}).
Proof. move=> [t u] ? []; constructor. Qed.
Instance coeq_equiv_equivalence : Equivalence (≡@{coeq}).
Proof.
  split.
  - constructor.
  - move=> [t u] ? []; constructor.
  - move=> [t u] ? ? [] []; constructor.
Qed.

Definition sym : relation soln := ee_contextual coeq_equiv.
Definition syms : relation soln := rtc sym.
Definition soln_equiv_step : relation soln := sym ∪ᵣ swap.
Instance soln_equiv : Equiv soln := rtc soln_equiv_step.
Instance pexp_equiv : Equiv pexp := fun p q => coeqs p ≡ coeqs q /\ terms p = terms q.

Instance: Proper ((≡) ==> (=) ==> (≡)) Pexp.
Proof. firstorder. Qed.
Instance: Proper ((≡) ==> (≡)) coeqs.
Proof. firstorder. Qed.
Instance: Proper ((≡) ==> (=)) terms.
Proof. firstorder. Qed.

Instance sym_sym : Symmetric sym.
Proof. apply: ee_Symmetric. Qed.
Instance soln_equiv_step_sym : Symmetric soln_equiv_step.
Proof.
  move=> ? ? [] H.
  - left; by apply: ee_Symmetric.
  - by right.
Qed.
Instance syms_equivalence : Equivalence syms.
Proof. apply: rtc_equivalence. Qed.
Instance soln_equiv_equivalence : Equivalence (≡@{soln}).
Proof. apply: rtc_equivalence. Qed.
Instance pexp_equiv_equivalence : Equivalence (≡@{pexp}).
Proof.
  constructor.
  - done.
  - move=> p q [? ?].
      by split.
  - move=> p q r [? ?] [? ?].
    split; by etransitivity.
Qed.

Lemma swap_equiv : forall {Θ Θ'}, swap Θ Θ' -> Θ ≡ Θ'.
Proof. move=> *; by apply/rtc_once/rel_inr. Qed.
Lemma sym_equiv : forall {Θ Θ'}, sym Θ Θ' -> Θ ≡ Θ'.
Proof. move=> *; by apply/rtc_once/rel_inl. Qed.
Lemma swaps_equiv : forall {Θ Θ'}, swaps Θ Θ' -> Θ ≡ Θ'.
Proof. move=> ? ?; apply/rtc_subrel=> ? ?; apply: rel_inr. Qed.
Lemma syms_equiv : forall {Θ Θ'}, syms Θ Θ' -> Θ ≡ Θ'.
Proof. move=> ? ?; apply/rtc_subrel=> ? ?; apply: rel_inl. Qed.

Lemma syms_alt : syms ≡ᵣ Forall2 (≡).
Proof.
  setoid_rewrite <- (Forall2_proper _ _ (@rtc_redundant _ coeq_equiv
                    Equivalence_Reflexive Equivalence_Transitive)).
  apply: ee_contextuals_alt.
Qed.

Lemma soln_equiv_step_commute : commute sym swap.
Proof. apply: swap_commute. Qed.

Lemma soln_equiv_alt : (≡) ≡ᵣ swaps >> syms.
Proof. apply/rtc_ell/subcommute_weaken/soln_equiv_step_commute. Qed.
Lemma soln_equiv_alt_alt : (≡) ≡ᵣ (≡ₚ) >> Forall2 (≡).
Proof. rewrite soln_equiv_alt swaps_alt syms_alt //. Qed.

Lemma soln_equiv_nil1 : forall {Θ}, [] ≡ Θ <-> [] = Θ.
Proof.
  move=> Θ; split=> [
    /soln_equiv_alt_alt [Θ' /Permutation_sym /Permutation_nil -> /Forall2_nil_inv_l] | ->] //.
Qed.
Lemma soln_equiv_nil2 : forall {Θ}, Θ ≡ [] <-> Θ = [].
Proof.
  move=> Θ; split=> [/soln_equiv_alt_alt [Θ' Perm /Forall2_nil_inv_r E] | ->] //.
  move: E Perm => -> /Permutation_nil //.
Qed.

Instance permute_coeqs : (≡ₚ) ⊆ᵣ (≡@{soln}).
Proof.
  move=> Θ Ξ Perm.
  apply/soln_equiv_alt; exists Ξ.
  - by apply/swaps_alt.
  - reflexivity.
Qed.

Instance: Proper ((≡ₚ) ==> (=) ==> (≡)) Pexp.
Proof. move=> ? ? /permute_coeqs. firstorder. Qed.

Instance: Proper ((≡) ==> (≡) ==> (≡)) (@app coeq).
Proof.
  rewrite soln_equiv_alt => Θ Θ'' [Θ' Swaps Syms] Ξ Ξ'' [Ξ' Swaps' Syms'].
  eexists.
  - by apply: app_lls.
  - by apply: app_ees.
Qed.

Instance: Proper ((≡) ==> (≡) ==> (≡)) (@cons coeq).
Proof.
  rewrite soln_equiv_alt => C C' /(rtc_once _ _) Sym1 Θ Θ'' [Θ' Swaps' Syms'].
  eexists.
  - by apply: cons_lls.
  - by apply: cons_ees.
Qed.

Lemma soln_equiv_subcommute_ee : forall {R} `{!Proper ((≡) ==> (=) ==> iff) R},
    subcommute (≡@{soln}) (ee_contextual R).
Proof.
  move=> R Pro Θ Θ''' [Θ'' /soln_equiv_alt_alt [Θ' Perm All] CStep].
  case: Θ'' Θ''' / CStep All Perm => Θ'1 ? C' Θ'2 Step
    /Forall2_app_inv_r [Θ1] [_] [/syms_alt Syms1]
    [/(Forall2_cons_inv_r _ _ _ _) [C] [Θ2] [E] [/syms_alt Syms2] -> ->] Perm.
  move: E Step => <- Step.
  move: (Perm >>> ee_Contextual _ _ _ _ Step) => /ee_perm [Ξ CStep Perm'].
  eexists; first by done.
  rewrite Perm' (syms_equiv Syms1) (syms_equiv Syms2) //.
Qed.

Lemma soln_equiv_subcommute_el : forall {R} `{!Proper ((≡) ==> (=) ==> iff) R},
    subcommute (≡@{soln}) (el_contextual R).
Proof.
  move=> R Pro Θ Θ''' [Θ'' /soln_equiv_alt_alt [Θ' Perm All] CStep].
  case: Θ'' Θ''' / CStep All Perm => Θ'1 ? Ξ Θ'2 Step
    /Forall2_app_inv_r [Θ1] [?] [/syms_alt Syms1]
    [/(Forall2_cons_inv_r _ _ _ _) [C] [Θ2] [E] [/syms_alt Syms2] -> ->] Perm.
  move: E Step => <- Step.
  move: (Perm >>> el_Contextual _ _ _ _ Step) => /el_perm [Λ CStep Perm'].
  eexists; first by done.
  rewrite Perm' (syms_equiv Syms1) (syms_equiv Syms2) //.
Qed.


Definition var_count : Type := gmultiset name.
Definition substitution : Type := Pmap term.
Class Vars (A : Type) := vars : A -> var_count.
Class Subst (A : Type) := sim_subst : substitution -> A -> A.
Arguments vars {_ _} !_ /.
Arguments sim_subst : simpl never.
Definition count `{Vars A} (x : name) (a : A) : nat := multiplicity x (vars a).
Notation "x # a" := (count x a = 0) (no associativity, at level 50).
Notation subst t x := (sim_subst {[x:=t]}).

Instance term_vars : Vars term :=
  fix term_vars t :=
    let _ := term_vars : Vars term in
    match t with
    | var x => {[x]}
    | star => ∅
    | costar => ∅
    | ttens l r => vars l ⊎ vars r
    | tpar l r => vars l ⊎ vars r
    end.
Instance term_subst : Subst term :=
  fix term_subst (σ : substitution) (u : term) {struct u} :=
    let _ := term_subst : Subst term in
    match u with
    | var x => default (var x) (σ !! x)
    | star => star
    | costar => costar
    | ttens l r => ttens (sim_subst σ l) (sim_subst σ r)
    | tpar l r => tpar (sim_subst σ l) (sim_subst σ r)
    end.
Ltac clear_sing := rewrite ?lookup_singleton ?lookup_singleton_ne //.
Lemma lookup_singleton' : forall {A} {x x' : name} {a : A},
    ({[x := a]} : Pmap A) !! x' = if decide (x = x') is left _ then Some a else None.
Proof. move=> *; case: (decide _) => [<- | NE]; clear_sing. Qed.

Definition multiplicityE :=
  (multiplicity_disj_union (A:=name), multiplicity_empty (A:=name),
   multiplicity_singleton (A:=name), multiplicity_singleton' (A:=name)).
Ltac prove_count := move=> *; rewrite /count /= multiplicityE //.
Lemma count_var_eq : forall {x}, count x (var x) = 1.
Proof. prove_count. Qed.
Lemma count_var_ne : forall {x x'}, x <> x' -> count x (var x') = 0.
Proof. prove_count; rewrite decide_False //. Qed.
Lemma count_var : forall {x x'}, count x (var x') = if decide (x = x') is left _ then 1 else 0.
Proof. prove_count. Qed.
Lemma count_star : forall {x}, count x star = 0.
Proof. done. Qed.
Lemma count_costar : forall {x}, count x costar = 0.
Proof. done. Qed.
Lemma count_ttens : forall {x t u}, count x (ttens t u) = count x t + count x u.
Proof. prove_count. Qed.
Lemma count_tpar : forall {x t u}, count x (tpar t u) = count x t + count x u.
Proof. prove_count. Qed.
Definition count_term := (@count_var_eq, @count_var, @count_star, @count_costar,
                          @count_ttens, @count_tpar).
(*
Lemma count_term : forall {x} {t : term},
    count x t =
    match t with
    | var x' => if decide (x = x') is left _ then 1 else 0
    | star => 0
    | costar => 0
    | ttens l r => count x l + count x r
    | tpar l r => count x l + count x r
    end.
Proof. move=> x [] // *; rewrite /count /= multiplicityE //. Qed.
*)

Instance vars_name : Vars name := singleton.
Instance coeq_vars : Vars coeq := fun '(Coeq t u) => vars t ⊎ vars u.
Instance coeq_subst : Subst coeq := fun σ '(Coeq t u) => Coeq (sim_subst σ t) (sim_subst σ u).
Instance prod_vars `{Vars A, Vars B} : Vars (A * B) := fun '(t, u) => vars t ⊎ vars u.
Instance prod_subst `{Subst A, Subst B} : Subst (A * B)
  := fun σ '(t, u) => (sim_subst σ t, sim_subst σ u).
Notation Msum := (foldr (⊎) ∅).
Instance list_vars `{Vars A} : Vars (list A) := fun l => Msum (vars <$> l).
Instance list_subst `{Subst A} : Subst (list A) := fun σ => fmap (sim_subst σ).
Instance pexp_vars : Vars pexp := fun '(Pexp Θ ts) => vars Θ ⊎ vars ts.
Instance pexp_subst : Subst pexp := fun σ '(Pexp Θ ts) => Pexp (sim_subst σ Θ) (sim_subst σ ts).

Instance vars_list_proper `{Vars A} : Proper ((≡ₚ) ==> (=)) (vars (A:=list A)).
Proof. apply: compose_proper; [apply: foldr_permutation_proper' | apply: fmap_Permutation]. Qed.
Instance count_list_proper `{Vars A} : Proper ((=) ==> (≡ₚ) ==> (=)) (count (A:=list A)).
Proof. move=> ? ? <- ? ? E; rewrite /count E //. Qed.

(* feels like this should be in stdpp o.O *)
Lemma monoid_fold `{Equivalence A R, !Assoc R f, !LeftId R e f, !Proper (R ==> R ==> R) f} :
  forall l1 l2, R (foldr f e (l1 ++ l2)) (f (foldr f e l1) (foldr f e l2)).
Proof.
  elim=> [| x xs IH] l2.
  - by symmetry.
  - rewrite /= -assoc.
    f_equiv.
      by rewrite IH.
Qed.
Corollary monoid_foldL `{Assoc A eq f, !LeftId eq e f} :
  forall l1 l2, foldr f e (l1 ++ l2) = f (foldr f e l1) (foldr f e l2).
Proof. apply: monoid_fold. Qed.

Lemma varsc_def : forall t u, vars (Coeq t u) = vars t ⊎ vars u. Proof. done. Qed.
Lemma varspr_def : forall `{Vars A, Vars B} (t : A) (u : B), vars (t, u) = vars t ⊎ vars u.
Proof. done. Qed.
Definition varsl_nil `{Vars A} : vars (A:=list A) [] = ∅ := erefl.
Definition varsl_cons `{Vars A} (t : A) ts : vars (t :: ts) = vars t ⊎ vars ts := erefl.
Lemma varspe_def : forall Θ ts, vars (Pexp Θ ts) = vars Θ ⊎ vars ts.
Proof. done. Qed.
Lemma varsl_app : forall `{Vars A} (ts us : list A), vars (ts ++ us) = vars ts ⊎ vars us.
Proof. move=> *. rewrite /vars /list_vars fmap_app monoid_foldL //. Qed.
Definition varsE := (varsc_def, @varspr_def, @varsl_nil, @varsl_cons, varspe_def, @varsl_app).

Lemma count_coeq : forall {x t u}, count x (Coeq t u) = count x t + count x u.
Proof. prove_count. Qed.
Lemma count_pair : forall `{Vars A, Vars B} {x} {t : A} {u : B}, count x (t, u) = count x t + count x u.
Proof. prove_count. Qed.
Lemma count_pexp : forall {x Θ ts}, count x (Pexp Θ ts) = count x Θ + count x ts.
Proof. prove_count. Qed.
Lemma count_nil : forall `{Vars A} {x}, count (A:=list A) x [] = 0.
Proof. done. Qed.
Lemma count_cons : forall `{Vars A} {x} {a : A} {as_}, count x (a :: as_) = count x a + count x as_.
Proof. prove_count. Qed.
Lemma count_app : forall `{Vars A} {x} {as1 as2 : list A},
    count x (as1 ++ as2) = count x as1 + count x as2.
Proof. move=> *; rewrite /count /= varsE multiplicityE //. Qed.
Definition countE := (count_term, @count_pair, @count_coeq, @count_pexp,
                      @count_nil, @count_cons, @count_app).

(* TODO This should really be in Common *)
Instance fmap_f_proper {A B} {R : relation A} {S : relation B}
  : Proper ((R ==> S) ==> Forall2 R ==> Forall2 S) fmap.
Proof.
  move=> f f' Proper_f.
  elim.
  - move=> l; inversion 1; constructor.
  - move=> x l IH l'.
    inversion 1; rewrite !fmap_cons; constructor; auto.
Qed.
Instance forall2L : @LeibnizEquiv (list A) (Forall2 eq).
Proof.
  move=> A; elim.
  - move=> l; by inversion 1.
  - move=> x l IH l'; inversion 1; firstorder.
Qed.

Instance varsc_proper : Proper ((≡) ==> (=)) (vars (A:=coeq)).
Proof. move=> [t u] _ [] //=; apply: comm. Qed.
Instance: Proper ((≡) ==> (=)) (vars (A:=soln)).
Proof.
  rewrite soln_equiv_alt_alt => Θ Θ'' [Θ' Perm E].
  etransitivity.
  - apply: (foldr_permutation_proper' eq).
    rewrite Perm //.
  - rewrite (forall2L _ _ (fmap_f_proper _ _ varsc_proper _ _ E)) //.
Qed.
Instance: Proper ((≡) ==> (=)) (vars (A:=pexp)).
Proof. move=> [Θ ts] [Ξ us] [/= -> ->] //. Qed.
Instance countc_proper : Proper ((=) ==> (≡) ==> (=)) (count (A:=coeq)).
Proof. move=> ? ? <- ? ? E; rewrite /count E //. Qed.
Instance: Proper ((=) ==> (≡) ==> (=)) (count (A:=soln)).
Proof. move=> ? ? <- ? ? E; rewrite /count E //. Qed.
Instance: Proper ((=) ==> (≡) ==> (=)) (count (A:=pexp)).
Proof. move=> ? ? <- ? ? E; rewrite /count E //. Qed.

Lemma sim_subst_var : forall {σ x}, sim_subst σ (var x) = default (var x) (σ !! x).
Proof. done. Qed.
Lemma subst_var_eq : forall {t x}, subst t x (var x) = t.
Proof. move=> *; rewrite /sim_subst /=; clear_sing. Qed.
Lemma subst_var_ne : forall {t x x'}, x <> x' -> subst t x (var x') = var x'.
Proof. move=> *; rewrite /sim_subst /=; clear_sing. Qed.
Lemma subst_var : forall {t x x'},
    subst t x (var x') = if decide (x = x') is left _ then t else var x'.
Proof. move=> *; rewrite /sim_subst /= lookup_singleton'; by case: (decide _). Qed.
Lemma sim_subst_star : forall {σ}, sim_subst σ star = star.
Proof. done. Qed.
Lemma sim_subst_costar : forall {σ}, sim_subst σ costar = costar. Proof. done. Qed.
Lemma sim_subst_ttens : forall {σ t u},
    sim_subst σ (ttens t u) = ttens (sim_subst σ t) (sim_subst σ u).
Proof. done. Qed.
Lemma sim_subst_tpar : forall {σ t u},
    sim_subst σ (tpar t u) = tpar (sim_subst σ t) (sim_subst σ u).
Proof. done. Qed.

Lemma sim_substpr_def : forall `{Subst A, Subst B} σ (t : A) (u : B),
    sim_subst σ (t, u) = (sim_subst σ t, sim_subst σ u).
Proof. done. Qed.
Lemma sim_substc_def : forall σ t u, sim_subst σ (Coeq t u) = Coeq (sim_subst σ t) (sim_subst σ u).
Proof. done. Qed.
Definition sim_substl_nil `{Subst A} σ : sim_subst (A:=list A) σ [] = [] := erefl.
Lemma sim_substl_cons : forall `{Subst A} σ (a : A) as_,
    sim_subst σ (a :: as_) = sim_subst σ a :: sim_subst σ as_.
Proof. reflexivity. Qed.
Lemma sim_substpe_def : forall σ Θ ts,
    sim_subst σ (Pexp Θ ts) = Pexp (sim_subst σ Θ) (sim_subst σ ts).
Proof. done. Qed.
Lemma sim_substl_app : forall `{Subst A} σ (ts us : list A),
    sim_subst σ (ts ++ us) = sim_subst σ ts ++ sim_subst σ us.
Proof. move=> *. apply: fmap_app. Qed.

Definition substE := (@subst_var_eq, @subst_var, (* @sim_subst_var*) @sim_subst_star,
                      @sim_subst_costar, @sim_subst_ttens, @sim_subst_tpar,
                      sim_substc_def, @sim_substpr_def, @sim_substl_nil, @sim_substl_cons,
                      sim_substpe_def, @sim_substl_app).

Instance sim_substc_proper : Proper ((=) ==> (≡) ==> (≡)) (sim_subst (A:=coeq)).
Proof. move=> ? _ <- [t' u] _ [] //=; constructor. Qed.
Instance sim_substl_proper `{Subst A}
  : Proper ((=) ==> (≡ₚ) ==> (≡ₚ)) (sim_subst (A:=list A)).
Proof. move=> ? _ <- t x; apply: fmap_Permutation. Qed.
Instance sim_substcl_proper : Proper ((=) ==> (≡) ==> (≡)) (sim_subst (A:=soln)).
Proof.
  rewrite soln_equiv_alt_alt => ? _ <- Θ Θ'' [Θ' Swaps Syms].
  eexists.
  - rewrite Swaps //.
  - by apply/fmap_f_proper; first apply: sim_substc_proper.
Qed.
Instance sim_substpe_proper : Proper ((=) ==> (≡) ==> (≡)) (sim_subst (A:=pexp)).
Proof. move=> ? _ <- [? ?] [? ?] [/= E1 E2]; rewrite substE E1 E2 //. Qed.

Lemma not_elem_of_multiplicity : forall (x : name) X, x ∉ X <-> multiplicity x X = 0.
Proof.
  move=> x X.
  rewrite elem_of_multiplicity -Nat.le_ngt Nat.le_0_r //.
Qed.
Lemma elem_of_count : forall `{Vars A} (x : name) (a : A), x ∈ vars a <-> count x a > 0.
Proof. move=> *; apply: elem_of_multiplicity. Qed.
Lemma not_elem_of_count : forall `{Vars A} (x : name) (a : A), x ∉ vars a <-> x # a.
Proof. move=> *; apply: not_elem_of_multiplicity. Qed.

Lemma elem_of_vars : forall `{Vars A} {a : A} {l : list A}, a ∈ l -> vars a ⊆ vars l.
Proof.
  move=> ? ? t l /(elem_of_list_split _ _) [l1 [l2 ->]].
  rewrite !varsE [vars t ⊎ vars l2]comm assoc.
  apply: gmultiset_disj_union_subseteq_r.
Qed.
Lemma elem_mult : forall `{Vars A} {x a} {l : list A}, a ∈ l -> count x a <= count x l.
Proof. move=> *. by apply: elem_of_vars. Qed.
Lemma elem_mult0 : forall `{Vars A} {x a} {l : list A}, x # l -> a ∈ l -> x # a.
Proof. move=> ? ? x a l NotIn /(elem_mult (x:=x)); rewrite NotIn; lia. Qed.

Definition subst_fresh `{Vars A} (σ : substitution) (a : A) : Prop :=
  map_Forall (fun x _ => x # a) σ.
Lemma sim_subst_nop : forall σ (u : term), subst_fresh σ u -> sim_subst σ u = u.
Proof.
  move=> σ u.
  rewrite /subst_fresh /map_Forall.
  setoid_rewrite <- not_elem_of_count.
  elim: u.
  - move=> /= x'; setoid_rewrite not_elem_of_singleton => H.
    rewrite sim_subst_var; case E: (σ !! x'); firstorder.
  - auto.
  - auto.
  - move=> t IH_t u IH_u; setoid_rewrite gmultiset_elem_of_disj_union => H.
    rewrite substE IH_t; last rewrite IH_u //; naive_solver.
  - move=> /= t IH_t u IH_u; setoid_rewrite gmultiset_elem_of_disj_union => H.
    rewrite substE IH_t; last rewrite IH_u //; naive_solver.
Qed.
Lemma subst_nop : forall t x (u : term), x # u -> subst t x u = u.
Proof. move=> *; by apply/sim_subst_nop/map_Forall_insert. Qed.
Lemma sim_substc_nop : forall σ (C : coeq), subst_fresh σ C -> sim_subst σ C = C.
Proof.
  move=> σ [t u] SF; rewrite substE !sim_subst_nop //;
    move=> x ? /SF; rewrite countE; lia.
Qed.
Lemma substc_nop : forall t x (C : coeq), x # C -> subst t x C = C.
Proof. move=> *; by apply/sim_substc_nop/map_Forall_insert. Qed.
Lemma sim_substl_nop : forall σ (us : list term), subst_fresh σ us -> sim_subst σ us = us.
Proof.
  move=> σ.
  elim.
  - auto.
  - cbn; move=> u us IH SF.
    rewrite substE sim_subst_nop; last rewrite IH //;
      move=> x ? /SF; rewrite countE; lia.
Qed.
Lemma substl_nop : forall t x (us : list term), x # us -> subst t x us = us.
Proof. move=> *; by apply/sim_substl_nop/map_Forall_insert. Qed.
Lemma sim_substcl_nop : forall σ (Θ : soln), subst_fresh σ Θ -> sim_subst σ Θ = Θ.
Proof.
  move=> σ.
  elim.
  - auto.
  - cbn; move=> C Θ IH SF.
    rewrite substE sim_substc_nop; last rewrite IH //;
      move=> x ? /SF; rewrite countE; lia.
Qed.
Lemma substcl_nop : forall t x (Θ : soln), x # Θ -> subst t x Θ = Θ.
Proof. move=> *; by apply/sim_substcl_nop/map_Forall_insert. Qed.
Lemma sim_substpe_nop : forall σ (p : pexp), subst_fresh σ p -> sim_subst σ p = p.
Proof.
  move=> σ [Θ ts] SF; rewrite substE sim_substcl_nop; last rewrite sim_substl_nop //;
    move=> x ? /SF; rewrite countE; lia.
Qed.
Lemma substpe_nop : forall t x (p : pexp), x # p -> subst t x p = p.
Proof. move=> *; by apply/sim_substpe_nop/map_Forall_insert. Qed.

Definition id_subst (σ : substitution) : Prop :=
  map_Forall (fun x a => a = var x) σ.
Lemma sim_subst_id : forall {σ} {u : term}, id_subst σ -> sim_subst σ u = u.
Proof.
  move=> σ u IS; elim: u => *; rewrite ?substE ?sim_subst_var; try congruence.
  case E: (σ !! _); auto.
Qed.
Lemma subst_id : forall {x} {u : term}, subst (var x) x u = u.
Proof. move=> *; by apply/sim_subst_id/map_Forall_insert. Qed.
Lemma sim_substc_id : forall {σ} {C : coeq}, id_subst σ -> sim_subst σ C = C.
Proof.
  move=> σ [t u] SF; rewrite substE !sim_subst_id //;
    move=> x ? /SF; rewrite countE; lia.
Qed.
Lemma substc_id : forall {x} {C : coeq}, subst (var x) x C = C.
Proof. move=> *; by apply/sim_substc_id/map_Forall_insert. Qed.
Lemma sim_substl_id : forall σ (us : list term), id_subst σ -> sim_subst σ us = us.
Proof.
  move=> σ.
  elim.
  - auto.
  - cbn; move=> u us IH IS.
    rewrite substE sim_subst_id; last rewrite IH //; firstorder.
Qed.
Lemma substl_id : forall x (us : list term), subst (var x) x us = us.
Proof. move=> *; by apply/sim_substl_id/map_Forall_insert. Qed.
Lemma sim_substcl_id : forall σ (Θ : soln), id_subst σ -> sim_subst σ Θ = Θ.
Proof.
  move=> σ.
  elim.
  - auto.
  - cbn; move=> C Θ IH IS.
    rewrite substE sim_substc_id; last rewrite IH //; firstorder.
Qed.
Lemma substcl_id : forall x (Θ : soln), subst (var x) x Θ = Θ.
Proof. move=> *; by apply/sim_substcl_id/map_Forall_insert. Qed.
Lemma sim_substpe_id : forall σ (p : pexp), id_subst σ -> sim_subst σ p = p.
Proof.
  move=> σ [Θ ts] IS; rewrite substE sim_substcl_id; last rewrite sim_substl_id //; firstorder.
Qed.
Lemma substpe_id : forall x (p : pexp), subst (var x) x p = p.
Proof. move=> *; by apply/sim_substpe_id/map_Forall_insert. Qed.

Lemma count_subst : forall {t x x'} {u : term},
    count x' (subst t x u) =
    (if decide (x' = x) is left _ then 0 else count x' u) +
    count x' t * count x u.
Proof.
  move=> t x x'.
  elim.
  - move=> y.
    rewrite substE !countE.
    case: (decide (x = y)) => [<- | NE] /=.
    + case: (decide _); lia.
    + rewrite count_var.
      case: (decide (x' = y)) => [-> | NE'] /=.
      * rewrite decide_False //; lia.
      * case: (decide _); lia.
  - rewrite /= !countE; case: (decide _) => //=.
  - rewrite /= !countE; case: (decide _) => //=.
  - case: (decide (x' = x)) => /= [-> | NE] l IH_l r IH_r /=; rewrite !countE IH_l IH_r; lia.
  - case: (decide (x' = x)) => /= [-> | NE] l IH_l r IH_r /=; rewrite !countE IH_l IH_r; lia.
Qed.

Lemma count_substl : forall {t x x'} {ts : list term},
    count x' (subst t x ts) =
    (if decide (x' = x) is left _ then 0 else count x' ts) +
    count x' t * count x ts.
Proof.
  move=> t x x'.
  elim.
  - rewrite /= !countE; case: (decide _) => //=.
  - move=> t' ts IH.
    rewrite substE !countE IH count_subst.
    case: (decide (x' = x)) => /=; lia.
Qed.
Lemma count_substcl : forall {t x x'} {Θ : soln},
    count x' (subst t x Θ) =
    (if decide (x' = x) is left _ then 0 else count x' Θ) +
    count x' t * count x Θ.
Proof.
  move=> t x x'.
  elim.
  - rewrite /= !countE; case: (decide _) => //=.
  - move=> [l r] Θ IH.
    rewrite substE !countE IH !count_subst.
    case: (decide (x' = x)) => /=; lia.
Qed.

Lemma subst_is_var : forall {t x u y},
    subst t x u = var y <-> u = var x /\ t = var y \/ u = var y /\ x <> y.
Proof.
  move=> t x u y.
  split.
  - case: u => //= x'.
    rewrite substE; case: (decide _) => [-> | NE] /=; firstorder.
  - case=> [[-> ->] | [-> NE]] /=; rewrite substE // decide_False //.
Qed.

Lemma subst_subst1 : forall {t t' x} (u : term),
    subst t x (subst t' x u) = subst (subst t x t') x u.
Proof.
  move=> t t' x.
  elim; try solve [move=> *; rewrite !substE; firstorder].
  move=> y; rewrite !substE.
  case: (decide (x = y)) => //= ?; rewrite substE decide_False //.
Qed.

Lemma subst_subst2 : forall {t t' : term} {x x'},
    x <> x' -> x' # t ->
    forall u : term, subst t x (subst t' x' u) = subst (subst t x t') x' (subst t x u).
Proof.
  move=> t t' x x' NE0 NotIn.
  elim=> *; rewrite !substE; try congruence.
  case: (decide _) => [E | NE]; case: (decide _) => [E' | NE']; subst; rewrite ?substE //.
  - rewrite subst_nop //.
  - rewrite !decide_False //.
Qed.

Lemma subst_subst3 : forall {t t' : term} {x x'},
    x <> x' -> x' # t -> x # t' ->
    forall u : term, subst t x (subst t' x' u) = subst t' x' (subst t x u).
Proof.
  move=> t t' x x' NE0 NotIn NotIn'.
  elim=> *; rewrite !substE; try congruence.
  case: (decide _) => [E | NE]; case: (decide _) => [E' | NE']; subst; rewrite ?substE //.
  - rewrite subst_nop //.
  - rewrite subst_nop //.
  - rewrite !decide_False //.
Qed.

Lemma substc_substc2 : forall {t t' : term} {x x'},
    x <> x' -> x' # t ->
    forall C : coeq, subst t x (subst t' x' C) = subst (subst t x t') x' (subst t x C).
Proof.
  move=> t t' x x' NE0 NotIn [? ?] /=.
  rewrite !substE subst_subst2 // [u in Coeq _ u]subst_subst2 //.
Qed.

Lemma substc_substc3 : forall {t t' : term} {x x'},
    x <> x' -> x' # t -> x # t' ->
    forall C : coeq, subst t x (subst t' x' C) = subst t' x' (subst t x C).
Proof.
  move=> t t' x x' NE0 NotIn NotIn' [? ?].
  rewrite !substE; do 2 rewrite [subst t x (subst _ _ _)]subst_subst3 //.
Qed.

Inductive have_tys : soln -> list (term * ty) -> Prop :=
| ax : forall {x A}, have_tys [] [(var x, neg A); (var x, A)]
| xchg_coeq : forall {Θ Θ' Γ}, Θ ≡ₚ Θ' -> have_tys Θ Γ -> have_tys Θ' Γ
| xchg_term : forall {Θ Γ Γ'}, Γ ≡ₚ Γ' -> have_tys Θ Γ -> have_tys Θ Γ'
| cut : forall {Θ Ξ Γ Δ A t u},
    vars Θ ∪ vars t ∪ vars Γ.*1 ## vars Ξ ∪ vars u ∪ vars Δ.*1 ->
    have_tys Θ ((t, A) :: Γ) -> have_tys Ξ ((u, neg A) :: Δ) ->
    have_tys (Coeq t u :: Θ ++ Ξ) (Γ ++ Δ)

| _one : have_tys [] [(star, one)]
| _perp : forall {Θ Γ}, have_tys Θ Γ -> have_tys Θ ((costar, perp) :: Γ)
| _tens : forall {Θ Ξ Γ Δ A B t u},
    vars Θ ∪ vars t ∪ vars Γ.*1 ## vars Ξ ∪ vars u ∪ vars Δ.*1 ->
    have_tys Θ ((t, A) :: Γ) -> have_tys Ξ ((u, B) :: Δ) ->
    have_tys (Θ ++ Ξ) ((ttens t u, tens A B) :: Γ ++ Δ)
| _par : forall {Θ Γ A B t u}, have_tys Θ ((t, A) :: (u, B) :: Γ) ->
                          have_tys Θ ((tpar t u, par A B) :: Γ).
Hint Constructors have_tys.
Definition sequent : Type := list ty.
Definition has_ty : pexp -> sequent -> Prop :=
  fun '(Pexp Θ ts) Seq => length ts = length Seq /\ have_tys Θ (zip ts Seq).

Example idfun_ty : has_ty (Pexp [] [tpar (var 1) (var 1)]%positive)
                          [lto (atom 1) (atom 1)]%positive.
Proof. repeat constructor. Qed.

Instance: Proper ((≡ₚ) ==> (≡ₚ) ==> iff) have_tys.
Proof.
  apply proper_sym_impl_iff_2; try typeclasses eauto.
  move=> Θ Θ' E_Θ Γ Γ' E_Γ Typing.
    by apply/(xchg_coeq E_Θ)/(xchg_term E_Γ).
Qed.
Instance: Proper ((≡) ==> (≡ₚ) ==> iff) have_tys.
Proof.
  apply proper_sym_impl_iff_2; try typeclasses eauto.
  move=> Θ0 Θ' /soln_equiv_alt [Θ /swaps_alt Perm Syms] Γ _ <- Typing.
  move: Typing Θ' Syms; rewrite {}Perm {Θ0}; elim: Θ Γ /.
  - move=> x A Θ' /ees_nil1 -> //.
  - move=> Θ Θ' Γ Perm Typing IH Θ'' Syms.
    move: (Perm >>> Syms) => /ees_perm [Θ'2 Syms' <-].
      by apply: IH.
  - move=> Θ Γ Γ' Perm Typing IH Θ' /IH Typing'.
      by rewrite -Perm.
  - move=> Θ Ξ Γ Δ A t u Disj Typing_Γ IH_Γ Typing_Δ IH_Δ Θ'
            /syms_alt /(Forall2_cons_inv_l _ _) [C'] [Θ''] [E]
            [/Forall2_app_inv_l [Θ''1] [Θ''2] [/syms_alt Syms1] [/syms_alt Syms2] ->] ->.
    case: E => /=.
    + apply: cut; only 2,3: by auto.
      rewrite -(syms_equiv Syms1) -(syms_equiv Syms2) //.
    + rewrite Permutation_app_comm [Γ ++ Δ]Permutation_app_comm.
      apply: cut.
      * rewrite -(syms_equiv Syms1) -(syms_equiv Syms2) //.
      * auto.
      * rewrite neg_involution; auto.
  - move=> Θ' /ees_nil1 -> //.
  - auto.
  - move=> Θ Ξ Γ Δ A B t u Disj Typing_Γ IH_Γ Typing_Δ IH_Δ Θ' /syms_alt /Forall2_app_inv_l
            [Θ'1] [Θ'2] [/syms_alt Syms1] [/syms_alt Syms2] ->.
    apply: _tens; only 2,3: by auto.
    rewrite -(syms_equiv Syms1) -(syms_equiv Syms2) //.
  - auto.
Qed.

Instance: Proper ((≡) ==> (=) ==> iff) has_ty.
Proof. move=> [? ?] [? ?] [/= E1 E2] ? ? E3. rewrite E1 E2 E3 //. Qed.

Definition linear (p : pexp) : Prop :=
  forall x, x # p \/ count x p = 2.
Theorem well_typed_linear : forall (p : pexp) (Seq : sequent), has_ty p Seq -> linear p.
Proof.
  move=> [Θ ts] Seq /= [E_length Typing] x.
  rewrite /= -(fst_zip ts Seq (Nat.eq_le_incl _ _ E_length)).
  move: {E_length Seq ts} (zip ts Seq) Typing => Γ.
  elim: Θ Γ /.
  - move=> x' A.
    rewrite !countE; case: (decide (x = x')); auto.
  - move=> Θ Θ' Γ -> //.
  - move=> Θ Γ Γ' E; rewrite !countE E //.
  - move=> Θ Ξ Γ Δ A t u Disj Typing_Γ IH_Γ Typing_Δ IH_Δ.
    move: IH_Γ IH_Δ.
    cbn; rewrite !fmap_app /= !countE => -[] H [] H'; try lia.
    move/elem_of_disjoint/(_ x) in Disj.
    case: Disj; rewrite !elem_of_union -!gmultiset_elem_of_disj_union
                        elem_of_multiplicity !multiplicity_disj_union -plus_assoc;
    [rewrite H | rewrite H']; auto.
  - auto.
  - move=> Θ Γ Typing; cbn; rewrite !countE //.
  - move=> Θ Ξ Γ Δ A B t u Disj Typing_Γ IH_Γ Typing_Δ IH_Δ.
    move: IH_Γ IH_Δ.
    cbn; rewrite !fmap_app !countE => -[] H [] H'; try lia.
    move/elem_of_disjoint/(_ x) in Disj.
    case: Disj; rewrite !elem_of_union -!gmultiset_elem_of_disj_union
                        elem_of_multiplicity !multiplicity_disj_union -plus_assoc;
    [rewrite H | rewrite H']; auto.
  - move=> Θ Γ A B t u Typing; cbn; rewrite !countE; lia.
Qed.

Instance: Proper ((≡) ==> iff) linear.
Proof.
  move=> [? ?] [? ?] [/= SolnEquiv ->].
  rewrite /linear; by setoid_rewrite SolnEquiv.
Qed.

Inductive reaction : ll_relation coeq :=
| communication : forall t x u, reaction [Coeq t (var x); Coeq (var x) u] [Coeq t u]
| unit_react : reaction [Coeq star costar] []
| pair_react : forall t u t' u', reaction [Coeq (ttens t u) (tpar t' u')] [Coeq t t'; Coeq u u'].
Hint Constructors reaction.

Inductive basic_step : relation pexp :=
| in_soln : forall Θ Θ' ts, ll_contextual reaction Θ Θ' -> basic_step (Pexp Θ ts) (Pexp Θ' ts)
| cleanup : forall x t Θ ts,
    x ∈ vars ts ->
    basic_step (Pexp (Coeq (var x) t :: Θ) ts) (Pexp Θ (subst t x ts)).
Hint Constructors basic_step.
Definition step : relation pexp := (≡) >> (basic_step >> (≡)).
Definition psteps : relation pexp := tc step.
Definition steps : relation pexp := (≡) ∪ᵣ psteps.

Lemma equiv_steps : forall {p q}, p ≡ q -> steps p q.
Proof. move=> *; by apply/rel_inl. Qed.
Lemma step_steps : forall {p q}, step p q -> steps p q.
Proof. move=> *; by apply/rel_inr/tc_once. Qed.

Instance: Proper ((≡) ==> (≡) ==> iff) step.
Proof.
  apply proper_sym_impl_iff_2; try typeclasses eauto.
  move=> p p' E1 q q' E2 [p_ E_ [_q BasicStep _E]].
  eexists; first by transitivity p.
  eexists; last by transitivity q.
  done.
Qed.
Instance: Proper ((≡) ==> (≡) ==> iff) psteps.
Proof.
  apply proper_sym_impl_iff_2; try typeclasses eauto.
  move=> p q E1 p' q' E2 PSteps.
  elim: p p' / PSteps q q' E1 E2.
  - move=> p p' Step q q' E1 E2.
    move: Step.
    rewrite E1 E2.
    apply: tc_once.
  - move=> p p' p'' Step PSteps IH q q' E1 E2.
    apply: tc_l.
    + rewrite -E1 //.
    + apply: IH; auto.
Qed.
Instance: Proper ((≡) ==> (≡) ==> iff) steps.
Proof.
  move=> p q E1 p' q' E2.
  rewrite /steps /relation_disjunction /predicate_union /= E1 E2 //.
Qed.

Instance steps_preorder : PreOrder steps.
Proof.
  constructor.
  - move=> p; left; reflexivity.
  - move=> p q r [E1 | PSteps1] [E2 | PSteps2].
    + left; rewrite E1 //.
    + right; rewrite E1 //.
    + right; rewrite -E2 //.
    + right; by etransitivity.
Qed.

Instance linear_step : Proper (step ==> impl) linear.
Proof.
  move=> _ _ [p -> [q Basic <-]].
  case: p q / Basic.
  - move=> _ _ ts [Θ1 _ _ Θ2 [t x u | | t u t' u']] /= Linear x';
            move: (Linear x'); rewrite !countE; lia.
  - move=> x t Θ ts /elem_of_count Reactive Linear x'.
    move: (Linear x) (Linear x').
    rewrite !countE count_substl => ?.
    have -> : count x ts = 1 by lia.
    case: (decide (x' = x)) => [-> | NE]; lia.
Qed.

Instance linear_steps : Proper (steps ==> impl) linear.
Proof.
  move=> p q [-> | PSteps] // Linear.
  elim: p q / PSteps Linear => [? ? /linear_step // | p q r /linear_step].
  auto.
Qed.

Definition cut_free (p : pexp) : Prop := let 'Pexp coeqs _ := p in coeqs = [].

Inductive reverse_typing1 : el_relation (term * ty) :=
| reverse_perp : reverse_typing1 (costar, perp) []
| reverse_par : forall t u A B, reverse_typing1 (tpar t u, par A B) [(t, A); (u, B)].
Definition reverse_typing : ll_relation (term * ty) := el_contextual reverse_typing1.

Lemma reverse_typing_vars : forall {Γ Λ}, reverse_typing Γ Λ -> vars Γ.*1 = vars Λ.*1.
Proof.
  move=> Γ Λ [Γ1 _ _ Γ2 [| t u A B]];
          rewrite !fmap_app /= !varsE /= ?left_id -?assoc //.
Qed.

Theorem reversible : Proper ((=) ==> reverse_typing ==> impl) have_tys.
Proof.
  move=> Θ _ <- Γ Λ Rev Typing.
  elim: Θ Γ / Typing Λ Rev.
  - move=> x A Λ /(@el_app1 _ _ [_] [_]) [[? _ /el_singleton] | [? _ /el_singleton]];
            inversion 1.
  - move=> Θ Θ' Γ Perm Typing IH Λ.
    rewrite -{}Perm; apply: IH.
  - move=> Θ Γ Γ' Perm Typing IH Λ Ex.
    move: (Perm >>> Ex) => /el_perm [Λ' Ex' <-].
      by apply: IH.
  - move=> Θ Ξ Γ Δ A t u Disj Typing_Γ IH_Γ Typing_Δ IH_Δ Λ /el_app1 [[Γ' -> Ex] | [Δ' -> Ex]].
    + apply: cut.
      * rewrite -(reverse_typing_vars Ex) //.
      * by apply/IH_Γ/cons_el2.
      * done.
    + apply: cut.
      * rewrite -(reverse_typing_vars Ex) //.
      * done.
      * by apply/IH_Δ/cons_el2.
  - move=> Λ /el_singleton; inversion 1.
  - move=> Θ Γ Typing IH Λ /el_cons [[? ->] | [Λ' -> Ex]]; last by auto.
      by inversion 1.
  - move=> Θ Ξ Γ Δ A B t u Disj Typing_Γ IH_Γ Typing_Δ IH_Δ Λ /el_cons [[? ->] | [Λ' ->]];
            first by inversion 1.
    move=> /el_app1 [[Γ' -> Ex] | [Δ' -> Ex]].
    + apply: _tens.
      * rewrite -(reverse_typing_vars Ex) //.
      * by apply/IH_Γ/cons_el2.
      * done.
    + apply: _tens.
      * rewrite -(reverse_typing_vars Ex) //.
      * done.
      * by apply/IH_Δ/cons_el2.
  - move=> Θ Γ A B t u Typing IH Λ /el_cons [[? ->] | [Λ' -> Ex]].
    + by inversion 1.
    + by apply/_par/IH/cons_el2/cons_el2.
Qed.
Existing Instance reversible.

Fixpoint weak_term_has_ty (t : term) (T : ty) : Prop :=
  (exists x, t = var x) \/
  match T with
  | atom _ => False
  | neg_atom _ => False
  | one => t = star
  | perp => t = costar
  | tens A B => exists l r, weak_term_has_ty l A /\ weak_term_has_ty r B /\ t = ttens l r
  | par A B => exists l r, weak_term_has_ty l A /\ weak_term_has_ty r B /\ t = tpar l r
  end.
Definition weak_matched (C : coeq) : Prop := exists A,
    weak_term_has_ty (lhs C) A /\ weak_term_has_ty (rhs C) (neg A).
(* TODO maybe include some notion of "instances of a variable having dual types" *)
Definition weak_have_tys (Θ : soln) (Γ : list (term * ty)) : Prop :=
  Forall weak_matched Θ /\ Forall (curry weak_term_has_ty) Γ.
Definition weak_has_ty : pexp -> sequent -> Prop :=
  fun '(Pexp Θ ts) Seq =>
    length ts = length Seq /\ weak_have_tys Θ (zip ts Seq).

Lemma wtht_expand : forall {t T},
    weak_term_has_ty t T <->
    (exists x, t = var x) \/
    match T with
    | atom _ => False
    | neg_atom _ => False
    | one => t = star
    | perp => t = costar
    | tens A B => exists l r, weak_term_has_ty l A /\ weak_term_has_ty r B /\ t = ttens l r
    | par A B => exists l r, weak_term_has_ty l A /\ weak_term_has_ty r B /\ t = tpar l r
    end.
Proof. move=> t [] //. Qed.

Theorem weak_typing_inversion : forall {Θ Γ}, have_tys Θ Γ -> weak_have_tys Θ Γ.
Proof.
  move=> Θ Γ; elim: Θ Γ /.
  - move=> x [] * /=; repeat econstructor.
  - rewrite /weak_have_tys => Θ _ Γ <- //.
  - rewrite /weak_have_tys => Θ Γ _ <- //.
  - move=> Θ Ξ Γ Δ A t u Disj _ [Matched_Θ /Forall_cons /= [WTHT_t WTHT_Γ]]
                             _ [Matched_Ξ /Forall_cons /= [WTHT_u WTHT_Δ]].
    split.
    + constructor.
      * eexists; eauto.
      * by apply/Forall_app.
    + by apply/Forall_app.
  - by repeat constructor.
  - move=> Θ Γ _ [Matched_Θ WTHT_Γ].
      by repeat econstructor.
  - move=> Θ Ξ Γ Δ A B t u Disj _ [Matched_Θ /Forall_cons /= [WTHT_t WTHT_Γ]]
                               _ [Matched_Ξ /Forall_cons /= [WTHT_u WTHT_Δ]].
    split.
    + by apply/Forall_app.
    + constructor.
      * right; eauto.
      * by apply/Forall_app.
  - move=> Θ Γ A B t u _ [Matched_Θ /Forall_cons /= [WTHT_t /Forall_cons /= [WTHT_u WTHT_Γ]]].
      by repeat econstructor.
Qed.
Corollary weak_typing_inversion' : forall {p Seq}, has_ty p Seq -> weak_has_ty p Seq.
Proof.
  move=> [Θ ts] /= Seq [Len Typing]; split.
  - done.
  - by apply: weak_typing_inversion.
Qed.
