Require Import VST.floyd.proofauto.
Require Import VST.floyd.library.
Require Import CPE2.
Require Import VST.msl.ghost.
Require Import VST.veric.own.
Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition Vprog : varspecs. mk_varspecs prog. Defined.

From Coq Require Import ssreflect ssrfun ssrbool.
Require Import stdpp.pmap.
Require Import stdpp.gmap.
Require Import Common.
Require Import PE2.
Require Import Transmission.

Open Scope logic.

(* TODO How much can I rely on enums being [tint]? *)

Definition Term : type := Tstruct _Term noattr.
Definition Node : type := tptr Term.
Definition LINK : val := Vint (Int.repr 0).
Definition UNIT : val := Vint (Int.repr 1).
Definition COUNIT : val := Vint (Int.repr 2).
Definition TENS : val := Vint (Int.repr 3).
Definition PAR : val := Vint (Int.repr 4).


Lemma Join_lower_None : forall {A} {a : A} {mx my mz},
    @sepalg.join _ (psepalg.Join_lower (sepalg_generators.Join_equiv A)) mx my mz ->
    mz = None <-> mx = None /\ my = None.
Proof.
  move=> A a mx my mz.
  case: mx my mz / => [| | ? ? ? []]; firstorder.
Qed.
Lemma Join_lower_Some : forall {A} {a : A} {mx my mz},
    @sepalg.join _ (psepalg.Join_lower (sepalg_generators.Join_equiv A)) mx my mz ->
    mz = Some a <-> mx = Some a \/ my = Some a.
Proof.
  move=> A a mx my mz.
  case: mx my mz / => [| | ? ? ? []]; intuition congruence.
Qed.
Lemma Join_lower_Some1 : forall {A} {x : A} {my mz},
    @sepalg.join _ (psepalg.Join_lower (sepalg_generators.Join_equiv A)) (Some x) my mz ->
    mz = Some x.
Proof. move=> A x my mz; inversion 1; firstorder. Qed.
Lemma Join_lower_Some2 : forall {A} {y : A} {mx mz},
    @sepalg.join _ (psepalg.Join_lower (sepalg_generators.Join_equiv A)) mx (Some y) mz ->
    mz = Some y.
Proof. move=> A x my mz; inversion 1; firstorder. Qed.
Lemma Join_lower_Some3 : forall {A} {z : A} {mx my},
    @sepalg.join _ (psepalg.Join_lower (sepalg_generators.Join_equiv A)) mx my (Some z) ->
    mx = Some z \/ my = Some z.
Proof. move=> A x my mz; inversion 1 as [| | ? ? ? []]; auto. Qed.

(* this is probably a misnomer *)
Definition affine (m : ShareMap.map address positive) : Prop := forall x adr1 adr2 adr3,
    ShareMap.map_val _ _ adr1 m = Some x -> ShareMap.map_val _ _ adr2 m = Some x ->
    ShareMap.map_val _ _ adr3 m = Some x ->
    [\/ adr1 = adr2, adr2 = adr3 | adr1 = adr3].
Program Instance link_naming_PCM : Ghost :=
  {| G := ShareMap.map address positive;
     (* TODO Should we add some kind of linearity constraint here? *)
     valid := affine;
     Join_G := ShareMap.Join_map _ _ (sepalg_generators.Join_equiv _);
     Sep_G := ShareMap.sa_map _ _ _;
     Perm_G := ShareMap.pa_map _ _ _ _ |}.
Next Obligation.
  move=> m1 m2 m /ShareMap.map_join_char J Aff x adr1 adr2 adr3 E1 E2 E3.
  move: E1 (J adr1) E2 (J adr2) E3 (J adr3) =>
    -> [_] /Join_lower_Some1 E1 -> [_] /Join_lower_Some1 E2 -> [_] /Join_lower_Some1 E3.
    by apply: Aff.
Qed.
Definition link_naming : Type := @G link_naming_PCM.
(* see examples/imp/hoare.v *)
Definition subsingleton_map (m : link_naming) (sh : share) (adr : address) (x : positive)
  : Prop :=
  ShareMap.map_share _ _ adr m = sh /\
  (forall adr' x', ShareMap.map_val _ _ adr' m = Some x' -> adr' = adr /\ x' = x).
  (* forall adr', adr' <> adr -> ShareMap.map_val _ _ adr' m  = None. *)
Definition named' (g : gname) (sh : share) (x : positive) (adr : address) : mpred :=
  EX m : link_naming, !! subsingleton_map m sh adr x &&
                      ghost_seplog.own g m compcert_rmaps.R.NoneP.
Definition named (g : gname) (sh : share) (x : positive) (p : val) : mpred :=
  EX adr : _, !! val2adr p adr && named' g sh x adr.

Lemma val2adr_inj : forall {p p' adr adr'},
    val2adr p adr -> val2adr p' adr' -> p = p' <-> adr = adr'.
Proof.
  move=> [] //= ? [? ?] [] //= ? [? ?] [? ?] [? ?] [-> ->] [-> ->].
  split=> [] [-> /Ptrofs.mkint_eq ->] //.
Qed.
Hint Resolve val2adr_inj.
Lemma val2adr_inj1 : forall {p adr adr'}, val2adr p adr -> val2adr p adr' -> adr = adr'.
Proof. move=> *; by apply/val2adr_inj. Qed.
Lemma val2adr_inj2 : forall {p p' adr}, val2adr p adr -> val2adr p' adr -> p = p'.
Proof. move=> *; by apply/val2adr_inj. Qed.
Hint Resolve val2adr_inj1 val2adr_inj2.
Lemma map_val_nonbot : forall {A B} {a : A} {m : ShareMap.map A B},
    ShareMap.map_share A B a m <> Share.bot <-> is_Some (ShareMap.map_val A B a m).
Proof.
  move=> A B a m.
  rewrite -ShareMap.map_val_bot.
  case E: (ShareMap.map_val _ _ _ _); firstorder eauto.
Qed.
Lemma subsingleton_map_exists : forall {sh adr x},
    exists m, subsingleton_map m sh adr x.
Proof.
  move=> sh adr x.
  case: (dec_share_nonunit sh) => [Non | /not_nonunit_bot ->].
  - exists (psepalg.insert_fpm _ adr (psepalg.mk_lifted sh Non, x) (psepalg.empty_fpm _ _)).
    split.
    + rewrite /ShareMap.map_share psepalg.fpm_gss //.
    + move=> adr' x' E; apply: dec_stable => /not_and_l_alt [NE | [NE E']];
        move: E; rewrite /ShareMap.map_val.
      * rewrite psepalg.fpm_gso //.
      * rewrite E' psepalg.fpm_gss => -[] //.
  - by exists (psepalg.empty_fpm _ _).
Qed.
Hint Resolve subsingleton_map_exists.
(*
Lemma singleton_map_nonidentity : forall {m sh adr x},
    singleton_map m sh adr x -> sepalg.nonidentity sh.
Proof.
  move=> m sh adr x [E1 [E2 U]] /identity_share_bot E_sh.
  move: E_sh E2 E1 => -> /ShareMap.map_val_bot -> //.
Qed.
 *)
Lemma subsingleton_map_valid : forall {m sh adr x},
    subsingleton_map m sh adr x -> valid m.
Proof. move=> m sh adr x [E U] y adr1 adr2 adr3 /U [-> ->] /U [-> _]; by constructor. Qed.
Hint Resolve subsingleton_map_valid.
(* this has gotta be possible to simplify *)
Lemma subsingleton_map_join : forall {m1 m2 m sh1 sh2 adr x},
    subsingleton_map m1 sh1 adr x -> subsingleton_map m2 sh2 adr x ->
    sepalg.join m1 m2 m <-> exists sh, sepalg.join sh1 sh2 sh /\ subsingleton_map m sh adr x.
Proof.
  move=> m1 m2 m sh1 sh2 adr x [E U] [E' U'].
  rewrite ShareMap.map_join_char /subsingleton_map; split.
  - move=> H.
    move: (H adr); rewrite E E' => -[J_sh _].
    eexists; repeat (split; first by auto).
    move=> adr' x' E_v; move: E_v (H adr') => -> -[_]; inversion 1 as [| | ? ? ? [] E1 E2].
    + by apply: U'.
    + by apply: U.
    + move: (U _ _ (esym E1)) (U' _ _ (esym E2)); firstorder.
  - move=> [sh [J [E'' U'']]] adr'.
    case E_y: (ShareMap.map_val _ _ adr' m) => [y |].
    + move: (U'' _ _ E_y) E E' E'' J (E_y) => [-> ->] <- <- <- J E_x.
      case E_y1: (ShareMap.map_val _ _ adr m1) => [y1 |];
       case E_y2: (ShareMap.map_val _ _ adr m2) => [y2 |].
      * move: (U _ _ E_y1) (U' _ _ E_y2) => [_ ->] [_ ->]; by split; last constructor.
      * move: (U _ _ E_y1) => [_ ->]; by split; last constructor.
      * move: (U' _ _ E_y2) => [_ ->]; by split; last constructor.
      * move: E_y1 E_y2 J E_x => /ShareMap.map_val_bot -> /ShareMap.map_val_bot ->
                                /bot_identity /esym /ShareMap.map_val_bot -> //.
    + move: E E' E'' E_y J => <- <- <- /ShareMap.map_val_bot.
      case E_y1: (ShareMap.map_val _ _ adr' m1) => [y1 |];
       case E_y2: (ShareMap.map_val _ _ adr' m2) => [y2 |].
      * move: (U _ _ E_y1) (E_y1) => [-> _] H ->
          /ghost_PCM.join_Bot [/ShareMap.map_val_bot NH _].
        move: H NH => -> //.
      * move: (U _ _ E_y1) (E_y1) => [-> _] H ->
          /ghost_PCM.join_Bot [/ShareMap.map_val_bot NH _].
        move: H NH => -> //.
      * move: (U' _ _ E_y2) (E_y2) => [-> _] H ->
          /ghost_PCM.join_Bot [_ /ShareMap.map_val_bot NH].
        move: H NH => -> //.
      * move: E_y1 E_y2 => /ShareMap.map_val_bot -> /ShareMap.map_val_bot -> -> _.
          by split; last constructor.
Qed.
Corollary subsingleton_map_join' : forall {m1 m2 m sh1 sh2 sh adr x},
    sepalg.join sh1 sh2 sh ->
    subsingleton_map m1 sh1 adr x -> subsingleton_map m2 sh2 adr x ->
    sepalg.join m1 m2 m <-> subsingleton_map m sh adr x.
Proof. move=> *; rewrite subsingleton_map_join //; firstorder. Qed.
Lemma subsingleton_map_join_exists : forall {m1 m2 sh1 sh2 sh adr x},
    sepalg.join sh1 sh2 sh ->
    subsingleton_map m1 sh1 adr x -> subsingleton_map m2 sh2 adr x ->
    exists m, sepalg.join m1 m2 m /\ subsingleton_map m sh adr x.
Proof.
  move=> m1 m2 sh1 sh2 sh adr x SSing1 SSing2 J.
  case: (@subsingleton_map_exists sh adr x) => [m SSing].
  exists m; rewrite subsingleton_map_join' //.
Qed.
Lemma named'_share_join : forall {g sh1 sh2 sh x adr},
    sepalg.join sh1 sh2 sh -> named' g sh1 x adr * named' g sh2 x adr = named' g sh x adr.
Proof.
  move=> g sh1 sh2 sh x adr J; rewrite /named'; apply: pred_ext.
  - do 2 (rewrite !(exp_sepcon1, exp_sepcon2); apply: exp_left; intros ?m).
    do 2 (rewrite ?(sepcon_andp_prop, sepcon_andp_prop'); apply: derives_extract_prop;
          intros ?SSing).
    move: (subsingleton_map_join_exists J SSing0 SSing) => [m' [J_m' SSing_m']].
    Exists m'; apply: andp_right; first by apply: prop_right.
    rewrite (own_op _ _ _ _ _ J_m'); apply: derives_refl.
  - apply: exp_left => m; apply: derives_extract_prop.
    move: (@subsingleton_map_exists sh1 adr x) (@subsingleton_map_exists sh2 adr x) =>
      [m1 SSing1] [m2 SSing2] /(subsingleton_map_join' J SSing1 SSing2) J_m.
    rewrite (own_op _ _ _ _ _ J_m); apply: sepcon_derives;
      EExists; (apply: andp_right; [by apply: prop_right | apply: derives_refl]).
Qed.
Lemma named_share_join : forall {g sh1 sh2 sh x p},
    sepalg.join sh1 sh2 sh -> named g sh1 x p * named g sh2 x p = named g sh x p.
Proof.
  move=> g sh1 sh2 sh x p J; rewrite /named; apply: pred_ext.
  - Intros adr1 adr2; EExists; entailer!; first by done.
    rewrite (val2adr_inj1 H H0) (named'_share_join (sh:=sh)) //.
  - Intros adr; EExists; EExists; entailer!; first by done.
    rewrite (named'_share_join (sh:=sh)) //.
Qed.
Hint Resolve named_share_join.
Lemma named'_collision : forall {g sh1 sh2 sh3 x adr1 adr2 adr3},
    sh1 <> Share.bot -> sh2 <> Share.bot -> sh3 <> Share.bot ->
    named' g sh1 x adr1 * named' g sh2 x adr2 * named' g sh3 x adr3 |--
      !! [\/ adr1 = adr2, adr2 = adr3 | adr1 = adr3].
Proof.
  move=> g sh1 sh2 sh3 x adr1 adr2 adr3 Non1 Non2 Non3.
  rewrite /named'; Intros m1 m3 m2; move: H H1 H0 => SSing1 SSing2 SSing3.
  assert_PROP (valid_2 m1 m2) as H. {
    apply: (derives_trans _ _ _ (sepcon_derives _ _ _ _ (ghost_valid_2 _ _ _ _)
                                                (TT_right _))).
    rewrite prop_sepcon; entailer.
  }
  move: H => [m [J _]]; rewrite -[L in L * _](ghost_op _ _ _ _ _ J).
  assert_PROP (valid_2 m m3) as H by apply: ghost_valid_2.
  move: H => [m' [J' Aff']].
  move: J J' => /ShareMap.map_join_char /all_and2 [_ J]
               /ShareMap.map_join_char /all_and2 [_ J'].
  move: SSing1 SSing2 SSing3 Non1 Non2 Non3 => [<- U] [<- U'] [<- U'']
    /map_val_nonbot [?] E1 /map_val_nonbot [?] E2 /map_val_nonbot [?] E3.
  move: E1 E2 E3 (E1) (E2) (E3) => /U [_ ->] /U' [_ ->] /U'' [_ ->] E1 E2 E3.
  move: E1 E2 E3 (J adr1) (J adr2) (J' adr1) (J' adr2) (J' adr3) => -> -> ->
    /Join_lower_Some1 -> /Join_lower_Some2 ->
    /Join_lower_Some1 E1 /Join_lower_Some1 E2 /Join_lower_Some2 E3.
    by apply/prop_right/Aff'.
Qed.
Lemma named_collision : forall {g sh1 sh2 sh3 x p1 p2 p3},
    sh1 <> Share.bot -> sh2 <> Share.bot -> sh3 <> Share.bot ->
    named g sh1 x p1 * named g sh2 x p2 * named g sh3 x p3 |--
      !! [\/ p1 = p2, p2 = p3 | p1 = p3].
Proof.
  move=> g sh1 sh2 sh3 x p1 p2 p3 Non1 Non2 Non3.
  rewrite /named; Intros adr1 adr2 adr3.
  rewrite (val2adr_inj H H1) (val2adr_inj H1 H0) (val2adr_inj H H0).
    by apply: named'_collision.
Qed.
Hint Resolve named_collision.

Definition Full_sh : share := Ews.
(* akin to conclib.v's gsh1 and gsh2 *)
Definition near := fst (slice.cleave Full_sh).
Definition far := snd (slice.cleave Full_sh).
Lemma near_readable : readable_share near.
Proof. apply: slice.cleave_readable1; auto. Qed.
Lemma far_readable : readable_share far.
Proof. apply: slice.cleave_readable2; auto. Qed.
Hint Resolve near_readable far_readable.
Lemma near_nonbot : near <> Share.bot.
Proof. move: bot_unreadable => H; contradict H; rewrite -H; auto. Qed.
Lemma far_nonbot : far <> Share.bot.
Proof. move: bot_unreadable => H; contradict H; rewrite -H; auto. Qed.
Hint Resolve near_nonbot far_nonbot.
(*
Definition Full_split (sh1 sh2 : share) : Prop :=
  (readable_share sh1 /\ readable_share sh2) /\ sepalg.join sh1 sh2 Full_sh.
*)

Notation mt := (malloc_token Full_sh Term).
Notation term_at := (data_at Full_sh Term).
Definition repr_var (g : gname) (x : positive) (p : val) : mpred := EX q : _,
  !! (q <> p) && named g near x p * named g far x q *
  data_at near Term (LINK, (q, nullval)) p * data_at far Term (LINK, (p, nullval)) q.
Fixpoint repr_term (g : gname) (t : term) (p : val) : mpred :=
  mt p *
  match t with
  | var x => repr_var g x p
  | star => term_at (UNIT, (nullval, nullval)) p
  | costar => term_at (COUNIT, (nullval, nullval)) p
  | ttens l r => EX p_t : _, EX p_u : _,
    term_at (TENS, (p_t, p_u)) p * repr_term g l p_t * repr_term g r p_u
  | tpar l r => EX p_t : _, EX p_u : _,
    term_at (PAR, (p_t, p_u)) p * repr_term g l p_t * repr_term g r p_u
  end.
(* Hmm, this repetition is ugly. *)
Definition repr_term' (g : gname) (t : term) (p : val) : mpred :=
  match t with
  | var x => repr_var g x p
  | star => term_at (UNIT, (nullval, nullval)) p
  | costar => term_at (COUNIT, (nullval, nullval)) p
  | ttens l r => EX p_t : _, EX p_u : _,
    term_at (TENS, (p_t, p_u)) p * repr_term g l p_t * repr_term g r p_u
  | tpar l r => EX p_t : _, EX p_u : _,
    term_at (PAR, (p_t, p_u)) p * repr_term g l p_t * repr_term g r p_u
  end.
Lemma get_mt (g : gname) (t : term) (p : val)
  : repr_term g t p = mt p * repr_term' g t p.
Proof. by case: t. Qed.
Definition repr_nonvar (g : gname) (t : term) (v : reptype Term) : mpred :=
  match t with
  | var x => FF
  | star => !! (v = (UNIT, (nullval, nullval))) && emp
  | costar => !! (v = (COUNIT, (nullval, nullval))) && emp
  | ttens l r =>
    !! (v.1 = TENS) && (repr_term g l v.2.1 * repr_term g r v.2.2)
  | tpar l r =>
    !! (v.1 = PAR) && (repr_term g l v.2.1 * repr_term g r v.2.2)
  end.
Lemma repr_term_var : forall {g x p}, repr_term g (var x) p = mt p * repr_var g x p.
Proof. done. Qed.
Definition is_nonvar (t : term) : Prop :=
  match t with
  | var _ => False
  | _ => True
  end.
Instance decide_is_nonvar {t} : Decision (is_nonvar t).
Proof. case: t => /=; typeclasses eauto. Defined.
Lemma repr_term_nonvar : forall {g t p},
    is_nonvar t -> repr_term g t p = mt p * EX v : _,
      term_at v p * repr_nonvar g t v.
Proof.
  move=> g [x | | | l r | l r] p //= _; apply/pred_ext.
  - EExists; entailer!.
  - entailer!.
  - EExists; entailer!.
  - entailer!.
  - Intros p_t p_u; EExists; by entailer!.
  - Intros v; destruct v as (? & p_t & p_u); simpl in *; subst.
    Exists p_t p_u; entailer!.
  - Intros p_t p_u; EExists; by entailer!.
  - Intros v; destruct v as (? & p_t & p_u); simpl in *; subst.
    Exists p_t p_u; entailer!.
Qed.
Lemma repr_term_alt : forall {g t p},
    repr_term g t p = mt p * ((EX x : _, !! (t = var x) && repr_var g x p) ||
                              !! is_nonvar t &&
                              (EX v : _, term_at v p * repr_nonvar g t v)).
Proof.
  move=> g t p; case: (decide (is_nonvar t)).
  - move=> NV; rewrite (repr_term_nonvar NV).
    rewrite true_eq // TT_andp.
    replace (EX _ : positive, _) with FF; first by rewrite FF_orp.
    apply/pred_ext; first by apply: FF_left.
    Intros x; by subst.
  - case: t => //= x _.
    f_equal; apply/pred_ext.
    + apply/orp_right1; Exists x; entailer!.
    + apply/orp_left; last by entailer!.
      Intros x'; inversion H; by subst.
Qed.
Definition tag_for (t : term) : val :=
  Vint (Int.repr match t with
                 | var _ => 0
                 | star => 1
                 | costar => 2
                 | ttens _ _ => 3
                 | tpar _ _ => 4
                 end).
Definition tag_for' (t : term) : val :=
  match t with
  | var _ => LINK
  | star => UNIT
  | costar => COUNIT
  | ttens _ _ => TENS
  | tpar _ _ => PAR
  end.
Lemma tag_for_fold (t : term) : tag_for t = tag_for' t.
Proof. case: t => //. Qed.
Lemma tag_for_tc (t : term) : tc_val tint (tag_for' t).
Proof. case: t => //. Qed.
Lemma tag_for_tc' (t : term) : is_int I32 Signed (tag_for' t).
Proof. case: t => //. Qed.
Hint Resolve tag_for_tc'.
Notation repr_tag t p := (field_at near Term (DOT _class) (tag_for t) p).
Lemma read_tag (g : gname) (t : term) (p : val)
  : repr_term g t p = repr_tag t p * (repr_tag t p -* repr_term g t p).
Proof.
  apply: pred_ext; last by apply: modus_ponens_wand.
  case: (decide (is_nonvar t)).
  - move=> /repr_term_nonvar ->.
    (Intros v); compute in v.
    case: v => Cl [v1 v2].
    assert_PROP (Cl = tag_for t) as E. {
      case: t => /= *; first by entailer!.
      all: Intros; try injection H; entailer!.
    }
    subst Cl.
    saturate_local.
    sep_apply data_at_field_at_cancel.
    rewrite -(field_at_share_join _ _ _ _ _ _ _ (slice.cleave_join _)) -/near -/far.
    sep_apply (gfield_ramif near Term [] (StructField _class)
                            (tag_for t, (v1, v2)) (tag_for t)).
    + by (exists erefl).
    + apply/field_compatible_cons; split; first compute; auto.
    + cancel; allp_left (tag_for t); rewrite -wand_sepcon_adjoint; cancel.
      (Exists (tag_for t, (v1, v2))); entailer!.
      rewrite /= sepcon_comm -sepcon_assoc 2!field_at_data_at_cancel'.
      sep_apply modus_ponens_wand.
      rewrite (data_at_share_join _ _ _ _ _ _ (slice.cleave_join _)); apply: derives_refl.
  - case: t => //= x _.
    rewrite /repr_var; Intros q.
    saturate_local.
    sep_apply data_at_field_at_cancel.
    sep_apply (gfield_ramif near Term [] (StructField _class) (LINK, (q, nullval)) LINK).
    + by (exists erefl).
    + apply/field_compatible_cons; split; first compute; auto.
    + cancel; allp_left LINK; rewrite -wand_sepcon_adjoint; cancel.
      Exists q; entailer!; rewrite sepcon_comm; apply: modus_ponens_wand.
Qed.

Definition repr_coeq (g : gname) (C : coeq) (pq : val * val) : mpred :=
  let '(Coeq t u) := C in let '(p, q) := pq in repr_term g t p * repr_term g u q.

Import iter_sepcon.
Fixpoint list_remove' `{EqDecision A} (x : A) (l : list A) : list A :=
  match l with
  | [] => []
  | y :: l' => if decide (x = y) is left _ then l' else y :: list_remove' x l'
  end.
Lemma list_remove'_alt `{EqDecision A} (x : A) (l : list A)
  : list_remove' x l = default l (list_remove x l).
Proof. elim: l => // a l /= ->; case: (decide _) => //; case: list_remove => //. Qed.
Fixpoint list_multiset_difference `{EqDecision A} (l1 l2 : list A) : list A :=
  match l2 with
  | [] => l1
  | a :: l2' => list_multiset_difference (list_remove' a l1) l2'
  end.
Lemma list_remove'_submseteq `{EqDecision A} (x : A) (l : list A)
  : list_remove' x l ⊆+ l.
Proof. elim: l => //= ? ? IH; case: (decide _); by constructor. Qed.
Instance list_remove'_mono `{EqDecision A}
  : Proper ((=) ==> submseteq ++> submseteq) (list_remove' (A:=A)).
Proof.
  move=> x _ <- l1 l2; elim: l1 l2 / => //=.
  - move=> y l1 l2 D; case: (decide _) => //=; by constructor.
  - move=> y1 y2 l.
    case: (decide _); case: (decide _) => //=.
    + move=> <- <- //.
    + constructor.
  - move=> y l1 l2 D IH.
    case: (decide _); last by constructor.
    move=> _; setoid_rewrite <- D; apply: list_remove'_submseteq.
  - by etransitivity.
Qed.
Instance list_remove'_proper `{EqDecision A}
  : Proper ((=) ==> (≡ₚ) ==> (≡ₚ)) (list_remove' (A:=A)).
Proof.
  move=> x _ <- l1 l2 E; apply: anti_symm.
  - by setoid_rewrite (Permutation_submseteq _ _ E).
  - by setoid_rewrite (Permutation_submseteq _ _ (symmetry E)).
Qed.
Instance list_diff_mono `{EqDecision A}
  : Proper (submseteq ++> submseteq --> submseteq) (list_multiset_difference (A:=A)).
Proof.
  have ? : Proper (submseteq ++> (=) ==> submseteq) (list_multiset_difference (A:=A)). {
    move=> l1 l1' D1 l2 _ <-; elim: l2 l1 l1' D1 => //= a l2 IH l1 l1' D1.
    apply: IH; by setoid_rewrite D1.
  }
  move=> l1 l1' D1 l2 l2' /= D2; setoid_rewrite D1.
  elim: l2' l2 / D2 l1' {D1 l1} => //=.
  - move=> a1 a2 l2 l1; f_equiv; apply: reflexive_eq; elim: l1 => //= a3 l1.
    repeat case: (decide _) => //=; congruence.
  - move=> a l2 l2' D IH l1; setoid_rewrite <- IH.
      by setoid_rewrite list_remove'_submseteq.
  - move=> ? ? ? _ IH1 _ IH2 ?; by setoid_rewrite <- IH1.
Qed.
Instance list_diff_proper `{EqDecision A}
  : Proper ((≡ₚ) ==> (≡ₚ) ==> (≡ₚ)) (list_multiset_difference (A:=A)).
Proof.
  move=> l1 l1' E1 l2 l2' E2; apply: anti_symm;
    by apply: list_diff_mono; apply: Permutation_submseteq.
Qed.
Local Instance list_diff `{EqDecision A} : Difference (list A)
  := list_multiset_difference.
Local Arguments difference {_ _} _ !_ /.
Lemma extends_remove `{EqDecision A} (l1 l2 : list A)
  : from_option (eq (l1 ∖ l2)) True (list_remove_list l2 l1).
Proof.
  elim: l2 l1 => // a l2' IH l1; cbn.
  case E: (list_remove a l1) => [l1' |] //=.
  rewrite list_remove'_alt E //=.
Qed.
Notation "'SALL'  x '∈' l ',' P " := (iter_sepcon (fun x => P) l)
  (at level 65, x pattern, right associativity) : logic.
Lemma list_diff_cancel `{EqDecision A} (l : list A) : l ∖ l = [].
Proof. elim: l => //= ? ?; rewrite decide_True //. Qed.
Lemma list_diff_action `{EqDecision A} (l l1 l2 : list A)
  : l ∖ (l1 ++ l2) = (l ∖ l1) ∖ l2.
Proof. elim: l1 l => // ? ? IH ?; cbn; rewrite IH //. Qed.
Lemma list_remove'_perm `{EqDecision A} (a : A) (l : list A)
  : a ∈ l -> l ≡ₚ a :: list_remove' a l.
Proof.
  elim: a l /.
  - move=> a l /=; rewrite decide_True //.
  - move=> a a' l In E /=; case: (decide _) => [-> | NE] //.
    rewrite {1}E; constructor.
Qed.
Lemma list_diff_perm `{EqDecision A} (l l' : list A)
  : l ⊆+ l' -> l' ≡ₚ l ++ (l' ∖ l).
Proof.
  elim: l l' => // a l IH l' Sub.
  rewrite {1}(list_remove'_perm a _ (elem_of_submseteq _ _ _ _ Sub));
    last by constructor.
  generalize (list_remove'_mono _ _ (eq_refl a) _ _ Sub).
  rewrite /= decide_True // => /IH {1}-> //.
Qed.
Definition list_mult `{EqDecision A} (a : A) (l : list A) : nat
  := count_occ (fun a1 a2 => decide (a1 = a2)) l a.
Instance list_mult_mono `{EqDecision A}
  : Proper ((=) ==> submseteq ++> le) (list_mult (A:=A)).
Proof.
  move=> ? _ <- l1 l2; elim: l1 l2 / => //= *; repeat case: (decide _) => //=; lia.
Qed.
Instance list_mult_proper `{EqDecision A}
  : Proper ((=) ==> (≡ₚ) ==> (=)) (list_mult (A:=A)).
Proof.
  move=> l1 l1' E1 l2 l2' E2; apply: (anti_symm le);
    by apply: list_mult_mono => //; apply: Permutation_submseteq.
Qed.
Lemma list_remove'_mult `{EqDecision A} (a1 a2 : A) (l : list A)
  : list_mult a1 (list_remove' a2 l) =
    if decide (a1 = a2) is left _ then Nat.pred (list_mult a1 l) else list_mult a1 l.
Proof.
  elim: l => /=; first by case: (decide _).
  move=> a' l; case: (decide _) => // [-> | NE].
  - rewrite (decide_iff _ _ _ _ (symmetry_iff _ _ _)); case: (decide _) => // NE.
    rewrite /= decide_False //.
  - case: (decide _) => [<- | NE'] /=.
    + rewrite decide_False //.
    + case: (decide _) => [<- | NE''] //= -> //.
Qed.
Lemma list_remove'_mult_eq `{EqDecision A} (a : A) (l : list A)
  : list_mult a (list_remove' a l) = Nat.pred (list_mult a l).
Proof. rewrite list_remove'_mult decide_True //. Qed.
Lemma list_remove'_mult_ne `{EqDecision A} (a1 a2 : A) (l : list A)
  : a1 <> a2 -> list_mult a1 (list_remove' a2 l) = list_mult a1 l.
Proof. move=> ?; rewrite list_remove'_mult decide_False //. Qed.
Lemma list_diff_mult `{EqDecision A} (a : A) (l1 l2 : list A)
  : list_mult a (l1 ∖ l2) = (list_mult a l1 - list_mult a l2)%nat.
Proof.
  elim: l2 l1 => //; first by move=> ? /=; lia.
  move=> a' l2 IH l1.
  rewrite /= IH list_remove'_mult (decide_iff _ _ _ _ (symmetry_iff _ _ _)).
  case: (decide _) => //=; lia.
Qed.
Lemma list_mult_elem `{EqDecision A} (a : A) (l : list A)
  : a ∈ l <-> (list_mult a l > 0)%nat.
Proof. rewrite elem_of_list_In count_occ_In //. Qed.
Require Import stdpp.gmultiset.
Lemma list_mult_not_elem `{EqDecision A} (a : A) (l : list A)
  : a ∉ l <-> list_mult a l = 0%nat.
Proof. rewrite elem_of_list_In count_occ_not_In //. Qed.
Lemma list_mult_fmap `{EqDecision A, EqDecision B} (f : A -> B) (a : A) (l : list A)
  : (list_mult a l <= list_mult (f a) (f <$> l))%nat.
Proof.
  unfold list_mult; elim: l => // a' l IH; cbn.
  case: (decide (a' = a)) => [-> | NE]; first by rewrite decide_True //; apply: le_n_S.
    by case: (decide _) => [E | NE']; cbn; first constructor.
Qed.
Lemma elements_mult `{Countable A} (a : A) (m : gmultiset A)
  : list_mult a (elements m) = multiplicity a m.
Proof.
  elim/gmultiset_ind: m => // a' m E.
  rewrite gmultiset_elements_disj_union gmultiset_elements_singleton
          multiplicity_disj_union multiplicity_singleton' /= E
          (decide_iff _ _ _ _ (symmetry_iff _ _ _)).
  case: (decide _) => //=.
Qed.
Lemma list_remove'_not_elem `{EqDecision A} (a : A) (l : list A)
  : a ∉ l -> list_remove' a l = l.
Proof. elim: l => //= ? ? IH /not_elem_of_cons [? /IH ->]; rewrite decide_False //. Qed.
Lemma list_remove'_fmap `{EqDecision A, EqDecision B} (f : A -> B) (a : A) (l : list A)
  : exists a', [/\ a ∈ l -> a' = a, f a = f a' &
             list_remove' (f a) (f <$> l) ≡ₚ f <$> list_remove' a' l].
Proof.
  elim: l; cbn; first by (exists a).
  move=> a0 l [a' [IfA Ker E_r]].
  case: (decide (a = a0)) => [<- | NE]; last case: (decide (a ∈ l)) => [In | NotIn].
  - exists a; rewrite !decide_True //; by split; first right.
  - pose proof (IfA In); subst a'.
    exists a; split=> //.
    rewrite (decide_False _ _ NE) fmap_cons -E_r.
    case: (decide _) => // <-.
    rewrite -list_remove'_perm //; by apply: elem_of_list_fmap_1.
  - case: (decide (f a = f a0)) => // [E_f | NE_f].
    + exists a0; split=> //.
      * move=> /elem_of_cons [] //.
      * rewrite decide_True //.
    + exists a'; split=> //.
      * move=> /elem_of_cons [] //.
      * case: (decide _); first by congruence.
        rewrite E_r //.
Qed.
Lemma sall_elem `{EqDecision A, ClassicalSep X} {P : A -> X} (a : A) (l : list A)
  : a ∈ l -> iter_sepcon P l = P a * iter_sepcon P (list_remove' a l).
Proof. move=> /(list_remove'_perm _) {1}-> //. Qed.
Lemma sall_submseteq `{EqDecision A, ClassicalSep X} {P : A -> X} (l l' : list A)
  : l ⊆+ l' -> iter_sepcon P l' = iter_sepcon P l * iter_sepcon P (l' ∖ l).
Proof. move=> /list_diff_perm {1}->; rewrite iter_sepcon_app //. Qed.

Definition repr_soln (g : gname) (ΘK : list (coeq * (val * val))) : mpred :=
  SALL (C, pq) ∈ ΘK, repr_coeq g C pq.
Definition repr_body (g : gname) (tsK : list (term * val)) : mpred :=
  SALL (t, p) ∈ tsK, repr_term g t p.
Instance: Proper ((=) ==> (≡ₚ) ==> (=)) repr_soln.
Proof. move=> g _ <- ΘK ΘK' Perm; rewrite /repr_soln Perm //. Qed.
Instance: Proper ((=) ==> (≡ₚ) ==> (=)) repr_body.
Proof. move=> g _ <- tsK tsK' Perm; rewrite /repr_body Perm //. Qed.

Definition zpexp : Type := list (coeq * (val * val)) * list (term * val).
Definition lengths {A B} (ls : list A * list B) : nat * nat
  := (length ls.1, length ls.2).
Definition plengths (p : pexp) : nat * nat := let (Θ, ts) := p in lengths (Θ, ts).
Definition repr_pexp (g : gname) (pK : zpexp) : mpred
  := let (ΘK, tsK) := pK in repr_soln g ΘK * repr_body g tsK.
Instance: Proper ((=) ==> prod_relation (≡ₚ) (≡ₚ) ==> (=)) repr_pexp.
Proof.
  rewrite /repr_pexp.
  move=> g _ <- [ΘK tsK] [ΘK' tsK'] [/= Perm_Θ Perm_ts].
  rewrite Perm_Θ Perm_ts //.
Qed.
Definition pexp_zip (p : pexp) (pp : list (val * val) * list val) : zpexp
  := let (Θ, ts) := p in (zip Θ pp.1, zip ts pp.2).

Lemma repr_soln_app (g : gname) (ΘK ΞK : list (coeq * (val * val)))
  : repr_soln g (ΘK ++ ΞK) = repr_soln g ΘK * repr_soln g ΞK.
Proof. rewrite /repr_soln iter_sepcon_app //. Qed.
Lemma repr_body_app (g : gname) (tsK usK : list (term * val))
  : repr_body g (tsK ++ usK) = repr_body g tsK * repr_body g usK.
Proof. rewrite /repr_body iter_sepcon_app //. Qed.
Lemma repr_coeq_swap (g : gname) (t u : term) (t_p u_p : val)
  : repr_coeq g (Coeq t u) (t_p, u_p) = repr_coeq g (Coeq u t) (u_p, t_p).
Proof. apply: sepcon_comm. Qed.
Lemma sall_key `{EqDecision A, EqDecision B, ClassicalSep X} (a : A) (l : list (A * B))
  : a ∈ l.*1 -> exists2 b, (a, b) ∈ l & forall P : A * B -> X,
        iter_sepcon P l = P (a, b) * iter_sepcon P (list_remove' (a, b) l).
Proof.
  move=> /elem_of_list_fmap [[_ b] [/= <- In]].
  exists b => // P; apply: sall_elem In.
Qed.

Instance val_vars : Vars val := fun _ => ∅.
Instance val_subst : Subst val := fun _ v => v.
Lemma varsv_def : forall v : val, vars v = ∅. Proof. done. Qed.
Lemma countv_def : forall x (v : val), count x v = 0%nat. Proof. done. Qed.
Lemma substv_def : forall σ (v : val), sim_subst σ v = v. Proof. done. Qed.
Definition varsE := (varsE, varsv_def).
Definition countE := (countE, countv_def).
Definition substE := (substE, substv_def).
Class NoVars (A : Type) `{Vars A} := no_vars : forall (a : A), vars a = ∅.
Instance val_novars : NoVars val := fun _ => erefl.
Instance list_novars `{NoVars A} : NoVars (list A).
Proof. elim=> // ? ? E; rewrite varsE E no_vars left_id //. Qed.
Instance pair_novars `{NoVars A, NoVars B} : NoVars (A * B).
Proof. move=> [a b]; rewrite varsE 2!no_vars left_id //. Qed.
Class NoSubst (A : Type) `{Subst A} := no_subst : forall σ (a : A), sim_subst σ a = a.
Instance val_nosubst : NoSubst val := fun _ _ => erefl.
Instance list_subst `{NoSubst A} : NoSubst (list A).
Proof. move=> σ; elim=> // ? ? E; rewrite substE E no_subst //. Qed.
Instance pair_nosubst `{NoSubst A, NoSubst B} : NoSubst (A * B).
Proof. move=> σ [a b]; rewrite substE 2!no_subst //. Qed.
Lemma vars_vzip `{Vars A, NoVars B} (l1 : list A) (l2 : list B)
  : (length l1 <= length l2)%nat -> vars (zip l1 l2) = vars l1.
Proof.
  elim: l1 l2 => // a l1 IH [/Nat.nle_succ_0 // | v l2 LE].
  rewrite [zip _ _]/= !varsE no_vars right_id IH //; simpl in LE; lia.
Qed.
Lemma vars_pexp_zip (p : pexp) (pp : list (val * val) * list val)
  : (length (coeqs p) <= length pp.1)%nat -> (length (terms p) <= length pp.2)%nat ->
    vars (pexp_zip p pp) = vars p.
Proof. move: p pp => [? ?] [? ?] ? ?; rewrite /= !vars_vzip //. Qed.
Lemma subst_vzip `{Subst A, NoSubst B} (σ : substitution) (l1 : list A) (l2 : list B)
  : sim_subst σ (zip l1 l2) = zip (sim_subst σ l1) l2.
Proof. elim: l1 l2 => // a l1 IH [] //= v l2; rewrite !substE IH no_subst //. Qed.
Lemma subst_pexp_zip (σ : substitution) (p : pexp) (pp : list (val * val) * list val)
  : sim_subst σ (pexp_zip p pp) = pexp_zip (sim_subst σ p) pp.
Proof. move: p pp => [? ?] [? ?]; rewrite /= !substE 2!subst_vzip //. Qed.

Lemma repr_link : forall g x p q,
    repr_term g (var x) p * repr_term g (var x) q =
    named g Full_sh x p * named g Full_sh x q *
    mt p * mt q *
    term_at (LINK, (q, nullval)) p * term_at (LINK, (p, nullval)) q.
Proof.
  move=> g x p q; rewrite !repr_term_var /repr_var; apply: pred_ext.
  - cancel; Intros v v'.
    assert_PROP [\/ p = q, q = v' | p = v'] as EO. {
      rewrite -[[\/ _, _ | _]]and_True -sepcon_prop_prop.
      apply/self_ramify_trans.
      2: rewrite sepcon_prop_prop and_True;
        apply: (@named_collision _ near near far); auto.
      cancel.
    }
    assert_PROP [\/ q = p, p = v | q = v] as EO'. {
      rewrite -[[\/ _, _ | _]]and_True -sepcon_prop_prop.
      apply/self_ramify_trans.
      2: rewrite sepcon_prop_prop and_True;
        apply: (@named_collision _ near near far); auto.
      cancel.
    }
    assert_PROP (v' = p) as E. {
      case: EO H H0 => -> // ? ?; last by apply: prop_right.
      data_at_conflict q.
    }
    assert_PROP (v = q) as E' by (case: EO' H H0 => -> // ? ?; last by apply: prop_right).
    rewrite E E' -!(named_share_join (slice.cleave_join Full_sh))
            -!(data_at_share_join _ _ _ _ _ _ (slice.cleave_join Full_sh)) -/near -/far.
    cancel.
  - cancel.
    assert_PROP (p <> q) as NE. {
      apply: not_prop_right => ->; rewrite sepcon_assoc; apply: sepcon_FF_derives'.
      apply: data_at_conflict; [auto | done].
    }
    rewrite -!(named_share_join (slice.cleave_join Full_sh))
            -!(data_at_share_join _ _ _ _ _ _ (slice.cleave_join Full_sh)) -/near -/far.
    Exists q p; cancel; entailer!.
Qed.


Theorem repr_sim_subst : forall {g u p},
  repr_term g u p |-- EX L : list (positive * val),
  !! (L.*1 ≡ₚ elements (vars u)) && ((SALL (x, q) ∈ L, repr_term g (var x) q) *
  ALL σ : substitution, (SALL (x, q) ∈ L, repr_term g (sim_subst σ (var x)) q) -*
    repr_term g (sim_subst σ u) p).
Proof.
  move=> g; elim.
  - move=> x p; Exists [(x, p)].
    rewrite /=; apply/andp_right;
      first by apply: prop_right; rewrite gmultiset_elements_singleton.
    cancel; apply/allp_right => σ.
    rewrite sepcon_emp /sim_subst /=; apply: wand_refl_cancel_right.
  - move=> p; Exists (@nil (positive * val)); simpl; entailer!.
    apply/allp_right => σ; rewrite -wand_sepcon_adjoint; cancel.
  - move=> p; Exists (@nil (positive * val)); simpl; entailer!.
    apply/allp_right => σ; rewrite -wand_sepcon_adjoint; cancel.
  - move=> l IH_l r IH_r p; rewrite [L in L |-- _]/=; Intros p_t p_u.
    sep_apply IH_l; sep_apply IH_r; Intros L_r L_l.
    (Exists (L_l ++ L_r)); apply/andp_right. {
      apply: prop_right.
      rewrite /= fmap_app gmultiset_elements_disj_union; by f_equiv.
    }
    rewrite iter_sepcon_app; cancel; apply/allp_right => σ; allp_left σ; allp_left σ.
    rewrite -wand_sepcon_adjoint /=; Exists p_t p_u; cancel.
    rewrite wand_sepcon_adjoint iter_sepcon_app; apply: wand_join.
  - move=> l IH_l r IH_r p; rewrite [L in L |-- _]/=; Intros p_t p_u.
    sep_apply IH_l; sep_apply IH_r; Intros L_r L_l.
    (Exists (L_l ++ L_r)); apply/andp_right. {
      apply: prop_right.
      rewrite /= fmap_app gmultiset_elements_disj_union; by f_equiv.
    }
    rewrite iter_sepcon_app; cancel; apply/allp_right => σ; allp_left σ; allp_left σ.
    rewrite -wand_sepcon_adjoint /=; Exists p_t p_u; cancel.
    rewrite wand_sepcon_adjoint iter_sepcon_app; apply: wand_join.
Qed.

Instance val_eq_dec : EqDecision val := Memory.EqDec_val.

Theorem repr_sim_substc : forall {g C pq},
  repr_coeq g C pq |-- EX L : list (positive * val),
  !! (L.*1 ≡ₚ elements (vars C)) && ((SALL (x, q) ∈ L, repr_term g (var x) q) *
  ALL σ : substitution, (SALL (x, q) ∈ L, repr_term g (sim_subst σ (var x)) q) -*
    repr_coeq g (sim_subst σ C) pq).
Proof.
  move=> g [l r] [p q] /=.
  sep_apply @repr_sim_subst; sep_apply @repr_sim_subst; Intros L_r L_l.
  (Exists (L_l ++ L_r)); apply/andp_right. {
    apply: prop_right.
    rewrite /= fmap_app gmultiset_elements_disj_union; by f_equiv.
  }
  rewrite iter_sepcon_app /=; cancel; apply/allp_right => σ; allp_left σ; allp_left σ.
  rewrite iter_sepcon_app; apply: wand_join.
Qed.

Theorem repr_sim_substcl : forall {g ΘK},
  repr_soln g ΘK |-- EX L : list (positive * val),
  !! (L.*1 ≡ₚ elements (vars ΘK)) && ((SALL (x, q) ∈ L, repr_term g (var x) q) *
  ALL σ : substitution, (SALL (x, q) ∈ L, repr_term g (sim_subst σ (var x)) q) -*
    repr_soln g (sim_subst σ ΘK)).
Proof.
  move=> g; elim. {
    Exists (@nil (positive * val)).
    simpl; entailer!.
    apply/allp_right => _; apply/wand_sepcon_adjoint; cancel.
  }
  move=> [C [p_t p_u]] ΘK D.
  rewrite [L in L |-- _]/=; sep_apply D; sep_apply @repr_sim_substc; Intros L_C L_Θ.
  Exists (L_C ++ L_Θ); apply/andp_right. {
    apply: prop_right.
    rewrite !varsE fmap_app 2!right_id gmultiset_elements_disj_union; by f_equiv.
  }
  rewrite iter_sepcon_app /=; cancel; apply/allp_right => σ; allp_left σ; allp_left σ.
  rewrite iter_sepcon_app sepcon_comm; apply: wand_join.
Qed.
Theorem repr_sim_substl : forall {g tsK},
  repr_body g tsK |-- EX L : list (positive * val),
  !! (L.*1 ≡ₚ elements (vars tsK)) && ((SALL (x, q) ∈ L, repr_term g (var x) q) *
  ALL σ : substitution, (SALL (x, q) ∈ L, repr_term g (sim_subst σ (var x)) q) -*
    repr_body g (sim_subst σ tsK)).
Proof.
  move=> g; elim. {
    Exists (@nil (positive * val)).
    simpl; entailer!.
    apply/allp_right => _; apply/wand_sepcon_adjoint; cancel.
  }
  move=> [t p_t] tsK D.
  rewrite [L in L |-- _]/=; sep_apply D; sep_apply @repr_sim_subst; Intros L_t L_ts.
  Exists (L_t ++ L_ts); apply/andp_right. {
    apply: prop_right.
    rewrite !varsE fmap_app right_id gmultiset_elements_disj_union; by f_equiv.
  }
  rewrite iter_sepcon_app /=; cancel; apply/allp_right => σ; allp_left σ; allp_left σ.
  rewrite iter_sepcon_app sepcon_comm; apply: wand_join.
Qed.

Theorem repr_sim_substpe : forall {g pK},
  repr_pexp g pK |-- EX L : list (positive * val),
  !! (L.*1 ≡ₚ elements (vars pK)) && ((SALL (x, q) ∈ L, repr_term g (var x) q) *
  ALL σ : substitution, (SALL (x, q) ∈ L, repr_term g (sim_subst σ (var x)) q) -*
    repr_pexp g (sim_subst σ pK)).
Proof.
  move=> g [ΘK tsK].
  rewrite [L in L |-- _]/=; sep_apply @repr_sim_substcl; sep_apply @repr_sim_substl.
  Intros L_ts L_Θ; Exists (L_Θ ++ L_ts); apply/andp_right. {
    apply: prop_right.
    rewrite /= fmap_app gmultiset_elements_disj_union; by f_equiv.
  }
  rewrite iter_sepcon_app /=; cancel; apply/allp_right => σ; allp_left σ; allp_left σ.
  rewrite iter_sepcon_app; apply: wand_join.
Qed.

Definition aux : Type := val * val.
Definition aux_cond0 (g : gname) (t : term) : val -> mpred :=
  if decide (is_nonvar t) is left _ then const emp else repr_term g t.
Definition aux_pre (g : gname) (α : action) : aux -> mpred :=
  if α is transmit t x
  then (fun '(p, q) => repr_term g (var x) p * aux_cond0 g t q)
  else const emp.
Definition aux_post (g : gname) (α : action) : aux -> mpred :=
  if α is transmit t x
  then (fun '(p, q) => repr_term g t p * aux_cond0 g t q)
  else const emp.


Instance derives_pro `{NatDed X} : PreOrder (derives (A:=X))
  := {| PreOrder_Reflexive := derives_refl; PreOrder_Transitive := derives_trans |}.

Lemma zip_app : forall {A B} {l1 l2 : list A} {k1 k2 : list B},
    length l1 = length k1 ->
    zip (l1 ++ l2) (k1 ++ k2) = zip l1 k1 ++ zip l2 k2.
Proof. move=> A B; elim=> /= [| x l1 IH] l2 [| y k1] //= k2 [/IH] -> //. Qed.

Theorem ramify_tstep : forall {g Θ Θp C Cp CS Ξ Ξp ts tsp} α,
    length Θ = length Θp -> length Ξ = length Ξp -> length ts = length tsp ->
    linear (Pexp (Θ ++ [C] ++ Ξ) ts) ->
    weakly_acyclic (Pexp (Θ ++ [C] ++ Ξ) ts) ->
    coeq_tstep α C CS ->
    repr_pexp g (pexp_zip (Pexp (Θ ++ [C] ++ Ξ) ts) (Θp ++ [Cp] ++ Ξp, tsp)) |--
    EX q : aux,
      aux_pre g α q * repr_coeq g C Cp * ALL CSp : _,
        !! (length CS = length CSp) && (repr_soln g (zip CS CSp) * aux_post g α q) -*
        repr_pexp g (pexp_zip (apply α (Pexp (Θ ++ CS ++ Ξ) ts)) (Θp ++ CSp ++ Ξp, tsp)).
Proof.
  move=> g Θ Θp C [l_p r_p] CS Ξ Ξp ts tsp α Len1 Len2 Len3 Linear WA Step.
  rewrite /= zip_app // repr_soln_app /=.
  case: α C CS / Step Linear WA.
  - move=> [t u] CS VStep Linear WA /=.
    (Exists (Vundef, Vundef)); cancel.
    apply/allp_right => CSp; rewrite -wand_sepcon_adjoint; Intros.
    rewrite !zip_app // !repr_soln_app; cancel.
  - move=> [t0 x0] t x.
    wlog [-> ->] : t0 x0 t x l_p r_p / t0 = t /\ x0 = var x. {
      move=> Gen E; inversion E; subst; first by apply: Gen.
      rewrite {1 2}[Coeq x t]coeq_swap repr_coeq_swap; by apply: Gen.
    }
    move=> _ Linear WA.
    pose p := Pexp (Θ ++ Ξ) ts; pose pp := (Θp ++ Ξp, tsp).
    transitivity (repr_coeq g (Coeq t x) (l_p, r_p) * repr_pexp g (pexp_zip p pp));
      first by rewrite /= zip_app // repr_soln_app; cancel.
    sep_apply @repr_sim_substpe; Intros L_p; rename H into E.
    allp_left ({[x:=t]} : substitution).
    have /gmultiset_elem_of_elements : x ∈ vars (pexp_zip p pp). {
      rewrite elem_of_multiplicity vars_pexp_zip
              -/(count _ _) /= ?app_length; only 1, 2: lia.
      move: (Linear x); rewrite !countE (_ : x # t); last by lia.
      move: WA => /Forall_middle [] //.
    }
    rewrite -E => /(sall_key _) [p_x In E_A].
    rewrite !E_A substE /aux_pre /aux_post /aux_cond0.
    case: (decide (is_nonvar t)) => [NV |].
    + (Exists (p_x, Vundef); rewrite /const; cancel).
      apply/allp_right => CSp; rewrite -wand_sepcon_adjoint; Intros.
      have -> : CSp = [] by case: CSp H.
      rewrite subst_pexp_zip -[pexp_zip _ _ in R in _ |-- R]/(pexp_zip (subst t x p) pp)
              [repr_soln _ _]/=; cancel.
      rewrite sepcon_assoc sepcon_comm; apply: modus_ponens_wand'.
      cancel; apply/derives_refl'/iter_sepcon_func_strong => -[y p_y] /elem_of_list_In In'.
      rewrite subst_var_ne //; contradict In'; subst y.
      move=> /(elem_of_list_fmap_1 fst) /=.
      case: (list_remove'_fmap fst (x, p_x) L_p) => -[x' p_x'] [/(_ In) [-> ->] _ /= <-].
      rewrite -/(_ ∉ _) list_mult_not_elem list_remove'_mult_eq E elements_mult
              vars_pexp_zip -/(count _ _) /= ?app_length; only 1, 2: lia.
      move: (Linear x); rewrite /p /= !countE; lia.
    + case: t Linear WA => // y Linear WA _.
      have NE : x <> y. {
        move=> ?; subst y.
        move: WA => /Forall_middle [_] /=; rewrite countE //.
      }
      have : y ∈ (list_remove' (x, p_x) L_p).*1. {
        case: (list_remove'_fmap fst (x, p_x) L_p) => -[x' p_x'] [/(_ In) [-> ->] _ /= <-].
        rewrite list_mult_elem list_remove'_mult_ne // E elements_mult
                vars_pexp_zip -/(count _ _) /= ?app_length; only 1, 2: lia.
        move: (Linear y); rewrite /p /= !countE decide_False //; lia.
      }
      move=> /(sall_key _) [p_y In' E_A'].
      rewrite !E_A' substE decide_False //.
      Exists (p_x, p_y); cancel.
      apply/allp_right => CSp; rewrite -wand_sepcon_adjoint; Intros.
      have -> : CSp = [] by case: CSp H.
      rewrite subst_pexp_zip -[pexp_zip _ _ in R in _ |-- R]/(pexp_zip (subst _ x p) pp)
              [repr_soln _ _]/=; cancel.
      rewrite 2!sepcon_assoc sepcon_comm; apply: modus_ponens_wand'.
      cancel; apply/derives_refl'/iter_sepcon_func_strong => -[z p_z] /elem_of_list_In In3.
      rewrite subst_var_ne //; contradict In3; subst z.
      enough ((x, p_z) ∉ list_remove' (x, p_x) L_p) as NotIn. {
        contradict NotIn.
        apply/(elem_of_submseteq _ _ _ NotIn)/list_remove'_submseteq.
      }
      move=> /(elem_of_list_fmap_1 fst).
      case: (list_remove'_fmap fst (x, p_x) L_p) => -[x' p_x'] [/(_ In) [-> ->] _ /= <-].
      rewrite -/(_ ∉ _) list_mult_not_elem list_remove'_mult_eq E elements_mult
              vars_pexp_zip -/(count _ _) /= ?app_length; only 1, 2: lia.
      move: (Linear x); rewrite /p /= !countE; lia.
Qed.

(* ⊢ ; (x ⅋ ( * )) ⅋ (( * ⊗ * ) ⊗ x) : α ⊗ 1 ⊸ (1 ⊗ 1) ⊗ α *)
Definition demo_term : term :=
  (tpar (tpar (var 1) costar) (ttens (ttens star star) (var 1)))%positive.
Definition demo_ty : ty :=
  (lto (tens (atom 1) one) (tens (tens one one) (atom 1)))%positive.
Fact demo_typing : have_tys [] [(demo_term, demo_ty)].
Proof.
  do 2 constructor.
  apply: xchg_term. { rewrite perm_swap [R in _ :: R]perm_swap //. }
  constructor.
  apply: (_tens (Θ:=[]) (Ξ:=[]) (Γ:=[])).
  - rewrite !varsE /= !left_id.
    apply: sets.disjoint_empty_l.
  - apply: (_tens (Θ:=[]) (Ξ:=[]) (Γ:=[])).
    + rewrite !varsE /= !left_id.
      apply: sets.disjoint_empty_l.
    + constructor.
    + constructor.
  - change (atom 1)%positive with (neg (neg_atom 1))%positive.
    constructor.
Qed.

(* copied from VST *)
Definition surely_malloc_spec :=
  DECLARE _surely_malloc
   WITH t : type, gv : globals
   PRE [ _n OF size_t ]
       PROP (0 <= sizeof t <= Ptrofs.max_unsigned;
                complete_legal_cosu_type t = true;
                natural_aligned natural_alignment t = true)
       LOCAL (temp _n (Vptrofs (Ptrofs.repr (sizeof t))); gvars gv)
       SEP (mem_mgr gv)
    POST [ tptr tvoid ] EX p:_,
       PROP ()
       LOCAL (temp ret_temp p)
       SEP (mem_mgr gv; malloc_token Ews t p * data_at_ Ews t p).

(* TODO this spec is FALSE---fix it & then prove it!!!
   I think we need some kind of linearity constraint or something. *)
Definition free_rec_spec :=
  DECLARE _free_rec
    WITH g : gname, t : term, p : val, gv : globals
    PRE [_t OF Node]
      PROP ()
      LOCAL (temp _t p; gvars gv)
      SEP (repr_term g t p; mem_mgr gv)
    POST [tvoid]
      PROP ()
      LOCAL ()
      SEP (mem_mgr gv).

(* Specs for constructors. *)

(* crap, need prophecy variables or something? *)
Definition reserved (g : gname) (x : positive) : mpred :=
  EX p : _, EX q : _, named g Full_sh x p * named g Full_sh x q.
(* I'm not sure that this spec is provable; need to reformulate it somehow. *)
Definition mklink_spec :=
  DECLARE _mklink
    WITH g : gname, sh_a : share, sh_b : share, pp : val, qq : val,
         x : positive, gv : globals
    PRE [_a OF tptr Node, _b OF tptr Node]
      PROP (writable_share sh_a; writable_share sh_b)
      LOCAL (temp _a pp; temp _b qq; gvars gv)
      SEP (reserved g x; data_at_ sh_a Node pp; data_at_ sh_b Node qq; mem_mgr gv)
    POST [tvoid] EX p : _, EX q : _,
      PROP (isptr p; isptr q) (* TODO why is this necessary? ...IS it necessary? *)
      LOCAL ()
      SEP (data_at sh_a Node p pp; data_at sh_b Node q qq;
           repr_term g (var x) p; repr_term g (var x) q; mem_mgr gv).
Definition unit_spec :=
  DECLARE _unit
    WITH g : gname, gv : globals
    PRE []
      PROP ()
      LOCAL (gvars gv)
      SEP (mem_mgr gv)
    POST [Node] EX p : _,
      PROP ()
      LOCAL (temp ret_temp p)
      SEP (repr_term g star p; mem_mgr gv).
Definition counit_spec :=
  DECLARE _counit
    WITH g : gname, gv : globals
    PRE []
      PROP ()
      LOCAL (gvars gv)
      SEP (mem_mgr gv)
    POST [Node] EX p : _,
      PROP ()
      LOCAL (temp ret_temp p)
      SEP (repr_term g costar p; mem_mgr gv).
Definition tens_spec :=
  DECLARE CPE2._tens
    WITH g : gname, p_t : val, p_u : val, t : term, u : term, gv : globals
    PRE [_l OF Node, _r OF Node]
      PROP ()
      LOCAL (temp _l p_t; temp _r p_u; gvars gv)
      SEP (repr_term g t p_t; repr_term g u p_u; mem_mgr gv)
    POST [Node] EX p : _,
      PROP ()
      LOCAL (temp ret_temp p)
      SEP (repr_term g (ttens t u) p; mem_mgr gv).
Definition par_spec :=
  DECLARE CPE2._par
    WITH g : gname, p_t : val, p_u : val, t : term, u : term, gv : globals
    PRE [_l OF Node, _r OF Node]
      PROP ()
      LOCAL (temp _l p_t; temp _r p_u; gvars gv)
      SEP (repr_term g t p_t; repr_term g u p_u; mem_mgr gv)
    POST [Node] EX p : _,
      PROP ()
      LOCAL (temp ret_temp p)
      SEP (repr_term g (tpar t u) p; mem_mgr gv).


Definition copy1_spec :=
  DECLARE _copy1
    WITH sh_a : share, sh_b : share, cl : val, f : val, s : val, p : val, q : val
    PRE [_a OF Node, _b OF Node]
      PROP (writable_share sh_a; readable_share sh_b;
            tc_val tint cl; tc_val Node f; tc_val Node s)
      LOCAL (temp _a p; temp _b q)
      SEP (data_at_ sh_a Term p; data_at sh_b Term (cl, (f, s)) q)
    POST [tvoid]
      PROP ()
      LOCAL ()
      SEP (data_at sh_a Term (cl, (f, s)) p; data_at sh_b Term (cl, (f, s)) q).

(* this is false, I think---we need some kind of full acyclicity *)
Definition cut_spec :=
  DECLARE CPE2._cut
    WITH g : gname, Θ : soln, Θp : list (val * val),
                    C : coeq, Cp : val * val,
                    Ξ : soln, Ξp : list (val * val),
         ts : list term, tsp : list val, gv : globals
    PRE [_t OF Node, _u OF Node]
      PROP (length Θ = length Θp; length Ξ = length Ξp; length ts = length tsp;
            linear (Pexp (Θ ++ [C] ++ Ξ) ts); weakly_acyclic (Pexp (Θ ++ [C] ++ Ξ) ts))
      LOCAL (temp _t Cp.1; temp _u Cp.2; gvars gv)
      SEP (repr_pexp g (pexp_zip (Pexp (Θ ++ [C] ++ Ξ) ts) (Θp ++ [Cp] ++ Ξp, tsp));
           mem_mgr gv)
    POST [tvoid] EX p' : _,
      PROP (tsteps (Pexp (Θ ++ [C] ++ Ξ) ts) p')
      LOCAL ()
      SEP (repr_pexp g (pexp_zip p' (Θp ++ Ξp, tsp)); mem_mgr gv).


Definition Gprog : funspecs :=
  ltac:(with_library prog
    [surely_malloc_spec;
     mklink_spec; unit_spec; counit_spec; tens_spec; par_spec;
     copy1_spec; cut_spec]).


Lemma body_copy1 : semax_body Vprog Gprog f_copy1 copy1_spec.
Proof. start_function; repeat forward. Qed.

Theorem body_cut : semax_body Vprog Gprog f_cut cut_spec.
Proof.
  start_function.
  rename H2 into Linear; rename H3 into WA; destruct C as (t, u), Cp as (p_t, p_u).
  set zp := pexp_zip _ _.
  rewrite /= zip_app // repr_soln_app /= read_tag [repr_term _ u _]read_tag.
  Intros; forward; forward.
  do 2 (gather_SEP (field_at _ _ _ _ _) (_ -* _); rewrite /= -read_tag).
  gather_SEP (repr_soln g (zip Θ _)) (repr_term g t _) (repr_term g u _)
             (repr_soln g (zip Ξ _)) (repr_body _ _); simpl.
  replace (_ * _) with (repr_pexp g zp)
    by rewrite /= zip_app // repr_soln_app /= -!sepcon_assoc //.
  set cond := if decide (is_nonvar t /\ is_nonvar u) is left _
              then 0 else 1.
  forward_if (temp _t'1 (Vint (Int.repr cond))); last forward_if.
  - forward; entailer!.
    rewrite {}/cond; case: (decide _) => // -[]; case: t H2 {Linear WA zp} => //.
  - forward; entailer!.
    rewrite {}/cond; case: t H2 {Linear WA zp} => //; case: u H3 => //.
  - move: H2'; rewrite {1}/cond; case: (decide _) => // /not_and_l NV _.
    set v := if t is var _ then u else t.
    set x := if t is var x' then x' else if u is var x' then x' else 1%positive.
    set α := transmit v x.
    unfold zp; sep_apply (@ramify_tstep g Θ Θp (Coeq t u) (p_t, p_u) [] Ξ Ξp ts tsp α). {
      rewrite /α /v /x.
      case: NV.
      - destruct t => //; constructor; constructor.
      - destruct u => //; constructor; destruct t; constructor.
    }
    (Intros q); match goal with |- context [allp ?P] =>
                                sep_apply (allp_instantiate P (@nil (val * val))) end.
    case: q => p_x p_v.
    rewrite true_eq // TT_andp emp_sepcon 2![[] ++ _]/=.
    rewrite [aux_pre _ _ _]/= [aux_post _ _ _]/=.
Abort.

(* copied from VST *)
Lemma body_surely_malloc : semax_body Vprog Gprog f_surely_malloc surely_malloc_spec.
Proof.
  start_function.
  forward_call (* p = malloc(n); *)
     (t, gv).
  Intros p.
  forward_if
  (PROP ( )
   LOCAL (temp _p p)
   SEP (mem_mgr gv; malloc_token Ews t p * data_at_ Ews t p)).
*
  if_tac.
    subst p. entailer!.
    entailer!.
*
    forward_call tt.
    contradiction.
*
    if_tac.
    + forward. subst p. congruence.
    + Intros. forward. entailer!.
*
  forward. Exists p; entailer!.
Qed.

Definition link_naming_insert (adr : address) (sh : share) (x : positive)
           (m : link_naming) : link_naming :=
  if dec_share_nonunit sh is left Non
  then psepalg.insert_fpm _ adr (psepalg.mk_lifted sh Non, x) m
  else m.
Lemma body_mklink : semax_body Vprog Gprog f_mklink mklink_spec.
Proof.
  start_function.
  forward_call (Term, gv); [done | Intros p]; forward; forward; forward.
  forward_call (Term, gv); [done | Intros q]; forward; forward; forward.
  rewrite -/Full_sh -/LINK -[Vlong _]/nullval.
  assert_PROP (exists p_a, val2adr p p_a) as H_p. {
    entailer!; destruct p; try contradiction; simpl; eauto. }
  assert_PROP (exists q_a, val2adr q q_a) as H_q. {
    entailer!; destruct q; try contradiction; simpl; eauto. }
  move: H_p H_q => [p_a ?] [q_a ?].
  viewshift_SEP 5 (named g Full_sh x p * named g Full_sh x q). {
    rewrite /reserved; Intros p0 q0; go_lower.
    rewrite (_ : p0 = p); last rewrite (_ : q0 = q); last by apply: bupd_intro.
    (* spec is wrong *)
    - give_up.
    - give_up.
  }
  Intros; do 5 forward; Exists p q; entailer!.
  rewrite repr_link -/Full_sh; cancel.
Abort.
Lemma body_unit : semax_body Vprog Gprog f_unit unit_spec.
Proof.
  start_function.
  forward_call (Term, gv); [done | Intros p].
  forward; forward; forward; forward; Exists p.
  rewrite /= -/Full_sh; entailer!.
Qed.
Lemma body_counit : semax_body Vprog Gprog f_counit counit_spec.
Proof.
  start_function.
  forward_call (Term, gv); [done | Intros p].
  forward; forward; forward; forward; Exists p; entailer!.
  rewrite /= -/Full_sh; entailer!.
Qed.
Lemma body_tens : semax_body Vprog Gprog f_tens tens_spec.
Proof.
  start_function.
  forward_call (Term, gv); [done | Intros p].
  forward; forward; forward; forward; forward; forward; Exists p; entailer!.
  rewrite /= -/Full_sh; Exists p_t; Exists p_u; entailer!.
Qed.
Lemma body_par : semax_body Vprog Gprog f_par par_spec.
Proof.
  start_function.
  forward_call (Term, gv); [done | Intros p].
  forward; forward; forward; forward; forward; forward; Exists p; entailer!.
  rewrite /= -/Full_sh; Exists p_t; Exists p_u; entailer!.
Qed.
